<?php

/**
 * 
 * The Affiliate Controller provides the main interface that the rest of the application will use to execute methods on the
 * active affiliate provider
 *
 */
class MM_AffiliateController
{
    public static $AFFILIATE_EVENT_TRACK_INITIAL_COMMISSION = "mm_affiliate_track_initial_commission";
    public static $AFFILIATE_EVENT_TRACK_REBILL_COMMISSION = "mm_affiliate_track_rebill_commission";
    public static $AFFILIATE_EVENT_REVERSE_COMMISSION = "mm_affiliate_reverse_commission";
    public static $AFFILIATE_EVENT_CREATE_ACCOUNT = "mm_affiliate_create_account";

    public static function setup()
    {
        $ac = new MM_AffiliateController();
        add_action(MM_Event::$MEMBER_ADD, array($ac, 'createAffiliateAccount'));
        add_action(MM_Event::$PAYMENT_RECEIVED, array($ac, 'trackCommission'));
        add_action(MM_Event::$PAYMENT_REBILL, array($ac, 'trackRebillCommission'));
        add_action(MM_Event::$REFUND_ISSUED, array($ac, 'reverseCommission'));
    }

    public function trackCommission($data, $isRebill = false)
    {
        $orderNumber = $this->getOrderNumber($data);
        $affiliateId = isset($data["order_affiliate_id"]) ? $data["order_affiliate_id"] : "";

        $commissionProfile = $this->getCommissionProfile($data);
        $this->_trackCommission($commissionProfile, $affiliateId, $orderNumber, $data, $isRebill);

        // check if there are any partner payouts
        $partners = $this->getPartnerPayouts($data);

        foreach ($partners as $partner) {
            $commissionProfile = new MM_CommissionProfile($partner->commissionProfileId);

            if (!$commissionProfile->isValid()) {
                $commissionProfile = MM_CommissionProfile::getDefaultCommissionProfile();
            }

            $this->_trackCommission($commissionProfile, $partner->affiliateId, $orderNumber . "-{$partner->affiliateId}", $data, $isRebill);
        }
    }

    public function trackRebillCommission($data)
    {
        $this->trackCommission($data, true);
    }

    private function _trackCommission(MM_CommissionProfile $commissionProfile, $affiliateId, $orderNumber, $data, $isRebill)
    {
        $affiliateProvider = MM_AffiliateProviderFactory::getActiveProvider();
        $couponCode = "";
        $doTrackCommission = false;

        // check if commission should be tracked based on if initial or rebill commissions
        // are enabled on the profile, and if this is a rebill payment, if there is a limit
        // and if it has been reached.
        if (!$isRebill && $commissionProfile->initialCommissionEnabled()) {
            // check if commission should be tracked based on if there is an affiliate ID
            if (!empty($affiliateId)) {
                $doTrackCommission = true;
            }

            // check if commission should be tracked based on if there is a coupon associated with the order
            if (
                $affiliateProvider->supportsFeature(MM_AffiliateProviderFeatures::COMMISSION_TRACKING_BY_COUPON)
                && isset($data["order_coupons"])
            ) {
                $coupons = json_decode(stripslashes($data["order_coupons"]));

                if (is_array($coupons) && count($coupons) > 0) {
                    $doTrackCommission = true;

                    foreach ($coupons as $coupon) {
                        $couponCode = $coupon->code;

                        // only supports one coupon right now
                        break;
                    }
                }
            }

            // check if commission should be tracked based on IP address
            if (
                $affiliateProvider->supportsFeature(MM_AffiliateProviderFeatures::COMMISSION_TRACKING_BY_IP_ADDRESS)
                && $affiliateProvider->getIPTrackingEnabled() == "1"
            ) {
                $doTrackCommission = true;
            }
        } else if ($isRebill && $commissionProfile->rebillCommissionsEnabled()) {
            if ($commissionProfile->doLimitRebills()) {
                // check if rebill limit exceeded
                global $wpdb;
                $totalRebills = 0;

                $provider_entity = new MM_AffiliateProvider();
                $provider_entity->setToken($affiliateProvider->getToken());
                $provider_entity->getData();

                $sql = "SELECT count(*) as totalRebills FROM " . MM_TABLE_AFFILIATE_REBILL_COMMISSIONS . " WHERE ";
                $sql .= "affiliate_provider_id='{$provider_entity->getId()}' AND affiliate_id='{$affiliateId}' ";
                $sql .= "AND order_number='{$data["order_number"]}'";
                $row = $wpdb->get_row($sql);

                if (isset($row->totalRebills)) {
                    $totalRebills = intval($row->totalRebills);
                }

                if ($totalRebills < $commissionProfile->getRebillCommissionLimit()) {
                    $doTrackCommission = true;
                }
            } else {
                $doTrackCommission = true;
            }
        }

        if ($doTrackCommission) {
            $orderTotal = floatval($data["order_total"]) - floatval($data["order_shipping"]);

            $result = $affiliateProvider->trackCommission($commissionProfile, $affiliateId, $orderNumber, $orderTotal, $data, $isRebill);

            // track rebill commissions
            if (MM_Response::isSuccess($result) && $isRebill) {
                global $wpdb;

                $provider_entity = new MM_AffiliateProvider();
                $provider_entity->setToken($affiliateProvider->getToken());
                $provider_entity->getData();

                $rebillData = array();
                $rebillData['affiliate_provider_id'] = $provider_entity->getId();
                $rebillData['affiliate_id'] = $affiliateId;
                $rebillData['order_number'] = $data["order_number"];
                $rebillData['transaction_id'] = $data["order_transaction_id"];

                $wpdb->insert(MM_TABLE_AFFILIATE_REBILL_COMMISSIONS, $rebillData);
            }

            if (MM_Response::isSuccess($result)) {
                // dispatch push notifications
                if (!$isRebill) {
                    MM_Event::safeDoAction(MM_Event::$COMMISSION_INITIAL, MM_Event::packageAffiliateData($commissionProfile, $affiliateId, $orderNumber, $orderTotal, $data));
                } else {
                    MM_Event::safeDoAction(MM_Event::$COMMISSION_REBILL, MM_Event::packageAffiliateData($commissionProfile, $affiliateId, $orderNumber, $orderTotal, $data, true));
                }

                // log activity
                $user = new MM_User($data["member_id"]);

                if ($user->isValid()) {
                    $logParams = array();
                    $logParams[MM_ActivityLog::$PARAM_AFFILIATE_PROVIDER_NAME] = $affiliateProvider->getName();
                    $logParams[MM_ActivityLog::$PARAM_AFFILIATE_ORDER_NUMBER] = $orderNumber;
                    $logParams[MM_ActivityLog::$PARAM_AFFILIATE_ORDER_TOTAL] = $orderTotal;
                    $logParams[MM_ActivityLog::$PARAM_AFFILIATE_ORDER_IP_ADDRESS] = $data["order_ip_address"];

                    if (!$isRebill) {
                        $logParams[MM_ActivityLog::$PARAM_AFFILIATE_TRACKING_EVENT] = self::$AFFILIATE_EVENT_TRACK_INITIAL_COMMISSION;
                    } else {
                        $logParams[MM_ActivityLog::$PARAM_AFFILIATE_TRACKING_EVENT] = self::$AFFILIATE_EVENT_TRACK_REBILL_COMMISSION;
                    }

                    if (!empty($couponCode)) {
                        $logParams[MM_ActivityLog::$PARAM_AFFILIATE_COUPON_CODE] = $couponCode;
                    }

                    if (!empty($affiliateId)) {
                        $logParams[MM_ActivityLog::$PARAM_AFFILIATE_ID] = $affiliateId;
                    }

                    $logParams[MM_ActivityLog::$PARAM_AFFILIATE_TRACKING_URL] = $result->getData("url");

                    if ($affiliateProvider->getToken() != MM_AffiliateProvider::$DEFAULT_TOKEN) {
                        MM_ActivityLog::log($user, MM_ActivityLog::$EVENT_TYPE_AFFILIATE_TRACKING, $logParams);
                    }
                }
            }
        }
    }

    public function reverseCommission($data)
    {
        $affiliateProvider = MM_AffiliateProviderFactory::getActiveProvider();
        $orderNumber = $this->getOrderNumber($data);
        $commissionProfile = $this->getCommissionProfile($data);

        $logParams = array();
        $logParams[MM_ActivityLog::$PARAM_AFFILIATE_PROVIDER_NAME] = $affiliateProvider->getName();
        $logParams[MM_ActivityLog::$PARAM_AFFILIATE_TRACKING_EVENT] = self::$AFFILIATE_EVENT_REVERSE_COMMISSION;

        if ($commissionProfile->doReverseCommissions()) {
            $affiliateId = $data["order_affiliate_id"];
            $result = $affiliateProvider->reverseCommission($affiliateId, $orderNumber);

            if (MM_Response::isSuccess($result)) {
                // dispatch push notification
                do_action(MM_Event::$CANCEL_COMMISSION, MM_Event::packageAffiliateData($commissionProfile, $affiliateId, $orderNumber, "", $data));

                // log activity
                $user = new MM_User($data["member_id"]);

                if ($user->isValid()) {
                    $logParams[MM_ActivityLog::$PARAM_AFFILIATE_ORDER_NUMBER] = $orderNumber;
                    $logParams[MM_ActivityLog::$PARAM_AFFILIATE_TRACKING_URL] = $result->getData("url");

                    if ($affiliateProvider->getToken() != MM_AffiliateProvider::$DEFAULT_TOKEN) {
                        MM_ActivityLog::log($user, MM_ActivityLog::$EVENT_TYPE_AFFILIATE_TRACKING, $logParams);
                    }
                }
            }
        }

        // check if there are any partner payouts
        $partners = $this->getPartnerPayouts($data);

        foreach ($partners as $partner) {
            $commissionProfile = new MM_CommissionProfile($partner->commissionProfileId);

            if ($commissionProfile->isValid() && $commissionProfile->doReverseCommissions()) {
                $result = $affiliateProvider->reverseCommission($partner->affiliateId, $orderNumber . "-{$partner->affiliateId}");

                if (MM_Response::isSuccess($result)) {
                    // dispatch push notification
                    do_action(MM_Event::$CANCEL_COMMISSION, MM_Event::packageAffiliateData($commissionProfile, $partner->affiliateId, $orderNumber . "-{$partner->affiliateId}", "", $data));

                    // log activity
                    $user = new MM_User($data["member_id"]);

                    if ($user->isValid()) {
                        $logParams[MM_ActivityLog::$PARAM_AFFILIATE_ORDER_NUMBER] = $orderNumber . "-{$partner->affiliateId}";
                        $logParams[MM_ActivityLog::$PARAM_AFFILIATE_TRACKING_URL] = $result->getData("url");

                        if ($affiliateProvider->getToken() != MM_AffiliateProvider::$DEFAULT_TOKEN) {
                            MM_ActivityLog::log($user, MM_ActivityLog::$EVENT_TYPE_AFFILIATE_TRACKING, $logParams);
                        }
                    }
                }
            }
        }
    }

    public function createAffiliateAccount($data)
    {
        $affiliateProvider = MM_AffiliateProviderFactory::getActiveProvider();

        if ($affiliateProvider->supportsFeature(MM_AffiliateProviderFeatures::AFFILIATE_ACCOUNT_CREATION)) {
            $result = $affiliateProvider->createAffiliateAccount($data);

            if (MM_Response::isSuccess($result)) {
                // log activity
                $user = new MM_User($data["member_id"]);

                if ($user->isValid()) {
                    $logParams = array();
                    $logParams[MM_ActivityLog::$PARAM_AFFILIATE_PROVIDER_NAME] = $affiliateProvider->getName();
                    $logParams[MM_ActivityLog::$PARAM_AFFILIATE_TRACKING_EVENT] = self::$AFFILIATE_EVENT_CREATE_ACCOUNT;
                    $logParams[MM_ActivityLog::$PARAM_AFFILIATE_TRACKING_URL] = $result->getData("url");

                    if ($affiliateProvider->getToken() != MM_AffiliateProvider::$DEFAULT_TOKEN) {
                        MM_ActivityLog::log($user, MM_ActivityLog::$EVENT_TYPE_AFFILIATE_TRACKING, $logParams);
                    }

                    // send affiliate welcome email
                    if ($affiliateProvider->getSendAffiliateWelcomeEmail() == "1") {
                        $welcomeEmailSubject = $affiliateProvider->getAffiliateWelcomeEmailSubject();
                        $welcomeEmailBody = $affiliateProvider->getAffiliateWelcomeEmailBody();

                        // replace affiliate SmartTags 
                        $affiliateData = $result->getData("affiliateData");

                        $email = new MM_Email();
                        $employee = MM_Employee::getDefault();
                        $context = new MM_Context($user, $employee);

                        if ($affiliateData instanceof MM_AffiliateVO) {
                            $context->setAffiliate($affiliateData);
                        }

                        $email->setContext($context);
                        $email->setSubject($welcomeEmailSubject);
                        $email->setBody($welcomeEmailBody);

                        $email->setToName($user->getFullName());
                        $email->setToAddress($user->getEmail());
                        $email->setFromName($employee->getDisplayName());
                        $email->setFromAddress($employee->getEmail());
                        $response = $email->send();

                        if (MM_Response::isError($response)) {
                            // do nothing
                        }
                    }
                }
            }
        }
    }

    private function getOrderNumber($orderData)
    {
        $orderNumber = $orderData["order_number"];

        if (!empty($orderData["order_transaction_id"])) {
            $orderNumber .= "-{$orderData["order_transaction_id"]}";
        }

        return $orderNumber;
    }

    private function getCommissionProfile($orderData)
    {
        $commissionProfile = new MM_CommissionProfile();

        if (isset($orderData["order_products"])) {
            $products = json_decode($orderData["order_products"]);

            foreach ($products as $product) {
                $p = new MM_Product($product->id);

                if ($p->isValid()) {
                    return $p->getCommissionProfile();
                }
            }
        }

        return $commissionProfile;
    }

    private function getPartnerPayouts($orderData)
    {
        $partners = array();

        if (isset($orderData["order_products"])) {
            $products = json_decode($orderData["order_products"]);

            foreach ($products as $item) {
                $product = new MM_Product($item->id);

                if ($product->isValid()) {
                    $partners = $product->getPartners();

                    // only supports one product right now
                    break;
                }
            }
        }

        return $partners;
    }
}
