<?php

/**
 *
 * MemberMouse(TM) (http://www.membermouse.com)
 * (c) MemberMouse, LLC. All rights reserved.
 */

abstract class MM_ActionDispatchEngineBase
{

    protected $actions = null;
    protected $params = array();

    public function processNotification(MM_Action $action)
    {
        switch ($action->getActionType()) {
            case MM_Action::$MM_ACTION_SEND_EMAIL:
                $this->action_sendEmail($action);
                break;

            case MM_Action::$MM_ACTION_NOTIFY_ZAPIER:
                $this->action_notifyZapier($action);
                break;

            case MM_Action::$MM_ACTION_CALL_SCRIPT:
                $this->action_callScriptUrl($action);
                break;
        }
    }

    protected function performActions($useQueue = true)
    {
        if (is_array($this->actions)) {
            foreach ($this->actions as $action) {
                if ($action->isValid() && $action->isActive()) {
                    // check action attributes and process the notification
                    if ($this->checkEventAttributes($action)) {
                        $this->processNotification($action);
                    }
                }
            }
        }
    }


    protected abstract function checkEventAttributes(MM_Action $action);


    protected function action_sendEmail($action)
    {

        $context = $this->createEmailContext($action);
        if ($context) {
            $actionValue = $action->getActionValue();
            $email = $this->prepareEmail($context, $actionValue);
            $email->send();
            return new MM_Response();
        } else {
            return new MM_Response("Error sending push notification email.", MM_Response::$ERROR);
        }
    }


    protected function action_notifyZapier($action)
    {
        $eventTypeName = MM_Event::getName($action->getEventType());
        $actionValue = $action->getActionValue();
        $user = new MM_User($this->params["member_id"]);
        $fromEmployee = new MM_Employee($actionValue["emailFromId"]);
        $context = new MM_Context($user, $fromEmployee);

        $body = "";

        // TODO review this mechanism for effectiveness in parsing content
        // the contents of $this->params is generated by one of the package...() functions in the MM_Event class
        foreach ($this->params as $key => $value) {
            if (!empty($value)) {
                $decodedValue = json_decode($value);

                if (!is_array($decodedValue)) {
                    $body .= "{$key}: {$value}";
                } else if (!empty($decodedValue)) {
                    foreach ($decodedValue as $arrIndex => $arrValue) {
                        $body .= "{$key}[{$arrIndex}]: ";

                        if (is_object($arrValue)) {
                            $body .= "\n";

                            foreach ($arrValue as $objKey => $objValue) {
                                $body .= "-{$arrIndex}- {$objKey}: {$objValue}\n";
                            }
                        } else {
                            $body .= "{$arrValue}\n";
                        }
                    }
                }
            } else {
                $body .= "{$key}:";
            }

            $body .= "\n";
        }

        if (!empty($actionValue["zapierAdditionalInfo"])) {
            $body .= "\n================= ADDITIONAL INFO =================\n";
            $body .= $actionValue["zapierAdditionalInfo"];
        }

        $email = new MM_Email();
        $email->setContext($context);
        $email->setSubject("Zapier Notification ({$eventTypeName})");
        $email->setBody($body);
        $email->setFromName($fromEmployee->getDisplayName());
        $email->setFromAddress($fromEmployee->getEmail());

        $email->setToName("Zapier");
        $email->setToAddress($actionValue["zapierMailbox"]);

        $email->send();
        return new MM_Response();
    }


    protected function action_callScriptUrl($action)
    {
        $actionValue = $action->getActionValue();

        $url = "";

        if (is_array($actionValue) && isset($actionValue["scriptUrl"])) {
            $url = $actionValue["scriptUrl"];
        }

        if (!empty($url)) {
            $postvars = "event_type={$action->getEventType()}&";
            $postvars .= $this->arrayToQuerystring($this->params);

            if (preg_match("/(\?)/", $url)) {
                $url = $url . "&" . $postvars;
            } else {
                $url = $url . "?" . $postvars;
            }

            $ch = curl_init($url);
            curl_setopt($ch, CURLOPT_USERAGENT, MM_CURL_USER_AGENT);
            curl_setopt($ch, CURLOPT_POST, 0);
            curl_setopt($ch, CURLOPT_HEADER, 0);
            curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
            curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
            curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
            curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 1);
            curl_setopt($ch, CURLOPT_TIMEOUT, 2); //timeout in seconds
            $contents = curl_exec($ch);
            curl_close($ch);
        }
    }


    abstract protected function createEmailContext(MM_Action $action);


    protected function prepareEmail(MM_Context $context, array $actionValue)
    {
        $email = new MM_Email();
        $email->setContext($context);
        $email->setSubject($actionValue["emailSubject"]);
        $email->setBody($actionValue["emailBody"]);

        $fromEmployee = $context->getEmployee();
        $email->setFromName($fromEmployee->getDisplayName());
        $email->setFromAddress($fromEmployee->getEmail());

        $this->addCCRecipients($email, $actionValue);
        $this->setEmailRecipient($email, $actionValue, $context->getUser());

        return $email;
    }


    protected function addCCRecipients(MM_Email $email, array $actionValue)
    {
        if (isset($actionValue["emailCC"]) && !empty($actionValue["emailCC"])) {
            $ccEmails = explode(",", $actionValue["emailCC"]);
            foreach ($ccEmails as $ccEmail) {
                $email->addCC(trim($ccEmail));
            }
        }
    }


    protected function setEmailRecipient(MM_Email $email, array $actionValue, MM_User $user)
    {
        if ($actionValue["emailToId"] == MM_Action::$CURRENT_MEMBER_PLACEHOLDER) {
            $email->setToName($user->getFirstName());
            $email->setToAddress($user->getEmail());
        } else {
            $toEmployee = new MM_Employee($actionValue["emailToId"]);
            $email->setToName($toEmployee->getDisplayName());
            $email->setToAddress($toEmployee->getEmail());
        }
    }


    protected function arrayToQuerystring($params)
    {
        $querystring = "";
        foreach ($params as $key => $value) {
            if (!is_array($value)) {
                $value = ($value != null) ? $value : "";
                $querystring .= $key . "=" . urlencode($value) . "&";
            }
        }
        return $querystring;
    }
}
