<?php

use membermouse\resque\ScheduledPaymentJob;
use membermouse\resque\Resque;

/**
 *
 * MemberMouse(TM) (http://www.membermouse.com)
 * (c) MemberMouse, LLC. All rights reserved.
 */

class MM_LocalBillingEngine
{
    public static $backupDetectionInterval = "24 hours"; //the interval after which we assume a backup might have been restored
    public static $backupDetectionSampleSize = 5; //sample this many transactions

    public function __construct() {}


    /**
     * backupRestoreDetected checks the current time against the time of the last billing sweep. If this time exceeds the defined interval,
     * then a sample of the earliest pending transactions is taken and the payment gateways are queried to see if the transactions in the sample 
     * have already been billed. If this is confirmed, the method returns true. If the time since the last billing sweep does not exceed the interval, if 
     * the interval is not set, if no payment gateways support querying by our transaction id, or if no transaction ids are found remotely, this method returns false.
     * 
     * @return boolean true if a restore from a backup is suspected, false otherwise
     */
    public static function backupRestoreDetected()
    {
        global $wpdb;

        $lastBillingSweep = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_LAST_BILLING_SWEEP);
        if (empty($lastBillingSweep) || !is_numeric($lastBillingSweep)) {
            return false;
        }

        $currentTime = MM_Utils::getCurrentTime('timestamp');
        $interval = "-" . self::$backupDetectionInterval;
        $threshold = strtotime($interval, $currentTime);
        if ($lastBillingSweep > $threshold) {
            return false;
        }

        //the last billing sweep happened outside of the threshold, so a restore from an outdated backup is suspected
        //try to confirm through the payment gateways
        $sampleSize = intval(self::$backupDetectionSampleSize);
        $spTable = MM_TABLE_SCHEDULED_PAYMENTS;
        $seTable = MM_TABLE_SCHEDULED_EVENTS;
        $pending = implode(",", [MM_ScheduledPaymentEvent::$EVENT_PENDING, MM_ScheduledPaymentEvent::$LEGACY_EVENT]);
        $sql = "SELECT sp.* FROM {$spTable} sp INNER JOIN {$seTable} se ON (se.id = sp.id) " .
            "WHERE sp.billing_status IN ({$pending}) AND (se.processed_date IS NULL) AND (se.status = 0) " .
            "ORDER BY se.scheduled_date ASC LIMIT {$sampleSize}";
        $samples = $wpdb->get_results($sql);
        foreach ($samples as $aSample) {
            $ps = MM_PaymentServiceFactory::getPaymentServiceById($aSample->payment_service_id); //this is ok because the factory caches the services
            if (!$ps->supportsFeature(MM_PaymentServiceFeatures::TRANSACTION_METADATA_SEARCH)) {
                continue;
            }

            $metadata = ["scheduledPaymentId" => $aSample->id];
            $matches = $ps->searchTransactionsByMetadata($metadata, $aSample->user_id);
            if (is_array($matches) && (count($matches) > 0)) {
                MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "Scheduled Payment ID {$aSample->id} exists remotely, the site has possibly been restored from an old backup");
                return true;
            }
        }

        return false;
    }


    public static function configureResque($configObject)
    {
        if (is_object($configObject) && isset($configObject->cleanup) && isset($configObject->cleanup->num_retries)) {
            $configObject->cleanup->num_retries = 0; //retries are handled by the dunning system
        }
        return $configObject;
    }


    public static function disableBilling()
    {
        MM_OptionUtils::setOption(MM_OptionUtils::$OPTION_KEY_SCHEDULED_BILLING_ENABLED, "0");
        self::sendBillingDisabledEmail();
    }


    public static function enableBilling()
    {
        MM_OptionUtils::setOption(MM_OptionUtils::$OPTION_KEY_SCHEDULED_BILLING_ENABLED, "1");
    }


    public static function isBillingEnabled()
    {
        $billingEnabled = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_SCHEDULED_BILLING_ENABLED);
        return ($billingEnabled !== "0"); //if the installer does not run immediately, this value will be false, which should be interpreted as enabled
    }


    /**
     * Gets the messaging for the on-site rebilling paused notice.
     * @return array An array with elements "title" and "message", containing the title and message of the notice.
     */
    public static function getBillingPausedMessage()
    {
        $title = sprintf(
            '<i class="fa fa-file-text"></i> <strong>%s</strong>',
            __('MemberMouse On-site Rebilling Has Been Paused', 'membermouse'),
        );

        $message = sprintf(
            '<div style="margin-left:20px;"><p><i class="fa fa-caret-right"></i> %s ',
            __("MemberMouse automatically pauses rebilling when a restore from backup is detected, or the Site URL changes.", "membermouse"),
        );

        $message .= sprintf(
            __('%1$sPlease %2$scarefully review%3$s %4$sOn-site Rebilling has been Paused%5$s.%6$s For assistance, %7$splease contact support%8$s.%9$s', 'membermouse'),
            '<em>',
            '<strong>',
            '</strong>',
            '<a href="https://membermouse.com/docs/onsite-rebilling-has-been-paused">',
            '</a>',
            '</em>',
            '<a href="https://membermouse.com/resource/submit-a-new-support-ticket">',
            '</a>',
            '</p></div>',
        );

        return ["title" => $title, "message" => $message];
    }

    private static function sendBillingDisabledEmail()
    {
        if (function_exists("wp_mail")) {
            $siteUrl = site_url();
            $employee = MM_Employee::getDefault();

            $email = new MM_Email();
            $email->setContext(new MM_Context());
            /* translators: %s: site url */
            $email->setSubject(sprintf(__('ALERT - MemberMouse Local Rebilling Disabled For %s', 'membermouse'), $siteUrl));
            $email->setFromName(__("MemberMouse Notification", "membermouse"));
            $email->setFromAddress("do-not-reply@membermouse.com");
            $email->setToName($employee->getFirstName());
            $email->setToAddress($employee->getEmail());

            $upcomingPaymentsUrl = MM_ModuleUtils::getUrl(MM_MODULE_UPCOMING_PAYMENTS);
            $upcomingPaymentsLink = "<a href=\"{$upcomingPaymentsUrl}\">" . __("Upcoming Payments", "membermouse") . "</a>";

            $settingsUrl = MM_ModuleUtils::getUrl(MM_MODULE_GENERAL_SETTINGS, MM_MODULE_OTHER_SETTINGS);
            $settingsLink = "<a href=\"{$settingsUrl}\">" . __("General Settings", "membermouse") . "</a>";

            $clonedToStagingUrl = "https://membermouse.com/docs/production-databases-copied-to-staging-site";
            $clonedToStagingLink = "<a href=\"{$clonedToStagingUrl}\">" . __("Production Databases Copied to Staging Sites", "membermouse") . "</a>";

            /* translators: %1$s: site url, %2$s: settings link, %3$s: upcoming payments link, %4$s: article link */
            $emailText = __(
                'This is an automated notice to inform you that local rebilling for your MemberMouse ' .
                    'installation at %1$s has been disabled. This happens automatically if the plugin detects ' .
                    'that you may have restored a backup, or cloned your live site into a staging environment. This email ' .
                    'will also be sent if you manually disable the Local Billing Scheduler in %2$s.\n\n' .
                    'If you have restored a backup of your site, subscriptions that billed recently could be scheduled to bill again. ' .
                    'To prevent double billing, please review recent transactions in your payment gateway, and use the ' .
                    '%3$s page to skip any rebills that have already run. Once you are satisfied that the billing ' .
                    'schedule displayed in Upcoming Payments is correct, you should re-enable billing in %2$s.\n\n' .
                    'If you have received this warning from your staging environment after cloning from your production site, ' .
                    'we recommend that you leave the Local Billing Scheduler inactive. Please see our article on %4$s ' .
                    'for recommendations on how to modify your staging database to prevent billing in the staging environement.\n\n' .
                    'If you are unsure what to do, please contact MemberMouse Support at support@membermouse.com for assistance.',
                'membermouse'
            );

            $body = '<html>' . sprintf($emailText, $siteUrl, $settingsLink, $upcomingPaymentsLink, $clonedToStagingLink) . '</html>';

            $email->setBody($body);
            $email->send();
        } else {
            MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "Failed to send billing disabled email notification because wp_mail() is not available");
        }
    }


    public static function performLocalBillingTasks()
    {
        global $wpdb;

        if (self::isBillingEnabled()) {
            if (self::backupRestoreDetected()) {
                self::disableBilling();
                return;
            }

            $currentTimestamp = MM_Utils::getCurrentTime('timestamp');
            MM_OptionUtils::setOption(MM_OptionUtils::$OPTION_KEY_LAST_BILLING_SWEEP, $currentTimestamp);

            $dunningStates = [
                MM_ScheduledPaymentEvent::$EVENT_PENDING,
                MM_ScheduledPaymentEvent::$FIRST_REBILL_FAILED,
                MM_ScheduledPaymentEvent::$SECOND_REBILL_FAILED,
                MM_ScheduledPaymentEvent::$THIRD_REBILL_FAILED
            ];

            foreach ($dunningStates as $dunningState) {
                $scheduledPaymentIds = MM_ScheduledPaymentEvent::assignScheduledPaymentsToBatchAndRetrieve($dunningState);
                if (!empty($scheduledPaymentIds)) {
                    foreach ($scheduledPaymentIds as $scheduledPaymentId) {
                        $job = new ScheduledPaymentJob();
                        $job->scheduled_event_id = $scheduledPaymentId;
                        $jobId = $job->enqueue();
                        if (is_numeric($jobId)) {
                            //link the newly created job with the scheduled payment
                            $wpdb->insert(
                                MM_TABLE_SCHEDULED_PAYMENT_JOB_LINKS,
                                ["scheduled_payment_id" => $scheduledPaymentId, "job_id" => $jobId]
                            );
                        }
                    }
                }
            }
        } else {
            MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "Local billing scheduler did not run because billing is paused");
        }
    }


    /**
     * performDailyMaintenance performs daily maintenance tasks to ensure the integrity of the billing system tables
     * Subscriptions found to be in an invalid state are automatically corrected
     */
    public static function performDailyMaintenance()
    {
        static::repairInvalidBatchIds();
        static::repairCrashedSubscriptions();
    }


    /**
     * repairInvalidBatchIds replaces batch ids in the scheduled events table that are set to empty strings will null values
     * 
     */
    public static function repairInvalidBatchIds()
    {
        global $wpdb;

        $wpdb->update(MM_TABLE_SCHEDULED_EVENTS, ["batch_id" => null], ["batch_id" => ""]);
    }


    /**
     * repairCrashedSubscriptions repairs scheduled rebills that crashed during processing and left the db in an inconsistent state. This can happen when user code hooked to a 
     * payment event has a fatal error or exits the interpreter. The repair logic is as follows:
     * 
     * * Detect when rebills exist that started processing at least 24 hours prior to the last time the billing sweep ran, that are still marked as in-progress
     * * For each of those rebills, remove any existing Resque jobs that exist (these are probably 'stuck')
     * * Check to see if a successful payment was logged in the period starting 24 hours before, and ending 24 hours after the scheduled date
     * * * If a payment exists, reset the status and reschedule this rebill to the next date in the billing sequence
     * * * If a payment does not exist, reset the status and batch id, and queue it up to bill again immediately
     */
    public static function repairCrashedSubscriptions()
    {
        global $wpdb;

        $threshold = "24 hours";
        $lastBillingSweep = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_LAST_BILLING_SWEEP);

        if ($lastBillingSweep === false) {
            return; //billing has never run
        }

        //* All scheduled payments with status 1 that have scheduled date of [24 hours] prior to the last billing sweep are probably crashed
        $spTable = MM_TABLE_SCHEDULED_PAYMENTS;
        $seTable = MM_TABLE_SCHEDULED_EVENTS;
        $spjlTable = MM_TABLE_SCHEDULED_PAYMENT_JOB_LINKS;
        $jobsTable = $wpdb->prefix . membermouse\resque\SUB_PREFIX . 'jobs'; //Resque jobs table
        $transactionsTable = MM_TABLE_TRANSACTION_LOG;

        $upv = new MM_UpcomingPaymentsView();
        $thresholdDate = date("Y-m-d H:i:s", strtotime("-{$threshold}", $lastBillingSweep));
        $billingFloorDate = MM_ScheduledPaymentEvent::getBillingFloorDate();
        $sql = "SELECT sp.*,se.scheduled_date FROM {$spTable} sp INNER JOIN {$seTable} se ON (sp.id = se.id) " .
            "WHERE (se.status = 2) AND (se.scheduled_date >= '{$billingFloorDate}') AND (se.scheduled_date <= '{$thresholdDate}') ";
        $results = $wpdb->get_results($sql);
        foreach ($results as $row) {
            //remove any resque jobs that might be stuck
            $scheduledPaymentId = $row->id;
            $jobSQL = "DELETE j.* FROM {$spjlTable} spj INNER JOIN {$jobsTable} j ON (j.id = spj.job_id) " .
                "WHERE (spj.scheduled_payment_id = %d) AND (j.args LIKE '%\"scheduled_event_id\":\"%d\"%')";
            $wpdb->query($wpdb->prepare($jobSQL, $scheduledPaymentId, $scheduledPaymentId));

            //check if there was a successful payment, and if so, skip this rebill
            $baseTimestamp = strtotime($row->scheduled_date);
            $rangeStart = date("Y-m-d H:i:s", strtotime("-24 hours", $baseTimestamp));
            $rangeEnd = date("Y-m-d H:i:s", strtotime("+24 hours", $baseTimestamp));
            $tsCheckSQL = "SELECT count(*) FROM {$transactionsTable} WHERE (order_item_id = %d) AND (transaction_type = %d) AND (transaction_date BETWEEN %s AND %s)";
            $numPayments = $wpdb->get_var($wpdb->prepare($tsCheckSQL, $row->order_item_id, MM_TransactionLog::$TRANSACTION_TYPE_RECURRING_PAYMENT, $rangeStart, $rangeEnd));
            if (intval($numPayments) > 0) {
                //skip payment
                $upv->skipPayment(["scheduledPaymentId" => $scheduledPaymentId]);
            }

            //reset scheduled payment
            $scheduledPaymentEvent = new MM_ScheduledPaymentEvent($scheduledPaymentId);
            $scheduledPaymentEvent->clearStatusAndBatchId(true, true);
        }
    }
}
