<?php

/**
 *
 * MemberMouse(TM) (http://www.membermouse.com)
 * (c) MemberMouse, LLC. All rights reserved.
 * 
 * The Extentions Factory retrieves a requested extension based on a token, or all available extensions
 */

class MM_ExtensionsFactory
{
    //used to hold instances of extensions that can be reused
    private static $instances = array();

    /**
     * Retrieve all of the active extensions
     * 
     * @return array An array containing all the available extension objects, keyed by the token
     */
    public static function getActiveExtensions($autoloadDependencies = true)
    {
        $extensionsArray = self::getAvailableExtensions($autoloadDependencies);

        foreach ($extensionsArray as $token => $extension) {
            if (!$extension->isActive()) {
                unset($extensionsArray[$token]);
            }
        }

        return $extensionsArray;
    }

    /**
     * Returns a specific extension. Enforces the singleton nature of extensions and returns an already instantiated instance if the 
     * requested extension has already been created
     * 
     * @param String $extensionToken
     * @param boolean $autoloadDependencies(optional) Whether to load libraries required by the extension. For simple configuration tasks, this can be false
     * @param array|string (optional) $configData Configuration data that can be passed to the extension when it is created. This value is an empty string when it is unused
     * 
     * @return MM_Extension An instance of the extension matching the supplied token, null if one can't be found
     */
    public static function getExtension($extensionToken, $autoloadDependencies = true, $configData = "")
    {
        if (isset(MM_ExtensionsFactory::$instances[$extensionToken]) && (MM_ExtensionsFactory::$instances[$extensionToken] instanceof MM_Extension)) {
            $ext = MM_ExtensionsFactory::$instances[$extensionToken];
            if ($autoloadDependencies) {
                $ext->loadDependencies();
            }
            return $ext;
        }

        switch ($extensionToken) {
            case MM_Extension::$USERVOICE_TOKEN:
                MM_ExtensionsFactory::$instances[$extensionToken] = new MM_UserVoiceExtension($autoloadDependencies, $configData);
                break;
            case MM_Extension::$LINKEDIN_LOGIN_TOKEN:
                MM_ExtensionsFactory::$instances[$extensionToken] = new MM_LinkedInLoginExtension($autoloadDependencies, $configData);
                break;
            case MM_Extension::$FACEBOOK_LOGIN_TOKEN:
                MM_ExtensionsFactory::$instances[$extensionToken] = new MM_FacebookLoginExtension($autoloadDependencies, $configData);
                break;
            case MM_Extension::$GOOGLE_LOGIN_TOKEN:
                MM_ExtensionsFactory::$instances[$extensionToken] = new MM_GoogleLoginExtension($autoloadDependencies, $configData);
                break;
            case MM_Extension::$TWITTER_LOGIN_TOKEN:
                MM_ExtensionsFactory::$instances[$extensionToken] = new MM_TwitterLoginExtension($autoloadDependencies, $configData);
                break;
            case MM_Extension::$BBPRESS_TOKEN:
                MM_ExtensionsFactory::$instances[$extensionToken] = new MM_BBPressExtension($autoloadDependencies, $configData);
                break;
            case MM_Extension::$MM_COURSES_TOKEN:
                MM_ExtensionsFactory::$instances[$extensionToken] = new MM_MMCoursesExtension($autoloadDependencies, $configData);
                break;
            case MM_Extension::$PDF_RECEIPTS_TOKEN:
                MM_ExtensionsFactory::$instances[$extensionToken] = new MM_PDFReceiptsExtension($autoloadDependencies, $configData);
                break;
            case MM_Extension::$GOOGLE_ECOMMERCE_TOKEN:
                // TODO MM_ExtensionsFactory::$instances[$extensionToken] = new MM_GoogleECommerceExtension();
                break;
            default:
                return null;
                break;
        }

        return MM_ExtensionsFactory::$instances[$extensionToken];
    }


    public static function getAvailableExtensions($autoloadDependencies = true)
    {
        global $wpdb;

        $tokens = array(
            MM_Extension::$BBPRESS_TOKEN,
            MM_Extension::$USERVOICE_TOKEN,
            MM_Extension::$MM_COURSES_TOKEN,
            MM_Extension::$PDF_RECEIPTS_TOKEN
        );

        $extensions = array();
        foreach ($tokens as $token) {
            $ext = self::getExtension($token, $autoloadDependencies);
            if (($ext instanceof MM_Extension) && ($ext->hasPermission())) {
                $extensions[$token] = $ext;
            }
        }

        //handle social login extensions separately for efficiency reasons
        if (MM_MemberMouseService::hasPermission(MM_MemberMouseService::$EXTENSION_SOCIAL_LOGIN) == MM_MemberMouseService::$ACTIVE) {
            $tokens = array(
                MM_Extension::$LINKEDIN_LOGIN_TOKEN,
                MM_Extension::$FACEBOOK_LOGIN_TOKEN,
                MM_Extension::$GOOGLE_LOGIN_TOKEN,
                MM_Extension::$TWITTER_LOGIN_TOKEN
            );
            $tokenString = "'" . implode("','", $tokens) . "'";
            $configData = array();
            foreach ($tokens as $token) {
                $configData[$token] = array();
            }
            $socialLoginProviderSQL = "SELECT * FROM " . MM_TABLE_SOCIAL_LOGIN_PROVIDERS . " WHERE token IN ({$tokenString})";
            $results = $wpdb->get_results($socialLoginProviderSQL);
            foreach ($results as $row) {
                $configData[$row->token] = $row;
            }

            foreach ($configData as $token => $config) {
                $ext = self::getExtension($token, $autoloadDependencies, $config);
                if (($ext instanceof MM_Extension) && ($ext->hasPermission())) {
                    $extensions[$token] = $ext;
                }
            }
        }

        return $extensions;
    }
}
