<?php

/**
 * 
 *
 * MemberMouse(TM) (http://www.membermouse.com)
 * (c) MemberMouse, LLC. All rights reserved.
 *
 *
 * MM_FreeMemberWebformRequest is responsible for processing free member webform data.
 */

class MM_FreeMemberWebformRequest extends MM_OrderRequest
{
    protected $newUser = null;
    protected $membership = null;


    /**
     * This method validates the webform data by ensuring that required fields are present and 
     * that the membership level passed is valid. It then creates an instance of MM_User and 
     * popultes it with the webform data.
     *
     * @see MM_OrderRequest::processRequest()
     */
    public function processRequest()
    {
        if (!isset($this->requestData["email"]) || trim($this->requestData["email"]) == "") {
            $this->state->type = MM_Response::$ERROR;
            $this->state->message = __("Missing required field: email address is required.", "membermouse");
            return;
        } else if (!MM_Utils::validateEmail($this->requestData["email"])) {
            $this->state->type = MM_Response::$ERROR;
            /* translators: %s: an invalid email address */
            $this->state->message = sprintf(__('"%s" is not a valid email address.', 'membermouse'), $this->requestData["email"]);
            return;
        }

        if (!isset($this->requestData["member_type"]) && !isset($this->requestData["membership_level"])) {
            $this->state->type = MM_Response::$ERROR;
            $this->state->message = __('Missing required field: membership level is required.', 'membermouse');
            return;
        }

        $membershipLevelId = (isset($this->requestData["membership_level"])) ? $this->requestData["membership_level"] : $this->requestData["member_type"];

        $this->requestedMembership = new MM_MembershipLevel($membershipLevelId);

        if ($this->requestedMembership->isValid() && $this->requestedMembership->isFree()) {
            $this->newUser = new MM_User();
            $this->newUser->setEmail($this->requestData["email"]);

            if (isset($this->requestData["username"]) && $this->requestData["username"] != "") {
                $this->newUser->setUsername($this->requestData["username"]);
            }

            if (isset($this->requestData["password"]) && $this->requestData["password"] != "") {
                $this->newUser->setPassword($this->requestData["password"]);
            } else {
                $this->newUser->setPassword(MM_Utils::createRandomString(8));
            }

            if (isset($this->requestData["first_name"]) && $this->requestData["first_name"] != "") {
                $this->newUser->setFirstName($this->requestData["first_name"]);
            }

            if (isset($this->requestData["last_name"]) && $this->requestData["last_name"] != "") {
                $this->newUser->setLastName($this->requestData["last_name"]);
            }

            if (isset($this->requestData["phone"]) && $this->requestData["phone"] != "") {
                $this->newUser->setPhone($this->requestData["phone"]);
            }

            $this->newUser->validate();
        } else {
            $this->state->type = MM_Response::$ERROR;
            $this->state->message = __("Invalid membership level ID ", "membermouse") . $membershipLevelId;
            return;
        }
    }


    /**
     * This method attempts to create the new user account and assign the requested membership level them.
     * 
     * @see MM_OrderRequest::submitRequest()
     */
    public function submitRequest()
    {
        $result = $this->state;

        if (MM_Response::isSuccess($result)) {
            $result = MM_AccessControlEngine::assignMembershipToUser($this->newUser, $this->requestedMembership);

            if (MM_Response::isSuccess($result)) {
                $this->newUser = MM_User::findByEmail($this->requestData["email"]);

                if ($this->newUser->isValid()) {
                    // add custom field data
                    foreach ($this->requestData as $k => $v) {
                        if (preg_match("/(custom_)/", $k)) {
                            $fieldId = preg_replace("/[^0-9]+/", "", $k);
                            $result = $this->newUser->setCustomData($fieldId, $v);

                            if (MM_Response::isError($result)) {
                                break;
                            }
                        }
                    }
                }
            }
        }

        $this->handleResult($result);
    }


    /**
     * This method handles the result of the attempt to create the free member account. If the result is successful,
     * the customer will be redirected to the appropriate confirmation page. If there was an error processing the webform
     * or creating the account, the customer will be redirected to the default error page with an error message passed.
     *
     * @param MM_Response $result an instance of a response object that contains details about the result from the attempt to create the free member account
     *
     * @see MM_OrderRequest::handleResult()
     */
    protected function handleResult($result)
    {
        if ($result->type == MM_Response::$SUCCESS) {
            $params = array();
            $params["isFree"] = 1;
            $params["refType"] = MM_TYPE_MEMBERSHIP_LEVEL;
            $params["refId"] = $this->requestedMembership->getId();

            $url = MM_CorePageEngine::getUrl(MM_CorePageType::$FREE_CONFIRMATION, $params);
        } else {
            $url = MM_CorePageEngine::getUrl(MM_CorePageType::$ERROR);
            $url = MM_Utils::appendUrlParam($url, MM_Session::$PARAM_MESSAGE_KEY, $result->message, true);
        }

        wp_redirect($url);
        exit;
    }
}
