<?php

/**
 *
 * MemberMouse(TM) (http://www.membermouse.com)
 * (c) MemberMouse, LLC. All rights reserved.
 * 
 * MM_ScheduledEvent represents an event that was created by the MemberMouse scheduler. It is important to differentiate this 
 * local entity from the entity on central with the same name. Despite having identical name, and representing the same logical concept, 
 * they are very different objects because the interfaces were written from the perspective of the requestor in the former case, and the perspective 
 * of the dispatcher in the latter case.
 * 
 */
abstract class MM_ScheduledEvent extends MM_Entity
{

    protected $eventType     = "";
    protected $eventData     = "";
    protected $scheduledDate = "";
    protected $processedDate = null;
    protected $status = 0;
    protected $batchId = null;

    //constants representing event types
    public static $PAYMENT_SERVICE_EVENT = 1;
    public static $REMINDER_EVENT = 2;

    //constants representing commands
    public static $SCHEDULER_COMMAND_UPDATE = "UPDATE";
    public static $SCHEDULER_COMMAND_DELETE = "DELETE";

    public static $QUEUE_COMMAND_UPDATE = "0"; //numeric constants used for queueing, because that table could potentially become pretty big
    public static $QUEUE_COMMAND_DELETE = "1";

    //constants representing event statuses
    public static $EVENT_PENDING = 0;
    public static $EVENT_PROCESSED = 1;
    public static $EVENT_PROCESSING = 2;

    /**
     * Retrieve stored schedule information. 
     * 
     * @see MM_Entity::getData()
     */
    public function getData()
    {
        global $wpdb;

        if (intval($this->id) > 0) {
            $sql = "SELECT * from " . MM_TABLE_SCHEDULED_EVENTS . " where id={$this->id}";
            $result = $wpdb->get_row($sql);
            if ($result) {
                $this->setData($result);
            } else {
                parent::invalidate();
            }
        } else {
            parent::invalidate();
        }
    }


    public function getDataByOrderItemAndProcessingStatus($orderItemId)
    {
        global $wpdb;

        $sql = "SELECT sp.id FROM " . MM_TABLE_SCHEDULED_PAYMENTS . " sp LEFT JOIN " . MM_TABLE_SCHEDULED_EVENTS . " se on (sp.id = se.id) " .
            "WHERE (order_item_id=%s) AND (status = %s) ";

        $sql .= "ORDER BY id DESC LIMIT 1";
        $sql = $wpdb->prepare($sql, $orderItemId, MM_ScheduledEvent::$EVENT_PROCESSING);

        $eventId = $wpdb->get_var($sql);
        if ($eventId != null) {
            $this->id = $eventId;
            $sql = "SELECT * from " . MM_TABLE_SCHEDULED_EVENTS . " where id={$this->id}";
            $result = $wpdb->get_row($sql);
            if ($result) {
                $this->setData($result);
            } else {
                parent::invalidate();
            }
        } else {
            parent::invalidate();
        }
    }

    /**
     * Populate this entity based on the schedule information in object $data
     *
     * @param array $data An object of type stdclass attributes represent the columns of a row in the underlying database table
     * @see MM_Entity::setData()
     */
    public function setData($data)
    {
        try {
            $this->id = $data->id;
            $this->eventType = $data->event_type;
            $this->setScheduledDate($data->scheduled_date);
            $this->setProcessedDate($data->processed_date);
            $this->status = $data->status;
            $this->batchId = empty($data->batch_id) ? null : $data->batch_id;

            if (!empty($data->event_data)) {
                $unserialized_settings = unserialize($data->event_data);
                if (is_array($unserialized_settings)) {
                    foreach ($unserialized_settings as $k => $v) {
                        if (isset($this->$k)) {
                            $this->$k = $v; //don't create attributes, only restore existing ones
                        }
                    }
                }
            }

            parent::validate();
        } catch (Exception $ex) {
            parent::invalidate();
        }
    }


    /**
     * Commit the contents of this entity to the database
     * @see MM_Entity::commitData()
     *
     * @return MM_Response object representing success or failure
     * 
     */
    public function commitData()
    {
        global $wpdb;

        if (empty($this->scheduledDate)) {
            return new MM_Response("Error creating scheduled event: No date was set", MM_Response::$ERROR);
        }

        $isUpdate = (intval($this->id) > 0) ? true : false;

        $scheduleData = array(
            "event_data"          => $this->prepareSettings(),
            "event_type"          => $this->eventType,
            "scheduled_date"    => $this->getScheduledDate(),
            "status"            => $this->getStatus(),
            "batch_id"          => $this->getBatchId()
        );

        if (!empty($this->processedDate)) {
            $scheduleData["processed_date"] = $this->getProcessedDate();
        }

        if (!$isUpdate) {
            //this is a new record
            $result = $wpdb->insert(MM_TABLE_SCHEDULED_EVENTS, $scheduleData);
            if ($result === false) {
                return new MM_Response("ERROR: unable to create scheduled event: error = " . $wpdb->last_error, MM_Response::$ERROR);
            }
            $this->id = $wpdb->insert_id;
        } else {
            $updateWhere = array("id" => $this->id);
            $result = $wpdb->update(MM_TABLE_SCHEDULED_EVENTS, $scheduleData, $updateWhere);
            if ($result === false) {
                return new MM_Response("ERROR: unable to update scheduled event: error = " . $wpdb->last_error, MM_Response::$ERROR);
            }
        }
        parent::validate();
        return new MM_Response();
    }


    protected function getEventData()
    {
        return $this->eventData;
    }


    protected function setEventData($eventData)
    {
        $this->eventData = $eventData;
    }


    public function getEventType()
    {
        return $this->eventType;
    }


    protected function setEventType($eventType)
    {
        $this->eventType = $eventType;
    }


    public function getScheduledDate()
    {
        return date("Y-m-d H:i", $this->scheduledDate);
    }


    public function setScheduledDate($scheduledDate)
    {
        $this->scheduledDate = is_int($scheduledDate) ? $scheduledDate : strtotime($scheduledDate);
    }


    /**
     * getProcessedDate returns the date that this schedule was last processed by the billing system. When working with a scheduled payment, this property 
     * can be used with billingStatus to determine when the scheduled payment entered its current dunning status
     * 
     * @return NULL|string If there is no processed date set, null is returned, otherwise a MySQL formatted string representation of the processed date (in UTC) is returned
     */
    public function getProcessedDate()
    {
        return empty($this->processedDate) ? null : gmdate("Y-m-d H:i", $this->processedDate);
    }


    /**
     * setProcessedDate sets the date that this schedule was last processed by the billing system. Internally the date is stored as a UNIX timestamp
     * 
     * @param int|string $processedDate Either a UNIX timestamp or a string date representation
     */
    public function setProcessedDate($processedDate)
    {
        if (empty($processedDate)) {
            $this->processedDate = null;
        } else {
            $this->processedDate = is_int($processedDate) ? $processedDate : strtotime($processedDate);
        }
    }


    public function getStatus()
    {
        return $this->status;
    }


    public function setStatus($status)
    {
        $this->status = $status;
    }


    public function delete()
    {
        global $wpdb;

        if (intval($this->id) > 0) {
            $wpdb->query("DELETE FROM " . MM_TABLE_SCHEDULED_EVENTS . " WHERE id='{$this->getId()}'");
            parent::invalidate();
            return new MM_Response();
        } else {
            return new MM_Response("Unable to delete this scheduled event because it was never committed", MM_Response::$ERROR);
        }
    }


    /**
     * Each subclass of the ScheduledEvent entity can have specific settings. These settings are stored as attributes on the object. Prior to commit,
     * prepareSettings creates an array using these attributes and then serializes that array and stores it in $this->eventData, so that the settings are persisted
     * when commitData is called. Atrributes with names that start with "transient_" are ignored
     * 
     * @param $additionalExclusions array of additional class members to exclude from serialization
     * 
     * @return string representing a serialized array of all the visible class members, minus the exclusions, suitable for storage in the eventData field
     */
    protected function prepareSettings($additionalExclusions = array())
    {
        $exclusions = array('id', 'eventData', 'eventType', 'scheduledDate', 'processedDate', 'status', 'notifyServices', 'batchId') + $additionalExclusions;
        $settings = array();
        foreach ($this as $key => $value) {
            if (!in_array($key, $exclusions) && (strpos($key, "transient_") !== 0)) {
                $settings[$key] = $value;
            }
        }
        return serialize($settings);
    }

    public function getBatchId()
    {
        return $this->batchId;
    }


    public function setBatchId($batchId)
    {
        $this->batchId = $batchId;
    }


    public function clearStatusAndBatchId($clearStatus = true, $clearBatchId = false)
    {
        global $wpdb;

        if ($clearStatus || $clearBatchId) {
            $fields = [];
            if ($clearStatus) {
                $this->status = 0;
                $fields["status"] = 0;
            }
            if ($clearBatchId) {
                $this->batchId = null;
                $fields["batch_id"] = null;
            }

            $res = $wpdb->update(MM_TABLE_SCHEDULED_EVENTS, $fields, ["id" => $this->getId()]);
            if ($res === false) {
                return new MM_Response("Error clearing batch id from scheduled event ID: {$this->getId()}", MM_Response::$ERROR);
            }
        }
        return new MM_Response();
    }


    /**
     * Updates the event status immediately in the database. Used to indicate when the event is being dispatched to avoid concurrency issues
     * 
     * @param int $newStatus An integer representing event status
     * @return MM_Response indicating success or failure
     */
    public function updateStatus($newStatus)
    {
        global $wpdb;

        if (!in_array($newStatus, [self::$EVENT_PENDING, self::$EVENT_PROCESSED, self::$EVENT_PROCESSING])) {
            return new MM_Response("Invalid status passed to MM_ScheduledEvent::updateStatus", MM_Response::$ERROR);
        }

        $res = $wpdb->update(MM_TABLE_SCHEDULED_EVENTS, ["status" => $newStatus], ["id" => $this->getId()]);
        if ($res === false) {
            return new MM_Response("Query error in MM_ScheduledEvent::updateStatus, received value: {$newStatus}", MM_Response::$ERROR);
        }
        return new MM_Response();
    }
}
