<?php

/**
 *
 * MemberMouse(TM) (http://www.membermouse.com)
 * (c) MemberMouse, LLC. All rights reserved.
 *
 * MM_ScheduledPaymentEvent represents a recurring payment that was scheduled by the MemberMouse scheduling framework. It specifically represents 
 * a payment event
 *
 */
class MM_ScheduledPaymentEvent extends MM_ScheduledEvent
{
    protected $paymentNumber    = "";
    private   $orderItemId      = "";
    private   $userId           = "";
    private   $paymentServiceId = "";
    private   $eventTimestamp   = null;
    private   $billingStatus    = ""; //represents dunning status


    //pseudo-constants representing billing statuses 
    public static $PENDING_REBILL = 0;
    public static $REBILLED_SUCCESSFULLY = 1;
    public static $FIRST_REBILL_FAILED = 2;
    public static $SECOND_REBILL_FAILED = 3;
    public static $THIRD_REBILL_FAILED = 4;
    public static $PERMANENTLY_FAILED = 5;
    public static $LEGACY_EVENT = 9; //despite the name this is a billing status

    //constants representing the time periods in the dunning cycle before a failed rebill is retried. 
    //the default is 4 attempts: the first attempt, followed by attempts at 2 days, 5 days, and finally 7 days
    //the interval is the amount of time since the last attempt
    public static $FIRST_FAILURE_INTERVAL = 2;
    public static $SECOND_FAILURE_INTERVAL = 3;
    public static $THIRD_FAILURE_INTERVAL = 2;


    private static $MIN_SECONDS_ELAPSED_PROCESSING = 2;
    public const BILLING_FLOOR_DATE = "2024-01-01 00:00"; //don't automatically process scheduled payments that were scheduled prior to this date. Date is in mysql format

    public function __construct($id = "", $getData = true)
    {
        $this->setBillingStatus(MM_ScheduledPaymentEvent::$PENDING_REBILL); //events are pending by default
        $this->eventType = MM_ScheduledEvent::$PAYMENT_SERVICE_EVENT;
        parent::__construct($id, $getData);
    }

    public function saveStatus($status)
    {
        $this->setStatus($status);
        if ($status == self::$EVENT_PENDING) {
            $this->setBatchId(null);
        }
        $this->commitData();
    }

    public function isProcessing()
    {
        if ($this->getStatus() == MM_ScheduledEvent::$EVENT_PROCESSING) {
            if (!is_null($this->eventTimestamp) && !empty($this->eventTimestamp)) {
                $diff = strtotime("now") - strtotime($this->eventTimestamp);
                if ($diff < self::$MIN_SECONDS_ELAPSED_PROCESSING) {
                    return true;
                }
            }
            $this->saveStatus(MM_ScheduledEvent::$EVENT_PENDING);
        }
        return false;
    }

    public function getData()
    {
        global $wpdb;

        parent::getData();

        if (intval($this->id) > 0) {
            $sql = "SELECT * from " . MM_TABLE_SCHEDULED_PAYMENTS . " where id={$this->id}";
            $result = $wpdb->get_row($sql);
            if ($result) {
                $this->setAdditionalData($result);
            } else {
                parent::invalidate();
            }
        } else {
            parent::invalidate();
        }
    }


    public function setAdditionalData($data)
    {
        $this->orderItemId = $data->order_item_id;
        $this->userId = $data->user_id;
        $this->paymentServiceId = $data->payment_service_id;
        $this->eventTimestamp = $data->event_timestamp;
        $this->billingStatus = $data->billing_status;
    }


    public function commitData()
    {
        global $wpdb;

        //figure out if this is a new record or an update
        $newRecord = empty($this->id);

        //call the parent commit to generate the event
        $response = parent::commitData();
        if (MM_Response::isError($response)) {
            return $response;
        }

        //primary table updated, and schedule created, now create/update details table
        $scheduleDetailData = array(
            "user_id"              => $this->userId,
            "order_item_id"        => $this->orderItemId,
            "payment_service_id"  => $this->paymentServiceId,
            "event_timestamp"     => current_time('mysql', 1),
            "billing_status"      => $this->billingStatus
        );

        if ($newRecord) {
            //this is a new record
            $scheduleDetailData = array("id" => $this->id) + $scheduleDetailData;
            $result = $wpdb->insert(MM_TABLE_SCHEDULED_PAYMENTS, $scheduleDetailData);
            if ($result === false) {
                return new MM_Response("ERROR: unable to create scheduled event: error = " . $wpdb->last_error, MM_Response::$ERROR);
            }
        } else {
            $updateWhere = array("id" => $this->id);
            $result = $wpdb->update(MM_TABLE_SCHEDULED_PAYMENTS, $scheduleDetailData, $updateWhere);
            if ($result === false) {
                return new MM_Response("ERROR: unable to update scheduled event: error = " . $wpdb->last_error, MM_Response::$ERROR);
            }
        }

        parent::validate();
        return new MM_Response();
    }


    public function delete()
    {
        global $wpdb;

        if (intval($this->id) > 0) {
            $wpdb->query("DELETE FROM " . MM_TABLE_SCHEDULED_PAYMENTS . " WHERE id='{$this->id}'");
            $response = parent::delete();
            if (MM_Response::isError($response)) {
                return $response;
            }
            parent::invalidate();
            return new MM_Response();
        } else {
            return new MM_Response("Unable to delete this scheduled payment event because it is invalid", MM_Response::$ERROR);
        }
    }


    protected function prepareSettings($additionalExclusions = array())
    {
        $additionalExclusions[] = "orderItemId";
        $additionalExclusions[] = "userId";
        $additionalExclusions[] = "paymentServiceId";
        $additionalExclusions[] = "billingStatus";
        return parent::prepareSettings($additionalExclusions);
    }


    /**
     * Returns the payment number
     * 
     * @return integer representing the payment number
     */
    public function getPaymentNumber()
    {
        return $this->paymentNumber;
    }


    /**
     * Sets the payment number
     * 
     * @param int $paymentNumber
     */
    public function setPaymentNumber($paymentNumber)
    {
        if (is_numeric($paymentNumber)) {
            $this->paymentNumber = intval($paymentNumber);
        }
    }


    protected function setEventType($eventType)
    {
        //don't allow the event type to be changed
        $this->eventType = MM_ScheduledEvent::$PAYMENT_SERVICE_EVENT;
    }


    public function getOrderItemId()
    {
        return $this->orderItemId;
    }


    public function setOrderItemId($orderItemId)
    {
        $this->orderItemId = $orderItemId;
    }


    public function getUserId()
    {
        return $this->userId;
    }


    public function setUserId($userId)
    {
        $this->userId = $userId;
    }


    public function getPaymentServiceId()
    {
        return $this->paymentServiceId;
    }


    public function setPaymentServiceId($paymentServiceId)
    {
        $this->paymentServiceId = $paymentServiceId;
    }


    public function getBillingStatus()
    {
        return $this->billingStatus;
    }


    public function setBillingStatus($billingStatus)
    {
        $this->billingStatus = $billingStatus;
    }


    public function getEventTimestamp()
    {
        return $this->eventTimestamp;
    }


    /**
     * Finds the next scheduled payment event using an order item id 
     * 
     * @param string $orderItemId The order item id
     * @return MM_ScheduledPaymentEvent for the next scheduled event on that order item id. Returns boolean false if no next event exists
     */
    public static function findNextScheduledEventByOrderItemId($orderItemId, $futureOnly = true)
    {
        global $wpdb;

        $mmOrderItem = new MM_OrderItem($orderItemId);
        $order = new MM_Order($mmOrderItem->getOrderId());
        $paymentToken = $order->getPaymentMethod()->getToken();
        $paymentService = MM_PaymentServiceFactory::getPaymentService($paymentToken);
        if (method_exists($paymentService, "getNextBillingDate")) {
            $nextDate = $paymentService->getNextBillingDate($mmOrderItem);
            if ($nextDate !== false) {
                return $nextDate;
            }
        }

        $currentDatetime = MM_Utils::getCurrentTime();
        $sql = "SELECT sp.id FROM " . MM_TABLE_SCHEDULED_PAYMENTS . " sp LEFT JOIN " . MM_TABLE_SCHEDULED_EVENTS . " se on (sp.id = se.id) " .
            "WHERE (order_item_id=%s) AND (status = %s) ";
        if ($futureOnly) {
            $sql .= "AND (scheduled_date > '{$currentDatetime}') ";
        }
        $sql .= "ORDER BY id ASC LIMIT 1";
        $sql = $wpdb->prepare($sql, $orderItemId, MM_ScheduledEvent::$EVENT_PENDING);
        $eventId = $wpdb->get_var($sql);

        if ($eventId !== false) {
            return new MM_ScheduledPaymentEvent($eventId);
        }
        return false;
    }


    /**
     * Returns an array of ids for scheduled payments ready to bill immediately. If $setBatchId is provided, all of the schedules returned will be marked with that value in the
     * batchId field.
     *
     * @param int $dunningState The state of the Scheduled Payments that should be retrieved
     */
    public static function assignScheduledPaymentsToBatchAndRetrieve($dunningState)
    {
        global $wpdb;

        $recognizedDunningStates = [self::$EVENT_PENDING, self::$FIRST_REBILL_FAILED, self::$SECOND_REBILL_FAILED, self::$THIRD_REBILL_FAILED];
        if (!in_array($dunningState, $recognizedDunningStates)) {
            return false;
        }

        $targetDate = MM_Utils::getCurrentTime();
        if ($dunningState != self::$EVENT_PENDING) {
            $dunningInterval = self::getDunningInterval($dunningState);
            $targetDate = date("Y-m-d H:i:s", strtotime("-{$dunningInterval} days", strtotime($targetDate)));
            $billingStatusWhere = "(sp.billing_status = {$dunningState})";
        } else {
            $billingStatusWhere = $wpdb->prepare("(sp.billing_status IN (%d,%d))", self::$EVENT_PENDING, self::$LEGACY_EVENT);
        }
        $setBatchId = MM_Utils::createRandomString(32);
        $sql = "UPDATE " . MM_TABLE_SCHEDULED_EVENTS . " se INNER JOIN " . MM_TABLE_SCHEDULED_PAYMENTS . " sp ON " .
            "(se.id = sp.id) SET se.batch_id = %s WHERE (IFNULL(se.processed_date,se.scheduled_date) <= %s) AND (se.scheduled_date > %s) AND (se.status = %d) AND " .
            "{$billingStatusWhere} AND (se.batch_id IS NULL) AND (se.event_type = %d)";
        $sql = $wpdb->prepare($sql, $setBatchId, $targetDate, self::BILLING_FLOOR_DATE, MM_ScheduledEvent::$EVENT_PENDING, MM_ScheduledEvent::$PAYMENT_SERVICE_EVENT);
        $result = $wpdb->query($sql);
        if ($result !== false) {
            $sql = $wpdb->prepare("SELECT id FROM " . MM_TABLE_SCHEDULED_EVENTS . " WHERE batch_id = %s", $setBatchId);
            $currentBatchScheduleIds = $wpdb->get_col($sql);
            return $currentBatchScheduleIds;
        } else {
            MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "Error setting batch id for local event dispatch");
            return false;
        }
    }


    /**
     * Determine the dunning interval for each dunning state
     * 
     * @param int $dunningState One of the constants representing dunning state found in MM_ScheduledPaymentEvent
     * @return int The number of days to wait prior to the next billing attempt
     */
    public static function getDunningInterval($dunningState)
    {
        switch (intval($dunningState)) {
            case self::$FIRST_REBILL_FAILED:
                return apply_filters(MM_Event::$DUNNING_INTERVAL, self::$FIRST_FAILURE_INTERVAL, self::$FIRST_REBILL_FAILED);
                break;
            case self::$SECOND_REBILL_FAILED:
                return apply_filters(MM_Event::$DUNNING_INTERVAL, self::$SECOND_FAILURE_INTERVAL, self::$SECOND_REBILL_FAILED);
                break;
            case self::$THIRD_REBILL_FAILED:
                return apply_filters(MM_Event::$DUNNING_INTERVAL, self::$THIRD_FAILURE_INTERVAL, self::$THIRD_REBILL_FAILED);
                break;
            default:
                return 2; //received an unrecognized dunning state, return "2 days" as a reasonable default. Should not happen
                break;
        }

        return 2;
    }


    /**
     * execute attempts to immediately charge a scheduled payment, and return the results of the charge operation
     *
     * @param string|int $scheduledPaymentId The id of the scheduled payment to execute
     *
     * @return MM_PaymentServiceResponse indicating the results of the operation. If an error is encountered, this is returned as a response of type MM_PaymentServiceResponse::$ERROR
     */
    public function execute()
    {
        if (!$this->isValid()) {
            $error = "execute called on invalid ScheduledPaymentEvent";
            MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, $error);
            $response = new MM_PaymentServiceResponse($error, MM_PaymentServiceResponse::$ERROR);
            $response->data['immediately_fail'] = true; //hint for jobs queue to not retry
            return $response;
        }

        try {
            if ($this->isProcessing()) {
                MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "executeScheduledPaymentEvent: Event {$this->id} is currently being processed and will be ignored");
                return new MM_PaymentServiceResponse();
            }

            if ($this->getStatus() == MM_ScheduledEvent::$EVENT_PROCESSED) {
                MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "executeScheduledPaymentEvent: Event {$this->id} has already been processed");
                return new MM_PaymentServiceResponse();
            }

            $this->saveStatus(MM_ScheduledEvent::$EVENT_PROCESSING);

            $paymentService = MM_PaymentServiceFactory::getPaymentServiceById($this->getPaymentServiceId());
            if (is_null($paymentService)) {
                $this->saveStatus(MM_ScheduledEvent::$EVENT_PENDING);
                throw new Exception("executeScheduledPaymentEvent: Improper scheduled event configuration: Payment service with id {$this->getPaymentServiceId()} not found");
            }

            $response = $paymentService->processScheduledPaymentEvent($this);
            if (MM_PaymentServiceResponse::isError($response) || MM_PaymentServiceResponse::isFailed($response)) {
                $this->saveStatus(MM_ScheduledEvent::$EVENT_PENDING);
                MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "executeScheduledPaymentEvent: encountered an error when processing scheduled payment with id {$this->getId()}: {$response->message}");
                return $response;
            } else if (MM_PaymentServiceResponse::isSuccess($response)) {
                $this->saveStatus(MM_ScheduledEvent::$EVENT_PROCESSED);
                return $response;
            }
            //execution should not reach here
            throw new Exception("An unknown error was encountered when processing schedule with id {$this->id}");
        } catch (Exception $e) {
            $error = $e->getMessage() ?? "An unknown error was encountered";
            MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "executeScheduledPaymentEvent: {$error}");
            return new MM_PaymentServiceResponse($error, MM_PaymentServiceResponse::$ERROR);
        }
    }


    /**
     * processBillingFailure handles when a scheduled payment has failed or encountered an error (which is treated as a billing failure)
     *
     */
    public function processBillingFailure()
    {
        if ($this->status == MM_ScheduledPaymentEvent::$EVENT_PROCESSED) {
            //take no action if event is already marked as processed. Should not happen
            return;
        }

        if ($this->getBillingStatus() == MM_ScheduledPaymentEvent::$LEGACY_EVENT) {
            //Legacy events were created prior to local billing. 
            //State was all kept remote, so we have to infer what state the subscription is in by examining the data.
            //If the subscription has a null processed_date, then this is the first failure, set the status to pending so it is handled correctly later on

            $processedDate = $this->getProcessedDate();
            if (!empty($processedDate)) {
                //There is a processed date, so this subscription is somewhere in the dunning cycle
                //For simplicity, assume that the status is "first failure"
                $this->setBillingStatus(MM_ScheduledPaymentEvent::$FIRST_REBILL_FAILED);
            } else {
                $this->setBillingStatus(MM_ScheduledPaymentEvent::$PENDING_REBILL);
            }
        }

        //Now progress this event through the dunning cycle
        switch ($this->getBillingStatus()) {
            case MM_ScheduledPaymentEvent::$EVENT_PENDING:
                $this->setBillingStatus(MM_ScheduledPaymentEvent::$FIRST_REBILL_FAILED);
                break;
            case MM_ScheduledPaymentEvent::$FIRST_REBILL_FAILED:
                $this->setBillingStatus(MM_ScheduledPaymentEvent::$SECOND_REBILL_FAILED);
                break;
            case MM_ScheduledPaymentEvent::$SECOND_REBILL_FAILED:
                $this->setBillingStatus(MM_ScheduledPaymentEvent::$THIRD_REBILL_FAILED);
                break;
            case MM_ScheduledPaymentEvent::$THIRD_REBILL_FAILED:
            default:
                //cancel after 3rd failed billing attempt
                try {
                    //the try-catch here accounts for the scenario where the cancelling the subscription fails (ie. from a broken notification handler)
                    //Catching the error gives the app a chance to set the billing status to failed, so that it won't attempt to bill this event again
                    $orderItem = new MM_OrderItem($this->getOrderItemId());
                    if ($orderItem->isValid()) {
                        $paymentService = MM_PaymentServiceFactory::getPaymentServiceById($this->paymentServiceId);
                        if ($paymentService instanceof MM_PaymentService) {
                            $paymentService->cancelSubscription($orderItem);
                        }
                        MM_AccessControlEngine::revokeAccessForOrderItem($orderItem);
                    }
                } catch (Exception $e) {
                    MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "Exception cancelling order item {$this->getOrderItemId()}: " . $e->getMessage());
                } catch (Error $e) {
                    MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "Error cancelling order item {$this->getOrderItemId()}: " . $e->getMessage());
                }
                $this->setBillingStatus(MM_ScheduledPaymentEvent::$PERMANENTLY_FAILED);
                break;
        }
        $this->setProcessedDate(MM_Utils::getCurrentTime());
        $this->commitData(); //save the billing status and update processed date
    }


    public static function getBillingFloorDate()
    {
        return static::BILLING_FLOOR_DATE;
    }
}
