<?php

/**
 *
 * MemberMouse(TM) (http://www.membermouse.com)
 * (c) MemberMouse, LLC. All rights reserved.
 *
 * MM_ScheduledReminder represents a reminder that is scheduled to be dispatched
 *
 */

use membermouse\resque\ScheduledReminderJob;

class MM_ScheduledReminder extends MM_Entity
{
    protected $scheduledDate        = "";
    protected $status               = "";
    protected $processedDate        = "";
    protected $batchId              = null;
    protected $userId               = "";
    protected $orderId              = "";
    protected $orderItemId          = "";
    protected $reminderId           = "";
    protected $estimatedRebillDate  = "";

    //reminder status
    const PENDING_STATUS = 0;
    const PROCESSING_STATUS = 1;
    const PROCESSED_STATUS = 2;

    //resque job link types
    public static $REMINDER_JOB_TYPE = 2;

    //configurable window for reminder eligibility
    const DEFAULT_REMINDER_WINDOW_SIZE = 18; //in hours

    public function __construct($id = "", $getData = true)
    {
        parent::__construct($id, $getData);
    }


    public function getData()
    {
        global $wpdb;

        if (intval($this->id) > 0) {
            $sql = "SELECT * from " . MM_TABLE_SCHEDULED_REMINDERS . " where id={$this->id}";
            $result = $wpdb->get_row($sql);
            if ($result) {
                $this->setData($result);
            } else {
                parent::invalidate();
            }
        } else {
            parent::invalidate();
        }
    }


    public function setData($data)
    {
        $this->orderId = $data->order_id;
        $this->orderItemId = $data->order_item_id;
        $this->userId = $data->user_id;
        $this->reminderId = $data->reminder_id;
        $this->scheduledDate = $data->scheduled_date;
        $this->status = $data->status;
        $this->processedDate = $data->processed_date;
        $this->batchId = $data->batch_id;
        $this->estimatedRebillDate = $data->estimated_rebill_date;
        if ($this->id > 0) {
            $this->validate();
        }
    }


    public function commitData()
    {
        global $wpdb;

        //determine if this is a new record or an update
        $newRecord = empty($this->id);
        $errorMsg = __("Error scheduling reminder:", 'membermouse');

        //make sure all required fields are set
        if (empty($this->reminderId)) {
            $ridError = __("reminder id cannot be blank", "membermouse");
            MM_DiagnosticLog::log("Error committing scheduled reminder: reminderId cannot be blank", MM_DiagnosticLog::$MM_ERROR);
            return new MM_Response("{$errorMsg}: {$ridError}", MM_Response::$ERROR);
        }

        if (empty($this->scheduledDate)) {
            $sdateError = __("scheduled date cannot be blank", "membermouse");
            MM_DiagnosticLog::log("Error committing scheduled reminder: scheduledDate cannot be blank", MM_DiagnosticLog::$MM_ERROR);
            return new MM_Response("{$errorMsg}: {$sdateError}", MM_Response::$ERROR);
        }


        $reminderDetailData = array(
            "user_id"                    => $this->userId,
            "order_item_id"              => $this->orderItemId,
            "order_id"                  => $this->orderId,
            "reminder_id"                  => $this->reminderId,
            "status"                    => $this->status,
            "scheduled_date"            => $this->scheduledDate,
            "processed_date"            => $this->processedDate,
            "batch_id"                  => empty($this->batchId) ? null : $this->batchId,
            "estimated_rebill_date"     => $this->estimatedRebillDate

        );

        if ($newRecord) {
            //this is a new record
            $result = $wpdb->insert(MM_TABLE_SCHEDULED_REMINDERS, $reminderDetailData);
            if ($result === false) {
                return new MM_Response("ERROR: unable to create scheduled reminder: error = " . $wpdb->last_error, MM_Response::$ERROR);
            }
            $this->id = $wpdb->insert_id;
        } else {
            $result = $wpdb->update(MM_TABLE_SCHEDULED_REMINDERS, $reminderDetailData, ["id" => $this->id]);
            if ($result === false) {
                return new MM_Response("ERROR: unable to update scheduled reminder: error = " . $wpdb->last_error, MM_Response::$ERROR);
            }
        }

        parent::validate();
        return new MM_Response();
    }


    public function delete()
    {
        global $wpdb;

        if (intval($this->id) > 0) {
            $res = $wpdb->query("DELETE FROM " . MM_TABLE_SCHEDULED_REMINDERS . " WHERE id='{$this->id}'");
            if ($res !== false) {
                parent::invalidate();
                return new MM_Response();
            } else {
                return new MM_Response("Error deleting scheduled reminder", MM_Response::$ERROR);
            }
        } else {
            return new MM_Response("Unable to delete this scheduled reminder because it is invalid", MM_Response::$ERROR);
        }
    }


    /**
     * Returns an array of ids for scheduled reminders ready to be dispatched immediately. 
     */
    public static function assignScheduledRemindersToBatchAndRetrieve()
    {
        global $wpdb;

        $currentDate = MM_Utils::getCurrentTime();
        $batchId = MM_Utils::createRandomString(32);
        $sql = "UPDATE " . MM_TABLE_SCHEDULED_REMINDERS . " SET batch_id = %s WHERE (scheduled_date <= %s) AND (status = %d) AND " .
            "(batch_id IS NULL)";
        $sql = $wpdb->prepare($sql, $batchId, $currentDate, MM_ScheduledReminder::PENDING_STATUS);
        $result = $wpdb->query($sql);
        if ($result !== false) {
            $sql = $wpdb->prepare("SELECT id FROM " . MM_TABLE_SCHEDULED_REMINDERS . " WHERE batch_id = %s", $batchId);
            $currentBatchScheduleIds = $wpdb->get_col($sql);
            return $currentBatchScheduleIds;
        } else {
            MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "Error setting batch id for reminder event dispatch");
            return false;
        }
    }


    /**
     * Queries the active subscriptions to identify which ones are eligible to have reminders sent, based on the configuration of the active reminders.
     * Matching subscriptions are assigned a batch id and then sent to Resque
     */
    public static function scanForReminders()
    {
        global $wpdb;

        $reminderWindowLength = apply_filters(MM_Filters::$REMINDER_WINDOW_SIZE, self::DEFAULT_REMINDER_WINDOW_SIZE); //in hours
        $orderItemsTable = MM_TABLE_ORDER_ITEMS;
        $ordersTable = MM_TABLE_ORDERS;
        $reminderProductsTable = MM_TABLE_REMINDERS_PRODUCTS;
        $scheduledRemindersTable = MM_TABLE_SCHEDULED_REMINDERS;

        $sql = "SELECT * FROM " . MM_TABLE_REMINDERS . " WHERE (status = 1)";
        $billingReminders = $wpdb->get_results($sql);

        foreach ($billingReminders as $brRow) {
            //The query plan is more favorable when pre-calculating the current date/time and adding the reminder interval, and using that in a simple date comparison, vs 
            //taking the estimated_rebill_date - reminder interval, and comparing that to the current date/time
            $hasProductFilter = ($brRow->has_product_filter == 1);
            $interval = "{$brRow->unit_value} {$brRow->unit_of_measure}";
            $nowTS = MM_Utils::getCurrentTime("timestamp");
            $targetTS = strtotime("+{$interval}", $nowTS);
            $targetDate = date("Y-m-d H:i:s", $targetTS);
            $nowDate = date("Y-m-d H:i:s", $nowTS);
            $windowEnd = date("Y-m-d H:i:s", strtotime("-{$reminderWindowLength} hours", $targetTS));

            //if there is a product filter, use a join against mm_reminders_products to eliminate non-eligible rows
            $productJoin = "";
            if ($hasProductFilter) {
                //This will filter returned subscriptions by those matching the configured product filter
                $productJoin = "INNER JOIN {$reminderProductsTable} rp ON ((oi.item_type = " . MM_OrderItem::$ORDER_ITEM_TYPE_PRODUCT . ") " .
                    "AND (oi.item_id = rp.product_id) AND (rp.reminder_id = {$brRow->id})) ";
            }
            $sql = "SELECT oi.id AS order_item_id, oi.estimated_rebill_date ,o.id AS order_id, o.user_id " .
                "FROM {$orderItemsTable} oi INNER JOIN {$ordersTable} o ON (oi.order_id = o.id) " .
                "LEFT JOIN {$scheduledRemindersTable} sr ON ((sr.reminder_id = {$brRow->id}) AND " .
                "(oi.estimated_rebill_date = sr.estimated_rebill_date) AND (oi.id = sr.order_item_id)) " .
                "{$productJoin}" .
                "WHERE (sr.estimated_rebill_date IS NULL) AND (oi.status = 1) AND (oi.is_recurring = 1) AND (oi.item_type = " . MM_OrderItem::$ORDER_ITEM_TYPE_PRODUCT . ") " .
                "AND (oi.estimated_rebill_date IS NOT NULL) AND (oi.estimated_rebill_date BETWEEN '{$windowEnd}' AND '{$targetDate}') AND (oi.estimated_rebill_date > '{$nowDate}')";
            $subscriptions = $wpdb->get_results($sql);
            foreach ($subscriptions as $subscription) {
                $sr = new MM_ScheduledReminder();
                $sr->setOrderId($subscription->order_id);
                $sr->setOrderItemId($subscription->order_item_id);
                $sr->setUserId($subscription->user_id);
                $sr->setReminderId($brRow->id);
                $sr->setEstimatedRebillDate($subscription->estimated_rebill_date);
                $sr->setStatus(MM_ScheduledReminder::PENDING_STATUS);
                $sr->setScheduledDate($nowDate);
                $sr->commitData();
            }
            unset($subscriptions); //release memory held by $subscriptions
        }

        //Assign a batch id to pending reminders without a batch id, and then send the batch to Resque
        $scheduledReminderIds = static::assignScheduledRemindersToBatchAndRetrieve();
        if (!empty($scheduledReminderIds)) {
            foreach ($scheduledReminderIds as $scheduledReminderId) {
                $job = new ScheduledReminderJob();
                $job->scheduled_reminder_id = $scheduledReminderId;
                $jobId = $job->enqueue("now", 20);
                if (is_numeric($jobId)) {
                    //link the newly created job with the scheduled reminder
                    $wpdb->insert(
                        MM_TABLE_RESQUE_JOB_LINKS,
                        ["link_type" => self::$REMINDER_JOB_TYPE, "remote_id" => $scheduledReminderId, "job_id" => $jobId]
                    );
                }
            }
        }
    }


    /**
     * Event-handler for $BILLING_SUBSCRIPTION_CANCELED event. Removes any reminders attached to the cancelled order item
     * 
     * @param array $data
     */
    public static function subscriptionCancelled($data)
    {
        global $wpdb;

        if (!empty($data["order_item_id"]) && is_numeric($data["order_item_id"])) {
            $orderItemId = intval($data["order_item_id"]);
            $wpdb->delete(MM_TABLE_SCHEDULED_REMINDERS, ["order_item_id" => $orderItemId, "status" => MM_ScheduledReminder::PENDING_STATUS]);
        }
    }


    /**
     * Dispatches this reminder by setting up and emitting the MM_Event::$BILLING_REMINDER event
     * By default this event is hooked by the RemindersEngine
     */
    public function dispatch()
    {
        $orderItem = MM_OrderItem::create($this->orderItemId);
        if ($orderItem->isValid() && $orderItem->isRecurring() && ($orderItem->getStatus() == MM_OrderItem::$STATUS_RECURRING)) {
            //emit the billing reminder event
            MM_Event::safeDoAction(MM_Event::$BILLING_REMINDER, MM_Event::packageReminderData($this));
        }
        //consider the reminder processed if a job is created successfully
        $this->setStatus(MM_ScheduledReminder::PROCESSED_STATUS);
        $this->setProcessedDate(MM_Utils::getCurrentTime());
        $this->commitData();
    }


    public function getOrderId()
    {
        return $this->orderId;
    }


    public function setOrderId($orderId)
    {
        $this->orderId = $orderId;
    }


    public function getOrderItemId()
    {
        return $this->orderItemId;
    }


    public function setOrderItemId($orderItemId)
    {
        $this->orderItemId = $orderItemId;
    }


    public function getUserId()
    {
        return $this->userId;
    }


    public function setUserId($userId)
    {
        $this->userId = $userId;
    }


    public function getReminderId()
    {
        return $this->reminderId;
    }


    public function setReminderId($reminderId)
    {
        $this->reminderId = $reminderId;
    }


    /**
     * @return number
     */
    public function getEstimatedRebillDate()
    {
        return $this->estimatedRebillDate;
    }

    /**
     * @param number $estimatedRebillDate
     */
    public function setEstimatedRebillDate($estimatedRebillDate)
    {
        $this->estimatedRebillDate = $estimatedRebillDate;
    }
    /**
     * @return number
     */
    public function getScheduledDate()
    {
        return $this->scheduledDate;
    }

    /**
     * @return number
     */
    public function getStatus()
    {
        return $this->status;
    }

    /**
     * @return number
     */
    public function getProcessedDate()
    {
        return $this->processedDate;
    }

    /**
     * @return number
     */
    public function getBatchId()
    {
        return $this->batchId;
    }

    /**
     * @param number $scheduledDate
     */
    public function setScheduledDate($scheduledDate)
    {
        $this->scheduledDate = $scheduledDate;
    }

    /**
     * @param number $status
     */
    public function setStatus($status)
    {
        $this->status = $status;
    }

    /**
     * @param number $processedDate
     */
    public function setProcessedDate($processedDate)
    {
        $this->processedDate = $processedDate;
    }

    /**
     * @param number $batchId
     */
    public function setBatchId($batchId)
    {
        $this->batchId = $batchId;
    }
}
