<?php

/**
 *
 * MemberMouse(TM) (http://www.membermouse.com)
 * (c) Membermouse, LLC. All rights reserved.
 *
 * MM_ShippingMethod is the base class for all shipping methods that are supported by membermouse. The concept is that each shipping method
 * (no matter how it obtains its rates) can present shipping options to an external service along with a unique key. That key can be used to identify the
 * option selected and determine a price. 
 *
 */

abstract class MM_ShippingMethod extends MM_Entity
{

    protected $name;
    protected $active = false;
    protected $token = "";

    //shipping method tokens
    public static $FLATRATE_SHIPPING = "FLATRATE";

    abstract public function getShippingOptions($order = "");

    abstract public function getShippingOptionByKey(MM_Order $order, $shippingOptionKey);

    public function install()
    {
        require_once(ABSPATH . 'wp-admin/includes/upgrade.php');
    }


    public function getName()
    {
        return $this->name;
    }


    public function setName($name)
    {
        $this->name = $name;
    }


    public function getToken()
    {
        return $this->token;
    }


    public function __construct()
    {
        $this->getData();
    }

    /**
     * getData retrieves the shipping method information from the database by token, and then calls setData to populate the entity
     * @see MM_Entity::getData()
     */
    public function getData()
    {
        global $wpdb;

        $sql = "SELECT * FROM " . MM_TABLE_SHIPPING_METHODS . " WHERE token='" . $this->token . "'";
        $result = $wpdb->get_row($sql);

        if ($result) {
            $this->setData($result);
        } else {
            parent::invalidate();
        }
    }


    /**
     * setData populates the entity using an object with containing identically named fields which contain the data to use
     * @see MM_Entity::setData()
     */
    public function setData($data)
    {
        try {
            $this->id = $data->id;
            $this->name = $data->name;
            $this->active = $data->active;

            if (($data->settings != null) && ($data->settings != "")) {
                $unserialized_settings = unserialize($data->settings);
                if (is_array($unserialized_settings)) {
                    foreach ($unserialized_settings as $k => $v) {
                        if (isset($this->$k)) {
                            $this->$k = $v; //don't create attributes, only restore existing ones
                        }
                    }
                }
            }

            parent::validate();
        } catch (Exception $ex) {
            parent::invalidate();
        }
    }


    /**
     * Each implementation of the ShippingMethod class can have specific settings. This model is identical to the one used by MM_PaymentService and allows for unforseen attributes that need to be stored
     * to be persisted along with the shipping class, without the need for additional tables. For example, if a provider supplied rates via a web service, the shipping method could store the 
     * username/password necessary to make the service call. Settings are stored as attributes on the object. Prior to commit, an array is created using these attributes and then serialized and stored in $this->settings, 
     * which is persisted when commitData is called
     */
    protected function prepareSettings()
    {
        $exclusions = array('id', 'name', 'token', 'settings', 'active');
        $settings = array();
        foreach ($this as $key => $value) {
            if (!in_array($key, $exclusions) && (strpos($key, "transient_") !== 0)) {
                $settings[$key] = $value;
            }
        }
        return serialize($settings);
    }


    /**
     * commitData persists the state and settings of the ShippingMethod into the database. The prepareSettings method is used to gather all the method
     * specific settings and serialize them into one database field, allowing this generic commitData to be used for all future subclasses
     * @see MM_Entity::commitData()
     */
    public function commitData()
    {
        global $wpdb;

        $doUpdate = ($this->id > 0);
        $operation = ($doUpdate) ? "update" : "create"; //used in error messages

        if (!$doUpdate) {
            //there can only be one row with each token, check if it exists first
            $checkSql = "select id from " . MM_TABLE_SHIPPING_METHODS . " where token='{$this->token}'";
            $res = $wpdb->get_results($checkSql);
            if ($res) {
                //if it does, this is an update. set the id for good measure
                $doUpdate = true;
                $row = $res[0];
                $this->id = $row->id;
            }
        }

        MM_Transaction::begin();
        try {
            $setClause = "SET token='{$this->token}', name='%s',settings='%s',active='%d'";

            if (!$doUpdate) {
                $sql = "INSERT into " . MM_TABLE_SHIPPING_METHODS . " {$setClause}";
            } else {
                $sql = "UPDATE " . MM_TABLE_SHIPPING_METHODS . " {$setClause} where token='{$this->token}'";
            }

            $preparedSql = $wpdb->prepare($sql, $this->name, $this->prepareSettings(), ($this->active) ? 1 : 0);
            $result = $wpdb->query($preparedSql);

            if ($result === false) {
                MM_Transaction::rollback();
                return new MM_Response("ERROR: unable to {$operation} shipping method (" . $preparedSql . ")", MM_Response::$ERROR);
            }
        } catch (Exception $ex) {
            MM_Transaction::rollback();
            return new MM_Response("ERROR: unable to {$operation} shipping method", MM_Response::$ERROR);
        }

        MM_Transaction::commit();
        return new MM_Response();
    }


    /**
     * Returns all of the available shipping options across all of the active methods
     * 
     * @param MM_Order $order (optional) The order to calculate rates against
     * @return array containing all of the available shipping options.
     */
    public static function getAvailableShippingOptions($order = "")
    {
        global $wpdb;

        $availableShippingOptions = array();

        $lookupMethodSQL = "SELECT token FROM " . MM_TABLE_SHIPPING_METHODS . " WHERE active='1'";
        $methods = $wpdb->get_col($lookupMethodSQL);

        if (!is_null($methods)) {
            foreach ($methods as $shippingMethodToken) {
                $method = MM_ShippingMethod::getShippingMethodByToken($shippingMethodToken);
                if (!is_null($method)) {
                    $availableShippingOptions += $method->getShippingOptions($order);
                }
            }
        }

        return $availableShippingOptions;
    }


    /**
     * This is a static factory method used to obtain a specific implementor of MM_ShippingMethod
     * 
     * @param string $token The token reference
     * @return MM_ShippingMethod subclass identified by the token on success, null on failure
     */
    public static function getShippingMethodByToken($token)
    {
        switch ($token) {
            case MM_ShippingMethod::$FLATRATE_SHIPPING:
                return new MM_FlatRateShippingMethod();
                break;
            default:
                return null;
        }
        return null;
    }
}
