<?php

/**
 * Expression evaluator for MemberMouse.
 *
 * @package MemberMouse
 */
class MM_ExpressionEvaluator
{
    /**
     * Operator definitions with precedence and associativity.
     *
     * @var array<string, array<string, mixed>>
     */
    private const OPERATORS = [
        'AND' => [
            'precedence' => 1,
            'associativity' => 'left',
            'symbol' => '&',
            'name' => 'AND',
        ],
        'OR' => [
            'precedence' => 1,
            'associativity' => 'left',
            'symbol' => '|',
            'name' => 'OR',
        ],
        'NOT' => [
            'precedence' => 2,
            'associativity' => 'right',
            'symbol' => '!',
            'name' => 'NOT',
        ],
    ];

    /**
     * Special characters used in expressions.
     *
     * @var array<string>
     */
    private const SPECIAL_CHARS = ['(', ')', ' ', '&', '|', '!'];

    /**
     * Cache for evaluated expressions to improve performance.
     *
     * @var array<string, boolean>
     */
    private array $cache = [];

    /**
     * Evaluates a boolean expression against a set of values.
     *
     * @param string $expression    The expression to evaluate.
     * @param array  $compareValues The values to compare against.
     *
     * @return boolean The result of the evaluation.
     */
    public function evaluate(string $expression, array $compareValues): bool
    {
        $cacheKey = $expression . '_' . md5(serialize($compareValues));
        if (isset($this->cache[$cacheKey])) {
            return $this->cache[$cacheKey];
        }

        // Check if expression contains any operators or parentheses.
        if (!$this->hasOperators($expression)) {
            return $this->handleDirectComparison($expression, $compareValues, $cacheKey);
        }

        // Handle NOT keyword in direct comparison.
        if (strpos($expression, 'NOT') !== false) {
            $parts = explode('NOT', $expression);
            if (count($parts) === 2) {
                // Direct NOT expressions are considered invalid and return false.
                return $this->cacheAndReturn($cacheKey, false);
            }
        }

        // Process complex expressions.
        return $this->processComplexExpression($expression, $compareValues, $cacheKey);
    }

    /**
     * Checks if an expression contains any operators.
     *
     * @param string $expression The expression to check.
     *
     * @return boolean True if the expression contains operators.
     */
    private function hasOperators(string $expression): bool
    {
        foreach (self::OPERATORS as $operator) {
            if (
                strpos($expression, $operator['symbol']) !== false ||
                strpos($expression, $operator['name']) !== false
            ) {
                return true;
            }
        }
        return false;
    }

    /**
     * Handles direct comparison without operators.
     *
     * @param string $expression    The expression to evaluate.
     * @param array  $compareValues The values to compare against.
     * @param string $cacheKey      The cache key.
     *
     * @return boolean The result of the comparison.
     */
    private function handleDirectComparison(string $expression, array $compareValues, string $cacheKey): bool
    {
        $expression = trim($expression);
        foreach ($compareValues as $value) {
            if ((string)trim($value) === (string)$expression) {
                return $this->cacheAndReturn($cacheKey, true);
            }
        }
        return $this->cacheAndReturn($cacheKey, false);
    }

    /**
     * Processes a complex expression with operators.
     *
     * @param string $expression    The expression to evaluate.
     * @param array  $compareValues The values to compare against.
     * @param string $cacheKey      The cache key.
     *
     * @return boolean The result of the evaluation.
     */
    private function processComplexExpression(string $expression, array $compareValues, string $cacheKey): bool
    {
        // Convert the expression to use word operators and ensure proper spacing.
        $expression = str_replace(
            ['!', '|', '&'],
            ['NOT ', ' OR ', ' AND '],
            $expression
        );

        // Ensure proper spacing around parentheses.
        $expression = str_replace(['(', ')'], [' ( ', ' ) '], $expression);
        $expression = preg_replace('/\s+/', ' ', trim($expression));

        $tokens = $this->tokenize($expression);
        $postfix = $this->infixToPostfix($tokens);

        // If postfix is empty (invalid expression), return false.
        if (empty($postfix)) {
            return $this->cacheAndReturn($cacheKey, false);
        }

        $result = $this->evaluatePostfix($postfix, $compareValues);
        return $this->cacheAndReturn($cacheKey, $result);
    }

    /**
     * Caches and returns a result.
     *
     * @param string  $cacheKey The cache key.
     * @param boolean $result   The result to cache and return.
     *
     * @return boolean The cached result.
     */
    private function cacheAndReturn(string $cacheKey, bool $result): bool
    {
        $this->cache[$cacheKey] = $result;
        return $result;
    }

    /**
     * Tokenizes an expression into an array of tokens.
     *
     * @param string $expression The expression to tokenize.
     *
     * @return array<string> The array of tokens.
     */
    private function tokenize(string $expression): array
    {
        $tokens = [];
        $currentToken = '';
        $length = strlen($expression);

        for ($i = 0; $i < $length; $i++) {
            $char = $expression[$i];

            if (in_array($char, self::SPECIAL_CHARS, true)) {
                if ($currentToken !== '') {
                    $tokens[] = $currentToken;
                    $currentToken = '';
                }
                if ($char !== ' ') {
                    $tokens[] = $char;
                }
            } else {
                $currentToken .= $char;
            }
        }

        if ($currentToken !== '') {
            $tokens[] = $currentToken;
        }

        return $tokens;
    }

    /**
     * Converts an infix expression to postfix notation.
     *
     * @param array<string> $tokens The tokens of the infix expression.
     *
     * @return array<string> The tokens in postfix notation.
     */
    private function infixToPostfix(array $tokens): array
    {
        $output = [];
        $operators = [];

        foreach ($tokens as $token) {
            if ($token === '(') {
                $operators[] = $token;
            } elseif ($token === ')') {
                while (!empty($operators) && end($operators) !== '(') {
                    $output[] = array_pop($operators);
                }
                array_pop($operators);
            } elseif ($this->isOperator($token)) {
                while (
                    !empty($operators) &&
                    end($operators) !== '(' &&
                    (
                        self::OPERATORS[$token]['precedence'] > self::OPERATORS[$token]['precedence'] ||
                        (
                            self::OPERATORS[$token]['precedence'] === self::OPERATORS[$token]['precedence'] &&
                            self::OPERATORS[$token]['associativity'] === 'left'
                        )
                    )
                ) {
                    $output[] = array_pop($operators);
                }
                $operators[] = $token;
            } else {
                $output[] = $token;
            }
        }

        while (!empty($operators)) {
            if (end($operators) === '(') {
                return [];
            }
            $output[] = array_pop($operators);
        }

        return $output;
    }

    /**
     * Evaluates a postfix expression.
     *
     * @param array<string> $tokens        The tokens in postfix notation.
     * @param array         $compareValues The values to compare against.
     *
     * @return boolean The result of the evaluation.
     */
    private function evaluatePostfix(array $tokens, array $compareValues): bool
    {
        $stack = [];
        foreach ($tokens as $token) {
            if ($this->isOperator($token)) {
                if ($token === 'NOT') {
                    if (empty($stack)) {
                        return false;
                    }
                    $operand = array_pop($stack);
                    $stack[] = !$operand;
                } else {
                    if (count($stack) < 2) {
                        return false;
                    }
                    $right = array_pop($stack);
                    $left = array_pop($stack);
                    if ($token === 'AND') {
                        $stack[] = $left && $right;
                    } else {
                        $stack[] = $left || $right;
                    }
                }
            } else {
                $value = false;
                foreach ($compareValues as $compareValue) {
                    if ((string)trim($compareValue) === (string)$token) {
                        $value = true;
                        break;
                    }
                }
                $stack[] = $value;
            }
        }
        return !empty($stack) && end($stack);
    }

    /**
     * Checks if a token is an operator.
     *
     * @param string $token The token to check.
     *
     * @return boolean True if the token is an operator.
     */
    private function isOperator(string $token): bool
    {
        return in_array($token, ['NOT', 'AND', 'OR'], true);
    }
}
