<?php

/**
 *
 * MemberMouse(TM) (http://www.membermouse.com)
 * (c) MemberMouse, LLC. All rights reserved.
 */
class MM_AdminNotice
{
    private $id = "";
    private $type = "";
    private $title = false;
    private $message = "";
    private $options = [];
    private $displayed = false;

    const TYPE_INFO = "info";
    const TYPE_SUCCESS = "success";
    const TYPE_ERROR = "error";
    const TYPE_WARNING = "warning";
    const VALID_TYPES = [
        MM_AdminNotice::TYPE_INFO,
        MM_AdminNotice::TYPE_SUCCESS,
        MM_AdminNotice::TYPE_ERROR,
        MM_AdminNotice::TYPE_WARNING,
    ];

    const DEFAULT_TYPE = MM_AdminNotice::TYPE_INFO;
    const DEFAULT_TITLE = false;
    const DEFAULT_OPTIONS = [
        "customClasses"   => false,
        "behavior"          => "collapse",
        "retainStatus"      => true,
        "titleSize"          => "4",
        "capability"      => "edit_theme_options",
        "activeScreens"   => [],
        "inactiveScreens" => ["site-health", "update"],
    ];

    const STATUS_PREFIX = "mm_notice_status_";
    const STATUS_SET_ACTION = "mm_notice_set_status";
    const STATUS_DISMISSED = "dismissed";
    const STATUS_COLLAPSED = "collapsed";
    const STATUS_EXPANDED = "expanded";
    const VALID_STATUSES = [
        MM_AdminNotice::STATUS_DISMISSED,
        MM_AdminNotice::STATUS_COLLAPSED,
        MM_AdminNotice::STATUS_EXPANDED,
    ];


    /**
     * Constructor sets default options.
     */
    public function __construct()
    {
        $this->type = self::DEFAULT_TYPE;
        $this->title = self::DEFAULT_TITLE;
        $this->options = self::DEFAULT_OPTIONS;
    }


    /**
     * Setup admin notices by adding relevant actions.
     * @return void
     */
    public static function setup(): void
    {
        add_action("admin_notices", "MM_AdminNotice::displayAll");
        add_action("admin_enqueue_scripts", "MM_AdminNotice::enqueueAdminStyles");
        add_action("current_screen", "MM_AdminNotice::removeAllDisplayed");

        if (wp_doing_ajax()) {
            add_action("wp_ajax_mm_notice_set_status", "MM_AdminNotice::statusHandler");
        }
    }


    /**
     * Enqueue admin styles.
     * @return void
     */
    public static function enqueueAdminStyles(): void
    {
        wp_enqueue_style(
            "membermouse-admin-notice",
            MM_PLUGIN_URL . "/resources/css/admin/mm-adminnotice.css",
            array(),
            MemberMouse::getPluginVersion()
        );
    }

    /**
     * Add an admin notice. 
     * @param string	   $id		  A unique ID for the notice. Required.
     * @param string|bool  $title	  Title of the notice, or false to prevent a title from being displayed. Required. Note that if dismissible is set 
     *								  to false, a title must be specified for a desirable appearance.
     * @param string	   $message   The message. May contain HTML. If plain text, it will be enclosed in a paragraph tag. Required.
     * @param string	   $type	  The type of notice. Default "info", valid types are "info", "success", "error", or "warning".
     * @param array		   $options   An array of additional options, with defaults:
     *					   [
     *							"customClasses"   => (string|bool)  A space-deliminated string of additional classes to add to the notice. Default false.
     *							"behavior"		  => (string)		Notice behavior, values are "collapse" (default), "dismiss", or false for neither.
     *							"retainStatus"	  => (bool)			Whether to retain the notice dismissed/collapsed status across page loads. Default true.
     *							"titleSize"		  => (int|bool)		The size of the title, an integer from 1 to 6 or false. Default 4, corrosponding to
     *																<h4>. When a numeric is specified, the class is set to "mm-notice-title". Specifying
     *																false sets the default class "notice-title" which overrides font size settings.
     *							"capability"	  => (string)		The capability required to display the notice. Default "edit_theme_options".
     *							"activeScreens"   => (array)		An array of screen IDs that the notice should be displayed on. Default empty array.
     *							"inactiveScreens" => (array)		An array of screen IDs that the notice should not be displayed on. Default array
     *																containing "site-health".
     *					   ]
     * @return void
     */
    public static function add($id, $title, $message, $type = "info", $options = []): void
    {
        $notice = new MM_AdminNotice();

        $notice->setId($id);
        $notice->setTitle($title);
        $notice->setMessage($message);
        $notice->setType($type);
        $notice->setOptions($options);

        $notices = self::retrieveNotices();

        if (!$notices) {
            $notices = array();
        }

        $notices[$id] = $notice;
        self::storeNotices($notices);

        if ($notice->shouldDisplay()) {
            $notice->addManagementScript();
        }

        if (did_action("admin_notices")) {
            $message = "Admin Notice ID '{$id}' was added, but the admin_notices action has already fired. " .
                "The notice will display on the next page load.";
            MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, $message);
        }
    }


    /**
     * Remove an admin notice.
     * @param string $id The ID of the notice to remove. Required.
     * @return void
     */
    public static function remove($id): void
    {
        $notices = self::retrieveNotices();
        unset($notices[$id]);
        self::storeNotices($notices);
    }


    /**
     * Remove all admin notices.
     * @return void
     */
    public static function removeAll(): void
    {
        MM_Session::sessionStart();
        MM_Session::clear(MM_Session::$KEY_ADMIN_NOTICES);
    }


    /**
     * Remove all admin notices that have already displayed.
     * @return void
     */
    public static function removeAllDisplayed(): void
    {
        $notices = self::retrieveNotices();

        if ($notices) {
            foreach ($notices as $notice) {
                if ($notice->hasDisplayed()) {
                    unset($notices[$notice->getId()]);
                }
            }
            self::storeNotices($notices);
        }
    }


    /**
     * Display all admin notices.
     * @param bool $clearNotices Whether to clear the notices after displaying them. Default true.
     * @return void
     */
    public static function displayAll(): void
    {
        $notices = self::retrieveNotices();

        if ($notices) {
            foreach ($notices as $notice) {
                echo $notice->buildNotice();
                $notice->setDisplayed();
            }
            self::storeNotices($notices);
        }
    }


    /**
     * Build a single admin notice.
     * @return string HTML representing the notice, or empty string if the notice should not be displayed.
     */
    private function buildNotice(): string
    {
        if ($this->shouldDisplay()) {
            $classes  = $this->buildClasses();
            $content  = $this->buildTitle();
            $content .= $this->buildMessage();

            return sprintf(
                '<div id="%1$s" class="%2$s">%3$s</div>',
                'mm-notice-' . esc_attr($this->id),
                esc_attr($classes),
                $content,
            );
        } else {
            return "";
        }
    }


    /**
     * Build the classes for the notice, based on the options
     * @return string A space-deliminated string of CSS classes.
     */
    private function buildClasses(): string
    {
        $classes = "notice notice-" . $this->type . " ";
        if (!empty($this->options["customClasses"])) {
            $classes .= $this->options["customClasses"] . " ";
        }

        if ($this->options["behavior"] == "dismiss") {
            $classes .= "is-dismissible";
        } else if ($this->options["behavior"] == "collapse") {
            $classes .= "mm-is-collapsible";
        }

        return $classes;
    }


    /**
     * Build the notice title, based on the options.
     * @return string HTML representing the notice title, or empty string if notice has no title.
     */
    private function buildTitle(): string
    {
        if (!empty($this->title)) {
            $size = $this->options["titleSize"];
            if ($size) {
                $titleClass = "mm-notice-title";
                if (!($size == floor($size) && $size >= 1 && $size <= 6)) {
                    $size = 2;
                }
            } else {
                $titleClass = "notice-title";
                $size = 2;
            }
            return sprintf('<h%1$s class="%2$s">%3$s</h%1$s>', $size, $titleClass, $this->title);
        }

        return "";
    }


    /**
     * Build the notice message. If the notice message is plain text, it will be enclosed in a paragraph tag. If it is HTML, it will be left as-is.
     * Non-dismissible notices have a button added to collapse the notice.
     * @return string HTML representing the notice message
     */
    private function buildMessage(): string
    {
        $message = $this->message;
        $maybeCollapseButton = "";
        $maybeHiddenStyle = "";

        if ($message == strip_tags($message)) {
            $message = "<p>{$message}</p>";
        }

        if ($this->options["behavior"] == "collapse") {
            $readerText = __("Collapse this notice", "membermouse");
            $maybeCollapseButton = '<button type="button" class="mm-notice-collapse">';
            $maybeCollapseButton .= sprintf('<span class="screen-reader-text">%s</span></button>', $readerText);

            if ($this->getNoticeStatus() == MM_AdminNotice::STATUS_COLLAPSED) {
                $maybeHiddenStyle = ' style="display:none;"';
            }
        }

        return sprintf('<div class="notice-message"%1$s>%2$s</div>%3$s', $maybeHiddenStyle, $message, $maybeCollapseButton);
    }


    /**
     * Evaluate whether the notice should be displayed, based on options, the screen, and whether the user has dismissed it.
     * @return bool Whether the notice should be displayed
     */
    public function shouldDisplay(): bool
    {
        if (!current_user_can($this->options["capability"])) {
            return false;
        }

        if ($this->wasDismissed()) {
            return false;
        }

        $activeScreens = $this->options["activeScreens"];
        $inactiveScreens = $this->options["inactiveScreens"];
        $currentScreen = get_current_screen()->id;

        if (is_null($currentScreen)) {
            return false;
        }

        if (!$activeScreens || empty($activeScreens) || in_array($currentScreen, $activeScreens)) {
            if (in_array($currentScreen, $inactiveScreens)) {
                return false;
            }

            return true;
        }

        return false;
    }


    /**
     * Get the status of the notice for the current user.
     * @return string|false The status of the notice, or false if status has not been set.
     */
    public function getNoticeStatus()
    {
        return (get_user_meta(get_current_user_id(), self::STATUS_PREFIX . $this->id, true));
    }


    /**
     * Set status of the notice for the current user.
     * @param string $status The status of the notice, constants STATUS_DISMISSED and STATUS_COLLAPSED are acceptable values.
     * @return void
     */
    public function setNoticeStatus($status): void
    {
        update_user_meta(get_current_user_id(), self::STATUS_PREFIX . $this->id, $status);
    }


    /**
     * Evaluate whether the notice is dismissible and was previously dismissed for the current user.
     * @return bool Representing whether the notice was previously dismissed
     */
    private function wasDismissed(): bool
    {
        if ($this->options["behavior"] !== "dismiss") {
            return false;
        }

        return ($this->getNoticeStatus() == MM_AdminNotice::STATUS_DISMISSED);
    }


    /**
     * Evaluate whether a notice is configured to retain its status.
     * @return bool Representing whether the notice can retain its status
     */
    public function canRetainStatus(): bool
    {
        return ($this->options["retainStatus"] == true);
    }


    /**
     * Clear dismissed/collapsed status of the admin notice for the current user.
     * @param string $id The ID of the notice
     * @return void
     */
    public function clearNoticeStatus(): void
    {
        delete_user_meta(get_current_user_id(), self::STATUS_PREFIX . $this->id);
    }


    /**
     * Handler for AJAX requests from the management script to dismiss an admin notice.
     * @return void
     */
    public static function statusHandler(): void
    {
        if (!isset($_REQUEST["action"]) || $_REQUEST["action"] !== self::STATUS_SET_ACTION) {
            return;
        }

        if (!isset($_REQUEST["id"]) || !is_string($_REQUEST["id"])) {
            return;
        }

        if (!isset($_REQUEST["status"]) || !is_string($_REQUEST["status"]) || !in_array($_REQUEST["status"], self::VALID_STATUSES)) {
            return;
        }

        check_ajax_referer(self::STATUS_PREFIX . $_REQUEST["id"], "nonce", true);

        $notices = self::retrieveNotices();

        if (is_array($notices) && in_array($_REQUEST["id"], array_keys($notices))) {
            $notice = $notices[$_REQUEST["id"]];

            if ($notice->canRetainStatus()) {
                if ($_REQUEST["status"] == MM_AdminNotice::STATUS_DISMISSED) {
                    $notice->setNoticeStatus(MM_AdminNotice::STATUS_DISMISSED);
                    unset($notices[$_REQUEST["id"]]);
                    self::storeNotices($notices);
                }

                if ($_REQUEST["status"] == MM_AdminNotice::STATUS_COLLAPSED) {
                    $notice->setNoticeStatus(MM_AdminNotice::STATUS_COLLAPSED);
                }

                if ($_REQUEST["status"] == MM_AdminNotice::STATUS_EXPANDED) {
                    $notice->clearNoticeStatus();
                }
            }
        }
    }


    /**
     * Adds the management script for this notice inline after common.js.
     * @return bool|null Whether the script was added. True if added, false if failed to add, null if not required.
     */
    public function addManagementScript(): ?bool
    {
        $nonce = wp_create_nonce(self::STATUS_PREFIX . $this->id);
        $script = $this->buildManagementScript($nonce);
        $behavior = $this->options["behavior"];

        if (!empty($script)) {
            if (($behavior == "collapse") || ($behavior == "dismiss" && $this->canRetainStatus())) {
                return wp_add_inline_script("common", $script, "after");
            }
        }
        return null;
    }


    /**
     * Builds the management script for this notice. The script makes an AJAX request to statusHandler() which will persist either 
     * dismissed, or collapsde/expanded status of the notice.
     * @param string $nonce The nonce for the AJAX request
     * @return string The management script for this notice
     */
    private function buildManagementScript($nonce): string
    {
        $id = esc_attr($this->id);
        $ajaxUrl = esc_url(admin_url("admin-ajax.php"));
        $action = self::STATUS_SET_ACTION;
        $statusDismissed = MM_AdminNotice::STATUS_DISMISSED;
        $statusCollapsed = MM_AdminNotice::STATUS_COLLAPSED;
        $statusExpanded = MM_AdminNotice::STATUS_EXPANDED;
        $script = "";

        if ($this->options["behavior"] == "dismiss") {
            $script = <<<EOD
			jQuery(function() {
				var dismissButton = document.querySelector("#mm-notice-$id .notice-dismiss");
				dismissButton.addEventListener("click", async function(event){
					await fetch("$ajaxUrl", {
						method: "POST",
						headers:{
							"Content-Type": "application/x-www-form-urlencoded"
						},	
						body: new URLSearchParams({
							"id": "$id",
							"action": "$action",
							"status": "$statusDismissed",
							"nonce": "$nonce"
						})
					}).then((response) => {
						if (response.status >= 400 && response.status < 600) {
							throw new Error(response.status + " " + response.statusText);
						}
						return response;
					}).catch((error) => {
						console.log("MemberMouse Admin Notice AJAX Request Failed - " + error);
					});
				});
			});
			EOD;
        } else if ($this->options["behavior"] == "collapse") {
            $id = esc_attr($this->id);
            $script = <<<EOD
			jQuery(function() {
				var collapseButton = document.querySelector("#mm-notice-$id .mm-notice-collapse");
				collapseButton.addEventListener("click", async function(event){
					var message = document.querySelector("#mm-notice-$id .notice-message");
					var isDisplayed = message.style.display !== "none";
					var targetStatus = "";
					if(isDisplayed) {
						jQuery(message).slideUp();
						targetStatus = "$statusCollapsed";
					} else {
						jQuery(message).slideDown();
						targetStatus = "$statusExpanded";
					}
					await fetch("$ajaxUrl", {
						method: "POST",
						headers:{
							"Content-Type": "application/x-www-form-urlencoded"
						},	
						body: new URLSearchParams({
							"id": "$id",
							"action": "$action",
							"status": targetStatus,
							"nonce": "$nonce"
						})
					}).then((response) => {
						if (response.status >= 400 && response.status < 600) {
							throw new Error(response.status + " " + response.statusText);
						}
						return response;
					}).catch((error) => {
						console.log("MemberMouse Admin Notice AJAX Request Failed - " + error);
					});
				});
			});
			EOD;
        }

        return $script;
    }


    /**
     * Retrieves notices from session
     * @return array|false An array of MM_AdminNotice objects, or false if none in session
     */
    private static function retrieveNotices()
    {
        MM_Session::sessionStart();
        return MM_Session::value(MM_Session::$KEY_ADMIN_NOTICES);
    }


    /**
     * Store notices in session, or clear session if all notices have been removed
     * @param array $notices An array of MM_AdminNotice objects, or an empty array if none to be written
     * @return void
     */
    private static function storeNotices($notices): void
    {
        if (empty($notices)) {
            MM_Session::sessionStart();
            MM_Session::clear(MM_Session::$KEY_ADMIN_NOTICES);
        } else {
            // Setting a value starts session automatically
            MM_Session::value(MM_Session::$KEY_ADMIN_NOTICES, $notices);
        }
    }


    /**
     * Gets the ID of the notice.
     * @param string $id The unique ID of the notice.
     * @return string
     */
    public function getId(): string
    {
        return $this->id;
    }


    /**
     * Sets the ID of the notice.
     * @param string $id The unique ID of the notice.
     * @return void
     */
    public function setId($id): void
    {
        $this->id = $id;
    }


    /**
     * Sets the type of the notice.
     * @param string $type Notice type. Valid types are: success, info, warning, error. If an invalid type is specified no action will be taken.
     * @return void
     */
    public function setType($type): void
    {
        if (!empty($type) && is_string($type) && in_array($type, self::VALID_TYPES)) {
            $this->type = $type;
        }
    }


    /**
     * Sets the notice title.
     * @param string|bool $title The title of the notice, or false if no title should be shown.
     * @return void
     */
    public function setTitle($title): void
    {
        $this->title = $title;
    }


    /**
     * Sets the notice message.
     * @param string $message The message to be displayed. May be plain text or HTML.
     * @return void
     */
    public function setMessage($message): void
    {
        $this->message = $message;
    }


    /**
     * Sets the options for the notice.
     * @param array $options Overrides for the default options, @see MM_AdminNotice::add() for valid options and defaults.
     * @return void
     */
    public function setOptions($options): void
    {
        if (!empty($options) && is_array($options)) {
            $this->options = wp_parse_args($options, $this->options);
        }
    }


    /**
     * Sets a flag indicating notice display was attempted. Used to mark notices as eligible for removal next time a page is loaded.
     * @param bool $displayed Whether the notice was displayed, default true.
     * @return void
     */
    public function setDisplayed($displayed = true): void
    {
        $this->displayed = (bool)$displayed;
    }


    /**
     * Evaluate whether the notice has already been displayed.
     * @return bool Representing whether the notice was previously displayed
     */
    public function hasDisplayed(): bool
    {
        return ($this->displayed == true);
    }
}
