<?php

/**
 * 
 * MemberMouse(TM) (http://www.membermouse.com)
 * (c) MemberMouse, LLC. All rights reserved.
 */
class MM_TransactionHistoryView extends MM_RBACGeneralEmployeeView
{
    public static $MM_JSACTION_SAVE_TRANSACTION = "saveTransaction";
    public static $MM_JSACTION_SHOW_EDIT_TRANSACTION = "showEditDialog";
    public static $MM_JSACTION_RESEND_AFFILIATE_COMMISSION = "resendAffiliateCommission";
    public static $MM_JSACTION_REFUND_TRANSACTION = "refundTransaction";

    protected $headers = [];
    protected $renderForAdmin = true;


    public function __construct()
    {
        parent::__construct();
    }


    /**
     * AJAX handler
     * {@inheritDoc}
     * @see MM_View::performAction()
     * @param  array $post The post request
     * @return mixed	   The response
     */
    public function performAction($post)
    {
        $response = parent::performAction($post);

        if (!($response instanceof MM_Response)) {
            switch ($post[self::$MM_JSACTION]) {
                case self::$MM_JSACTION_SAVE_TRANSACTION:
                    return $this->saveTransaction($post);

                case self::$MM_JSACTION_SHOW_EDIT_TRANSACTION:
                    return $this->showEditTransactionDialog($post);

                case self::$MM_JSACTION_RESEND_AFFILIATE_COMMISSION:
                    return $this->resendAffiliateCommission($post);

                case self::$MM_JSACTION_REFUND_TRANSACTION:
                    return $this->refundTransaction($post);

                default:
                    return new MM_Response($response);
            }
        } else {
            return $response;
        }
    }


    /**
     * Save an "edited" transaction, meaning affiliate information may have been updated.
     * @param  array $post Data from the post request. At minimum, must contain order_id and order_item_id
     * @return mixed The result of the operation.
     */
    public function saveTransaction($post)
    {
        if (!isset($post["order_id"])) {
            return new MM_Response("Unable to update transaction. Invalid order passed.", MM_Response::$ERROR);
        }

        if (!isset($post["order_item_id"])) {
            return new MM_Response("Unable to update transaction. Invalid order item passed.", MM_Response::$ERROR);
        }

        $order = MM_Order::create($post["order_id"]);

        if ($order->isValid()) {
            if (isset($post["mm_affiliate_id"])) {
                $order->setAffiliateId($post["mm_affiliate_id"]);
            }

            if (isset($post["mm_sub_affiliate_id"])) {
                $order->setSubAffiliateId($post["mm_sub_affiliate_id"]);
            }

            $result = $order->commitData();

            if ($order->isValid()) {
                $orderItem = MM_OrderItem::create($post["order_item_id"]);

                do_action(
                    MM_Event::$AFFILIATE_INFO_CHANGED,
                    MM_Event::packageOrderData($order->getCustomer()->getId(), $order->getId(), $orderItem->getId())
                );
            }

            if (MM_Response::isSuccess($result)) {
                return new MM_Response(__("Transaction updated successfully.", "membermouse"));
            } else {
                return $result;
            }
        }
    }


    /**
     * Resend an affiliate commission
     * @param  array $post	Data from the post request. At minimum, must contain user_id, order_id, order_item_id, 
     *						and transaction_id
     * @return MM_Response	The result of the operation
     */
    public function resendAffiliateCommission($post)
    {
        global $wpdb;

        if (!isset($post["user_id"])) {
            return new MM_Response("Unable to send affiliate commission. Invalid user ID passed.", MM_Response::$ERROR);
        }

        if (!isset($post["order_id"])) {
            return new MM_Response("Unable to send affiliate commission. Invalid order passed.", MM_Response::$ERROR);
        }

        if (!isset($post["order_item_id"])) {
            return new MM_Response("Unable to send affiliate commission. Invalid order item passed.", MM_Response::$ERROR);
        }

        if (!isset($post["transaction_id"])) {
            return new MM_Response("Unable to send affiliate commission. Invalid transaction ID passed.", MM_Response::$ERROR);
        }

        $affiliateService = MM_AffiliateProviderFactory::getActiveProvider();

        if ($affiliateService->supportsFeature(MM_AffiliateProviderFeatures::COMMISSION_TRACKING)) {
            $orderData = MM_Event::packageOrderData($post["user_id"], $post["order_id"], $post["order_item_id"], $post["transaction_id"]);

            // determine if transaction in an initial or rebill payment
            $sql = "SELECT transaction_type FROM " . MM_TABLE_TRANSACTION_LOG . " WHERE id = %s";
            $result = $wpdb->get_row($wpdb->prepare($sql, $post["transaction_id"]));

            if (!is_null($result) && isset($result->transaction_type)) {
                $affiliateController = new MM_AffiliateController();
                $isRebill = false;
                $doSendCommission = false;

                if ($result->transaction_type == MM_TransactionLog::$TRANSACTION_TYPE_RECURRING_PAYMENT) {
                    $isRebill = true;
                }

                // determine if commission should be sent based on commission profile
                $commissionProfile = null;
                $orderItem = MM_OrderItem::create($post["order_item_id"]);

                if ($orderItem->isValid()) {
                    $product = MM_Product::create($orderItem->getItemId());

                    if ($product->isValid()) {
                        $commissionProfile = $product->getCommissionProfile();

                        if (!$commissionProfile->isValid()) {
                            $commissionProfile = null;
                        }
                    }
                }

                if (!is_null($commissionProfile)) {
                    if (!$isRebill) {
                        if ($commissionProfile->initialCommissionEnabled()) {
                            $doSendCommission = true;
                        } else {
                            return new MM_Response(__("Initial commissions are not enabled for the commission profile associated with this product.", "membermouse"), MM_Response::$ERROR);
                        }
                    } else {
                        if ($commissionProfile->rebillCommissionsEnabled()) {
                            $doSendCommission = true;
                        } else {
                            return new MM_Response(__("Rebill commissions are not enabled for the commission profile associated with this product.", "membermouse"), MM_Response::$ERROR);
                        }
                    }
                } else {
                    return new MM_Response("Unable to find commission profile.", MM_Response::$ERROR);
                }

                if ($doSendCommission) {
                    $affiliateController->trackCommission($orderData, $isRebill);
                    return new MM_Response("Affiliate commission request sent");
                }
            }
        } else {
            return new MM_Response("Unable to send affiliate commission. The active affiliate service doesn't support commission tracking.", MM_Response::$ERROR);
        }

        return new MM_Response("Unable to send affiliate commission.", MM_Response::$ERROR);
    }


    /**
     * Returns HTML for the template-generated Edit Transaction dialog as the data of an MM_Response
     * @param  array $post	Data from the post request. Minimum required data are order_id, order_item_id, 
     *						and transaction_id.
     * @return MM_Response	The response
     */
    public function showEditTransactionDialog($post)
    {
        if (!isset($post["order_id"])) {
            return new MM_Response("Error editing transaction. Order ID is required.", MM_Response::$ERROR);
        }

        if (!isset($post["order_item_id"])) {
            return new MM_Response("Error editing transaction. Order Item ID is required.", MM_Response::$ERROR);
        }

        if (!isset($post["transaction_id"])) {
            return new MM_Response("Error editing transaction. Transaction ID is required.", MM_Response::$ERROR);
        }

        $info = new stdClass();
        $info->orderId = $post["order_id"];
        $info->orderItemId = $post["order_item_id"];
        $info->transactionId = $post["transaction_id"];

        $data = MM_TEMPLATE::generate(MM_MODULES . "/" . MM_MODULE_MEMBER_DETAILS_TRANSACTION_HISTORY . ".dialog.php", $info);

        return new MM_Response($data);
    }


    /**
     * Processes a refund
     * @param  array $post	Data from the post request, minimum required data is transaction_id.
     * @return MM_Response	The response, indicating success or failure
     */
    public function refundTransaction($post)
    {
        if (!empty($post["transaction_id"])) {
            $transactionId = $post["transaction_id"];
            $transactionLookupResponse = MM_TransactionLog::getTransactionInfo($transactionId);
            if (MM_Response::isError($transactionLookupResponse)) {
                return $transactionLookupResponse;
            }
            $transactionInfo = $transactionLookupResponse->message;
            $paymentService = MM_PaymentServiceFactory::getPaymentServiceById($transactionInfo->payment_service_id);
            if (is_null($paymentService)) {
                return new MM_Response("Unable to refund transaction: Invalid payment method");
            }
            $refundPaymentServiceResponse = $paymentService->refundTransaction($transactionId);
            //translate the paymentServiceResponse into an MM_Response. All non-success states become errors
            if (MM_PaymentServiceResponse::isSuccess($refundPaymentServiceResponse)) {
                return new MM_Response(__("Refund was successful.", "membermouse"));
            } else {
                return new MM_Response($refundPaymentServiceResponse->getMessage(), MM_Response::$ERROR);
            }
        } else {
            //shouldnt ever happen through the interface
            return new MM_Response("Unable to refund transaction: No transaction id was supplied", MM_Response::$ERROR);
        }
    }


    /**
     * Retrieves data for the view. 
     * @param  integer		$userId		The ID of the member whose data will be retrieved
     * @param  MM_DataGrid	$dg			A datagrid object, used to determine sort order and limits.
     * @param  string		$startDate	Start date, not used in this view implementation
     * @return array		An array of stdClass objects representing rows, each with the data for one transaction.
     *						Properties: affiliateId, amount, currency, date, is_test, isRefunded, mmTransId, 
     *						order_item_id, orderId, ordernumber, productId, productname, pymtServiceName, 
     *						pymtServiceToken, recurringDiscount, subAffiliateId, total, transtype, user_id
     */
    public function getViewData($userId, MM_DataGrid $dg = null, $startDate = null)
    {
        global $wpdb;

        //tables used in the master query. Placing the constants into variables allows them to be embedded into the string, aiding in readability
        $transactionsTable    = MM_TABLE_TRANSACTION_LOG;
        $paymentServicesTable = MM_TABLE_PAYMENT_SERVICES;
        $ordersTable          = MM_TABLE_ORDERS;
        $orderItemsTable      = MM_TABLE_ORDER_ITEMS;

        $desiredTransactionTypes = implode(",", array(MM_TransactionLog::$TRANSACTION_TYPE_PAYMENT, MM_TransactionLog::$TRANSACTION_TYPE_RECURRING_PAYMENT, MM_TransactionLog::$TRANSACTION_TYPE_REFUND));
        $refundTransctionType = MM_TransactionLog::$TRANSACTION_TYPE_REFUND;

        // get total of query
        $queryWithTotal = "SELECT count(*) as total  ";
        $queryWithTotal .= "FROM {$transactionsTable} t LEFT JOIN {$paymentServicesTable} ps on (t.payment_service_id =  ps.id) " .
            "LEFT JOIN {$ordersTable} o on (t.order_id = o.id) " .
            "LEFT JOIN {$orderItemsTable} oi on (t.order_item_id = oi.id) " .
            "WHERE (t.transaction_type IN ({$desiredTransactionTypes})) AND (o.user_id = %s) ";
        $queryWithTotal = $wpdb->prepare($queryWithTotal, $userId);
        $countResult = $wpdb->get_row($queryWithTotal);

        // place total result into master query
        $masterQuery = "SELECT {$countResult->total} as 'total', o.id as orderId, o.order_number as ordernumber, t.id as mmTransId, t.is_test, o.user_id, t.order_item_id, t.transaction_date as date, ifnull(oi.description,'multiple products') as productname, oi.item_id as productId, oi.recurring_discount as recurringDiscount, " .
            "t.amount, t.currency, t.transaction_type as transtype, ps.name as pymtServiceName, ps.token as pymtServiceToken, o.affiliate_id as affiliateId, o.sub_affiliate_id as subAffiliateId, " .
            "if(((t.refund_id IS NOT NULL) AND (t.transaction_type != {$refundTransctionType})),true,false) as isRefunded " .
            "FROM {$transactionsTable} t LEFT JOIN {$paymentServicesTable} ps on (t.payment_service_id =  ps.id) " .
            "LEFT JOIN {$ordersTable} o on (t.order_id = o.id) " .
            "LEFT JOIN {$orderItemsTable} oi on (t.order_item_id = oi.id) " .
            "WHERE (t.transaction_type IN ({$desiredTransactionTypes})) AND (o.user_id = %s) ";
        $masterQuery = $wpdb->prepare($masterQuery, $userId);

        if (!is_null($dg) && !is_null($dg->sortBy) && !empty($dg->sortBy)) //handle the sorting
        {
            $masterQuery .= " ORDER BY {$dg->sortBy} {$dg->sortDir}";
        }

        $masterQuery .= " " . $dg->getLimitSql();
        $results = $wpdb->get_results($masterQuery);

        if (count($results) > 0) {
            //retrieve coupon info with one query - first divide the results up by order id. Using order id as the key eliminates duplicate order ids and gives us an easy reference back to the transaction rows
            $orderIndex = [];
            foreach ($results as $rowObject) {
                if (!is_null($rowObject->orderId)) {
                    $orderIndex[$rowObject->orderId][] = $rowObject;
                }
            }

            //now retrieve all of the order items representing coupons, joined with the coupon table, for each of the orders the above transactions belong to
            $orderIdstring = implode(",", array_keys($orderIndex));
            $couponsTable = MM_TABLE_COUPONS;
            $couponType = MM_OrderItem::$ORDER_ITEM_TYPE_COUPON;
            $couponSql = "SELECT oi.order_id, oi.total as discountAmount, c.coupon_name, c.recurring_billing_setting FROM {$orderItemsTable} oi LEFT JOIN {$couponsTable} c " .
                "ON (oi.item_id = c.id) WHERE (oi.item_type = '{$couponType}') AND (oi.order_id IN ({$orderIdstring}))";
            $couponResults = $wpdb->get_results($couponSql);

            //finally, loop through the coupons and add their data to the transaction rows. Essentially we are deriving which transactions represent payments for items that had coupons applied
            //NOTE: This assumes that if a coupon exists in an order, there is only one product in that order. The moment multiple products are allowed in a single order that assumption becomes invalid, and this code becomes unreliable (for coupon detection).
            foreach ($couponResults as $additionalCouponData) {
                $currentOrderId = $additionalCouponData->order_id;
                foreach ($orderIndex[$currentOrderId] as $currentTransaction) {
                    switch ($currentTransaction->transtype) {
                        case MM_TransactionLog::$TRANSACTION_TYPE_PAYMENT:
                            $currentTransaction->couponApplied = true;
                            $currentTransaction->couponName = $additionalCouponData->coupon_name;
                            $currentTransaction->couponDiscount = abs(floatval($additionalCouponData->discountAmount));
                            break;
                        case MM_TransactionLog::$TRANSACTION_TYPE_RECURRING_PAYMENT:
                            if ($additionalCouponData->recurring_billing_setting == "all") {
                                $currentTransaction->couponApplied = true;
                                $currentTransaction->couponName = $additionalCouponData->coupon_name;
                                $currentTransaction->couponDiscount = "";

                                if (!is_null($currentTransaction->recurringDiscount)) {
                                    $currentTransaction->couponDiscount = $currentTransaction->recurringDiscount;
                                } else {
                                    $product = new MM_Product($currentTransaction->productId);

                                    if ($product->isValid()) {
                                        $currentTransaction->couponDiscount = $product->getPrice(false) - abs(floatval($currentTransaction->amount));
                                    }
                                }
                                break;
                            }
                            //break is intentionally in the if. If that condition isnt matched, we want it to go to the default
                        default:
                            $currentTransaction->couponApplied = false;
                            $currentTransaction->couponName = "";
                            $currentTransaction->couponDiscount = 0.00;
                            break;
                    }
                }
            }

            // retrieve all order items representing prorations and other discounts
            $orderIdstring = implode(",", array_keys($orderIndex));
            $prorationType = MM_OrderItem::$ORDER_ITEM_TYPE_PRORATION;
            $otherDiscountType  = MM_OrderItem::$ORDER_ITEM_TYPE_DISCOUNT;
            $prorationDiscountSql = "SELECT oi.order_id, oi.total as discountAmount,oi.item_type FROM {$orderItemsTable} oi WHERE (oi.item_type IN ('{$prorationType}','{$otherDiscountType}')) AND (oi.order_id IN ({$orderIdstring}))";
            $prorationDiscountResults = $wpdb->get_results($prorationDiscountSql);

            //finally, loop through the prorations and add their data to the transaction rows.
            foreach ($prorationDiscountResults as $additionalTransactionData) {
                $currentOrderId = $additionalTransactionData->order_id;
                foreach ($orderIndex[$currentOrderId] as $currentTransaction) {
                    switch ($currentTransaction->transtype) {
                        case MM_TransactionLog::$TRANSACTION_TYPE_PAYMENT:
                            if ($additionalTransactionData->item_type == $prorationType) {
                                $currentTransaction->prorationApplied = true;
                                $currentTransaction->prorationAmount = abs(floatval($additionalTransactionData->discountAmount));
                            } else if ($additionalTransactionData->item_type == $otherDiscountType) {
                                $currentTransaction->otherDiscountApplied = true;
                                $currentTransaction->otherDiscountAmount = abs(floatval($additionalTransactionData->discountAmount));
                            }
                            break;

                        case MM_TransactionLog::$TRANSACTION_TYPE_RECURRING_PAYMENT:
                        default:
                            $currentTransaction->prorationApplied = false;
                            $currentTransaction->prorationAmount = 0.00;
                            $currentTransaction->otherDiscountApplied = false;
                            $currentTransaction->otherDiscountAmount = 0.00;
                            break;
                    }
                }
            }
        }
        return $results;
    }


    /**
     * Generates and filters data for column headers. 
     * A header is an associative, nested array, with the following structure:
     * ['columnId' => ['content' => 'Title', 'attr' => 'style="width:20px;"']
     * The 'attr' key and value are optional and specify params applied to the <th> tag.
     * Default headers: ordernumber, date, productname, amount, transtype, affiliate, actions.
     * When rendering My Account, affiliate and actions are not included
     * @return boolean Indicates whether headers returned through the filter were malformed and removed
     */
    public function generateHeaders(): bool
    {
        if ($this->renderForAdmin) {
            $headers = [
                'ordernumber' => [
                    'content' => '<a onclick="mmjs.sort(\'ordernumber\');" href="#">' . __('Order #', 'membermouse') . '</a>',
                    'attr'    => 'style="width:65px;"'
                ],
                'date' => [
                    'content' => '<a onclick="mmjs.sort(\'date\');" href="#">' . __('Date', 'membermouse') . '</a>',
                    'attr'    => 'style="width:180px;"'
                ],
                'productname' => [
                    'content' => '<a onclick="mmjs.sort(\'productname\');" href="#">' . __('Product Name', 'membermouse') . '</a>',
                ],
                'amount' => [
                    'content' => '<a onclick="mmjs.sort(\'amount\');" href="#">' . __('Amount', 'membermouse') . '</a>',
                    'attr'    => 'style="width:120px;"'
                ],
                'transtype' => [
                    'content' => '<a onclick="mmjs.sort(\'transtype\');" href="#">' . __('Type', 'membermouse') . '</a>',
                    'attr'    => 'style="width:50px;"'
                ],
                'affiliate' => [
                    'content' => __('Affiliate', 'membermouse'),
                    'attr'    => 'style="width:60px;"'
                ],
                'actions' => [
                    'content' => __('Actions', 'membermouse'),
                    'attr'    => 'style="width:140px;"'
                ],
            ];
        } else {
            $headers = [
                'ordernumber'  => ['content' => __("Order #", "membermouse")],
                'date'         => ['content' => __("Order Date", "membermouse")],
                'productname'  => ['content' => __("Description", "membermouse")],
                'amount'       => ['content' => __("Amount", "membermouse")],
                'transtype'    => ['content' => __("Type", "membermouse")],
            ];
        }

        $dataIntegrity = true;
        $headers = apply_filters(
            MM_Filters::$ACCOUNT_TRANSACTIONS_COLS,
            $headers,
            $this->renderForAdmin,
        );

        // $key must not contain characters disallowed in WP filter names
        // $val must be an specially formed array, but here we reject only incorrect datatype
        foreach ($headers as $key => $val) {
            if (!preg_match("/^[a-zA-Z0-9_-]+$/", $key) || !is_array($val)) {
                unset($headers[$key]);
                $dataIntegrity = false;
            }
        }

        $this->headers = $headers;
        return $dataIntegrity;
    }


    /** 
     * Generates the rows of the table, in an order determined by the headers
     * @param  array $data	The data produced by getViewData()
     * @return array		An array of rows, each of which is an array of records. A record is an associative 
     * 						array: ['content' => 'Content', 'attr' => 'style="white-space:nowrap;"']. Only
     * 						the actions column includes an 'attr' element, applied automatically.
     */
    public function generateRows($data)
    {
        $rows = [];
        $headers = $this->headers;
        $dateFormat = ($this->renderForAdmin) ? "M j, Y g:i a" : "M j, Y";

        foreach ($data as $item) {
            $currentRow = [];
            foreach ($headers as $columnId => $columnData) {
                switch ($columnId) {
                    case "ordernumber":
                        $content = '<span style="font-family:courier;">' . $item->ordernumber . '</span>';
                        break;

                    case "date":
                        $content = MM_Utils::dateToLocal($item->date, $dateFormat);
                        break;

                    case "productname":
                        $content = $item->productname;
                        break;

                    case "amount":
                        $content = $this->generateAmountRecord($item);
                        break;

                    case "transtype":
                        $content = $this->generateTransactionTypeRecord($item);
                        break;

                    case "affiliate":
                        $content = $this->generateAffiliateRecord($item);
                        break;

                    case "actions":
                        $content = $this->generateActionsRecord($item);
                        break;

                    default:
                        $content = "";
                        break;
                }

                // Filter, validate returned type, prepare structure
                $filteredContent = apply_filters(
                    MM_Filters::getTransactionsGridFilterName($columnId),
                    $content,
                    $item,
                    $this->renderForAdmin
                );
                $filteredContent = (is_string($filteredContent)) ? $filteredContent : $content;
                $currentRecord = ["content" => $filteredContent];

                // Add nowrap style for the actions column
                if ($columnId == "actions") {
                    $currentRecord["attr"] = 'style="white-space:nowrap;"';
                }

                $currentRow[] = $currentRecord;
            }

            $rows[] = $currentRow;
        }

        return $rows;
    }


    /** 
     * Generates the content for the Transaction Type column
     * @param  object	$item A stdClass object with data about the associated transaction and user
     * @return string	The HTML contents of the table data element, not including the element itself.
     */
    protected function generateTransactionTypeRecord($item)
    {
        $pymtMethodInfoStr = "";

        if ($this->renderForAdmin) {
            $pymtMethodInfoStr = !empty($item->pymtServiceName) ? "\n" . __("Payment Method", "membermouse") . " : " . $item->pymtServiceName . "\n" : "";
        }

        if (($item->transtype == MM_TransactionLog::$TRANSACTION_TYPE_PAYMENT) || ($item->transtype == MM_TransactionLog::$TRANSACTION_TYPE_RECURRING_PAYMENT)) {
            $transType = MM_Utils::getIcon('money', 'green', '1.4em', '2px', __("Successful Payment", "membermouse") . " " . $pymtMethodInfoStr);
        } else if ($item->transtype == MM_TransactionLog::$TRANSACTION_TYPE_REFUND) {
            $transType = MM_Utils::getIcon('money', 'red', '1.4em', '2px',  __("Successful Refund", "membermouse") . " " . $pymtMethodInfoStr);
        } else {
            $transType = MM_NO_DATA;
        }

        if ($item->is_test) {
            $transType .= MM_Utils::getIcon('flask', 'grey', '1.3em', '2px', __("Test Transaction", "membermouse"), "padding-left:4px;");
        }

        return $transType;
    }


    /**
     * Generates the content for the Affiliate Info column
     * @param  object $item	A stdClass object with data about the associated transaction and user
     * @return string		The HTML contents of the table data element, not including the element itself.
     */
    protected function generateAffiliateRecord($item)
    {
        $affiliateInfo = "";
        if (!empty($item->affiliateId) || !empty($item->subAffiliateId)) {
            $affiliateInfoStr = "";

            if (!empty($item->affiliateId)) {
                $affiliateInfoStr .= __("Affiliate ID", "membermouse") . ": " . $item->affiliateId . "\n";
            }

            if (!empty($item->subAffiliateId)) {
                $affiliateInfoStr .= __("Sub-affiliate ID", "membermouse") . ": " . $item->subAffiliateId . "\n";
            }

            $affiliateInfo = MM_Utils::getAffiliateIcon($affiliateInfoStr);
        }

        return $affiliateInfo;
    }


    /**
     * Generates the content for the Actions column
     * @param  object $item	A stdClass object with data about the associated transaction and user
     * @return string		The HTML contents of the table data element, not including the element itself.
     */
    protected function generateActionsRecord($item)
    {
        global $wpdb;

        $actions = "";
        $crntPaymentService = null;
        $crntAffiliateService = null;
        $affiliateActivityLogEntries = null;

        if (($item->isRefunded != '1') && (($item->transtype == MM_TransactionLog::$TRANSACTION_TYPE_PAYMENT) || ($item->transtype == MM_TransactionLog::$TRANSACTION_TYPE_RECURRING_PAYMENT))) {
            $actions = '<a title="' . __("Issue Refund", "membermouse") . '" onclick="mmjs.issueRefund(' . $item->mmTransId . ',\'' . number_format(abs($item->amount), 2, '.', '') . '\')" style="cursor:pointer;">' . MM_Utils::getIcon('undo', 'red', '1.3em', '2px') . '</a>';

            // add affiliate service-specific actions
            if (!empty($item->affiliateId)) {
                if (is_null($crntAffiliateService)) {
                    $crntAffiliateService = MM_AffiliateProviderFactory::getActiveProvider();
                }

                if (is_null($affiliateActivityLogEntries)) {
                    // retrieve affiliate activity log entries for the current user from the database
                    $eventLogTable = MM_TABLE_EVENT_LOG;
                    $affiliateEvent = MM_ActivityLog::$EVENT_TYPE_AFFILIATE_TRACKING;
                    $affiliateActivitySql = "SELECT * FROM {$eventLogTable} WHERE user_id = '{$item->user_id}' AND event_type = '{$affiliateEvent}' ";
                    $results = $wpdb->get_results($affiliateActivitySql);

                    if ($results) {
                        $affiliateActivityLogEntries = [];

                        foreach ($results as $activityLogEntry) {
                            $params = unserialize($activityLogEntry->additional_params);

                            if (
                                $params[MM_ActivityLog::$PARAM_AFFILIATE_TRACKING_EVENT] == MM_AffiliateController::$AFFILIATE_EVENT_TRACK_INITIAL_COMMISSION
                                || $params[MM_ActivityLog::$PARAM_AFFILIATE_TRACKING_EVENT] == MM_AffiliateController::$AFFILIATE_EVENT_TRACK_REBILL_COMMISSION
                            ) {
                                $affiliateActivityLogEntries[] = $params;
                            }
                        }
                    }
                }

                if ($crntAffiliateService->supportsFeature(MM_AffiliateProviderFeatures::COMMISSION_TRACKING)) {
                    $showAffiliateCommissionBtn = true;

                    // check if commission has already been sent and logged in the activity log
                    if (!is_null($affiliateActivityLogEntries)) {
                        foreach ($affiliateActivityLogEntries as $activityLogEntry) {
                            $crntAffilateId = $item->affiliateId;
                            $crntOrderNumber = $item->ordernumber . "-" . $item->mmTransId;

                            if (
                                $crntOrderNumber == $activityLogEntry[MM_ActivityLog::$PARAM_AFFILIATE_ORDER_NUMBER]
                                && $crntAffilateId == $activityLogEntry[MM_ActivityLog::$PARAM_AFFILIATE_ID]
                            ) {
                                $showAffiliateCommissionBtn = false;
                                break;
                            }
                        }
                    }

                    if ($showAffiliateCommissionBtn) {
                        /* translators: %s: affiliate service provider name */
                        $buttonMsg = sprintf(__('Send Affiliate Commission to %s', 'membermouse'), $crntAffiliateService->getName());
                        $actions .= ' <a title="' . $buttonMsg . '" onclick="mmjs.resendAffiliateCommission(' . $item->user_id . ', ' . $item->orderId . ', ' . $item->ordernumber . ', ' . $item->order_item_id . ', ' . $item->mmTransId . ', ' . $item->affiliateId . ')" style="cursor:pointer;">' . MM_Utils::getIcon('paper-plane', 'light-blue', '1.3em', '2px') . '</a>';
                    }
                }
            }

            // add advanced transaction tools if the customer has access to advanced subscription mgmt features
            if ((MM_MemberMouseService::hasPermission(MM_MemberMouseService::$FEATURE_ADV_SUBSCRIPTION_MGMT) == MM_MemberMouseService::$ACTIVE)) {
                $editActionUrl = "href='javascript:mmjs.editTransaction({$item->orderId}, {$item->order_item_id}, {$item->mmTransId});'";
                $actions .= " " . MM_Utils::getEditIcon(__("Edit Transaction", "membermouse"), '', $editActionUrl);
            } else if (MM_MemberMouseService::hasPermission(MM_MemberMouseService::$FEATURE_ADV_SUBSCRIPTION_MGMT) != MM_MemberMouseService::$ACTIVE) {
                $upgradeMsg = __("Additional transaction management features are only available on the Pro plan and above. Visit membermouse.com to upgrade your plan.", "membermouse");
                $actions .= " " . MM_Utils::getEditIcon($upgradeMsg, '', '', true);
            }
        } else if ($item->isRefunded == '1') {
            $refundPymtInfoStr = "";

            if (!empty($item->refundPymtServiceName)) {
                $refundPymtInfoStr .= __("Payment Method", "membermouse") . ": " . $item->refundPymtServiceName . "\n";

                if (!empty($item->refundRefId)) {
                    $refundPymtInfoStr .= $item->refundPymtServiceName . " " . __("Reference ID", "membermouse") . ": " . $item->refundRefId;
                }
            }

            $actions = MM_Utils::getCheckIcon("Refund Issued\n{$refundPymtInfoStr}") . " <em>" . __("Refund Issued", "membermouse") . "</em>";
        }

        // add payment service-specific actions
        if (is_null($crntPaymentService) || $crntPaymentService->getToken() != $item->pymtServiceToken) {
            $crntPaymentService = MM_PaymentServiceFactory::getPaymentService($item->pymtServiceToken);
        }

        if ($crntPaymentService->supportsFeature(MM_PaymentServiceFeatures::TRANSACTION_PAGE_ACTIONS)) {
            $actions .= " " . $crntPaymentService->getTransactionPageActions($item->mmTransId);
        }

        return $actions;
    }


    /**
     * Generates the content for the Amount column
     * @param  object $item	A stdClass object with data about the associated transaction and user
     * @return string		The HTML contents of the table data element, not including the element itself.
     */
    protected function generateAmountRecord($item)
    {
        if (floatval($item->amount) == 0) {
            $desc = __("When requesting a refund, the related transaction can be in a settled or unsettled state. " .
                "For settled transactions, the funds have actually been debited from the customer's account so it's " .
                "possible for them to be refunded. For unsettled transactions, the debit is still pending so in this " .
                "case there's nothing to refund and the transaction is simply voided which cancels the request to " .
                "transfer the funds.", "membermouse");
            $amount = "<span style='font-family:courier; font-variant: small-caps;'>VOIDED</span>" . MM_Utils::getInfoIcon($desc);
        } else {
            $amount = "<span style='font-family:courier;'>" . _mmf(abs($item->amount), $item->currency) . "</span>";
        }

        $additionalRowInfo = "";
        $item->couponApplied = isset($item->couponApplied) ? $item->couponApplied : false;
        $item->prorationApplied = isset($item->prorationApplied) ? $item->prorationApplied : false;
        $item->otherDiscountApplied = isset($item->otherDiscountApplied) ? $item->otherDiscountApplied : false;
        $anyDiscounts = $item->couponApplied || $item->prorationApplied || $item->otherDiscountApplied;

        if ($anyDiscounts) {
            $totalDiscount = 0.00;
            $discountDescription = __("DISCOUNTS APPLIED", "membermouse") . "\n";

            if ($item->couponApplied) {
                if (!empty($item->couponDiscount)) {
                    $discountDescription .= __("Coupon", "membermouse") . ": " . _mmf($item->couponDiscount, $item->currency) . " ({$item->couponName})\n";
                    $totalDiscount += $item->couponDiscount;
                }
            }

            if ($item->prorationApplied) {
                if (!empty($item->prorationAmount)) {
                    $discountDescription .= __("Proration", "membermouse") . ": " . _mmf($item->prorationAmount, $item->currency) . "\n";
                    $totalDiscount += $item->prorationAmount;
                }
            }

            if ($item->otherDiscountApplied) {
                if (!empty($item->otherDiscountAmount)) {
                    $discountDescription .= __("Other discounts", "membermouse") . ": " . _mmf($item->otherDiscountAmount, $item->currency) . "\n";
                    $totalDiscount += $item->otherDiscountAmount;
                }
            }

            $discountDescription .= "---------------------\n";
            $discountDescription .= __("Total discount", "membermouse") . ": " . _mmf($totalDiscount, $item->currency);
            $additionalRowInfo .= MM_Utils::getDiscountIcon($discountDescription);
        }

        return "{$amount} {$additionalRowInfo}";
    }


    /**
     * Gets column header data, for use by the grid.
     * @return array	An associative array of data used by the grid. The key is the element ID.
     *					The value is an associative array with keys "content" and optionally "attr"
     */
    public function getHeaders(): array
    {
        return $this->headers;
    }


    /**
     * Sets whether to render the grid for admin
     * @param boolean	True to render for admin, false to render for the member
     */

    public function setRenderForAdmin($renderForAdmin): void
    {
        $this->renderForAdmin = (bool)$renderForAdmin;
    }
}
