<?php

/**
 * @package MemberMouse
 * @version 3.0.20
 *
 * Plugin Name: MemberMouse Platform
 * Plugin URI: http://membermouse.com
 * Description: MemberMouse is an enterprise-level membership platform that allows you to quickly and easily manage a membership site or subscription business. MemberMouse is designed to deliver digital content, automate customer self-service and provide you with advanced marketing tools to maximize the profitability of your continuity business.
 * Author: MemberMouse, LLC
 * Version: 3.0.20
 * Author URI: http://membermouse.com
 * Text Domain: membermouse
 * Domain Path: /languages/
 * Copyright: 2009-2024 MemberMouse LLC. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * Also add information on how to contact you by electronic and paper mail.
 */

require_once("includes/mm-constants.php");
require_once("includes/mm-functions.php");
require_once("includes/init.php");
require_once('lib/resque/autoloader.php');

if (MM_DiagnosticLog::isEnabled()) {
    function membermouseDiagnosticCapture()
    {
        $error = @error_get_last();
        if (($error != null) && ($error['type'] & (E_ERROR | E_USER_ERROR | E_PARSE | E_CORE_ERROR | E_COMPILE_ERROR | E_RECOVERABLE_ERROR))) {
            @MM_DiagnosticLog::logItem($error['file'], MM_DiagnosticLog::$PHP_ERROR, $error['message'], $error['line']);
        }
    }
    register_shutdown_function('membermouseDiagnosticCapture');
    $previous = set_error_handler(function ($errno, $errstr, $errfile, $errline, $errcontext = []) use (&$previous) {
        MM_DiagnosticLog::logPHPErrors($errno, $errstr, $errfile, $errline, $errcontext);

        // If another error handler was defined, call it, otherwise use the default PHP error handler
        if ($previous) {
            return $previous($errno, $errstr, $errfile, $errline, $errcontext);
        }
        return false;
    });
}

if (!class_exists('MemberMouse', false)) {
    class MemberMouse
    {
        private static $menus = '';
        private $option_name = 'membermouse-settings';
        private $installerRan = false;
        private static $pluginVersion = '3.0.20';

        public function __construct()
        {
            // check if the plugin has been upgraded to a new major version and if so, run the installer
            if (!isset($_GET["release"])) {
                $crntMajorVersion = self::getPluginVersion();
                $lastMajorVersion = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_MAJOR_VERSION);
                if (!empty($lastMajorVersion) && (version_compare($lastMajorVersion, $crntMajorVersion) < 0)) {
                    $this->install();
                }
            }

            $license = MM_License::getCurrentLicense();
            if (strpos($_SERVER["PHP_SELF"], "plugins.php") === false) {
                MM_MemberMouseService::validateLicense($license);
            }

            require_once("includes/php_interface.php");

            $this->addActions();
            $this->addFilters();
        }


        public function addAssignAccessHandler()
        {
?>
            <script type='text/javascript'>
                //// if Bulk-edit list of page/posts is clicked in, lets be sure
                //// our remove access list is up-to-date.
                jQuery('#bulk-edit fieldset').click(function(event) {
                    accessrights_js.getAccessRightsForPosts();
                });


                ///// If the doaction option list for bulk edit selectes "Manage Access" 
                ///// lets be sure to modify the options for display in the UI to show that of the access options.
                jQuery(document).ready(function() {
                    jQuery("#doaction").click(function(event) {
                        if (jQuery("#bulk-action-selector-top :selected").text() == "<?php _e("Manage Access", "membermouse"); ?>") {
                            event.preventDefault();
                            var total = jQuery("body").find('input[name="post[]"]:checked').length;

                            /// if >=1 post/page is selected lets show the dialog.
                            if (total > 0) {
                                /// lets reuse the bulk-edit screen 
                                jQuery("#bulk-action-selector-top :selected").val("edit");

                                /// lets hide the irrelevant options 
                                jQuery(".inline-edit-categories").hide();
                                jQuery(".inline-edit-col-right").hide();

                                /// lets now show the bulk-edit screen
                                window.inlineEditPost.setBulk();

                                /// lets fill it with access management options
                                accessrights_js.getAccessRightsForPosts(true);
                            }
                        } else {
                            //// if our custom bulk options are showing
                            if (jQuery("#mm-custom-bulk-id").length) {
                                /// remove them as needed
                                jQuery("#mm-custom-bulk-id").remove();

                                /// now lets reset the existing options for the edit screen
                                if (jQuery("#bulk-action-selector-top :selected").val() == "edit") {
                                    jQuery(".inline-edit-categories").show();
                                    jQuery(".inline-edit-col-right").show();
                                }
                            }
                        }
                    });
                });
            </script>
            <?php
        }


        public function addFilters()
        {
            $post_hook = new MM_PostHooks();

            ///// add ability to handle new option by inline js
            add_action("admin_footer", array($this, "addAssignAccessHandler"));

            //// show new bulk edit option
            $bulkActionsClosure = function ($bulk_actions) {
                $bulk_actions['assign-access'] = __('Manage Access', 'membermouse');
                return $bulk_actions;
            };
            add_filter('bulk_actions-edit-page', $bulkActionsClosure);
            add_filter('bulk_actions-edit-post', $bulkActionsClosure);

            //// handle the update/save action from the bulk-edit inline dialog
            add_action('save_post', array($post_hook, 'saveAccessRightsFromBulkEdit'), 10, 2);

            remove_filter('the_title', 'do_shortcode', 9);
            remove_filter('wp_title', 'do_shortcode', 9);

            add_filter('the_title', 'MM_TagProcessor::processSmartTags', 9);
            add_filter('wp_title', 'MM_TagProcessor::processSmartTags', 9);
            add_filter('the_content', 'MM_TagProcessor::processSmartTags', 9);

            add_filter('document_title_parts', function ($title) {
                if (is_array($title)) {
                    $title = array_map(array("MM_TagProcessor", "processSmartTags"), $title);
                }
                return $title;
            }, 9, 1);


            $user_hooks = new MM_UserHooks();
            add_filter('login_redirect', array($user_hooks, 'loginRedirect'), 1, 3);
            add_filter('template_redirect', array($user_hooks, 'handlePageAccess'));

            $useMMLoginPage = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_USE_MM_LOGIN_PAGE);
            if ($useMMLoginPage == "1") {
                add_filter('login_url',  array($user_hooks, "loginUrl"),  1, 2);
            }

            add_filter('logout_url',  array($user_hooks, "logoutUrl"),  1, 2);

            // add WP menu filters
            $showLoginLogoutLink = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_SHOW_LOGIN_LOGOUT_LINK);
            if ($showLoginLogoutLink == "1") {
                add_filter('wp_nav_menu_items', array($user_hooks, 'showLoginLogoutLinkInMenu'), 10, 2);
            }

            $hideMenuItems = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_HIDE_PROTECTED_MENU_ITEMS);
            if ($hideMenuItems == "1") {
                add_filter('wp_setup_nav_menu_item', array($user_hooks, 'hideProtectedMenuItems'));
            }

            add_filter('the_title', 'do_shortcode', 9);
            add_filter('wp_title', 'do_shortcode', 9);

            // add MM link footer
            if (MM_MemberMouseService::hasPermission(MM_MemberMouseService::$SHOW_MM_FOOTER) == MM_MemberMouseService::$ACTIVE) {
                add_filter('wp_footer', array($this, 'addMMFooter'), 9);
            }


            add_filter('rest_prepare_post', array($post_hook, "doRestFilter"), 12, 3);
            add_filter('rest_prepare_page', array($post_hook, "doRestFilter"), 12, 3);
            add_filter('manage_posts_columns', array($post_hook, 'postsColumns'), 5);
            add_filter('manage_pages_columns', array($post_hook, 'pagesColumns'), 5);
            add_filter('posts_where', array($post_hook, 'handlePostWhere'), 1, 1);
            add_filter('wp_head', array($post_hook, "checkPosts"));

            add_filter('the_content', array($this, "filterContent"), 10);

            //prevent members with pending status from logging in
            add_filter('authenticate', array($user_hooks, 'checkLogin'), 100, 3);

            //prevent wptexturize from running on mm_form tags
            add_filter('no_texturize_shortcodes', function ($shortcodes) {
                $shortcodes[] = 'MM_Form';
                return $shortcodes;
            });

            //default password strength validation - minimum of 8 chars
            if (class_exists("MM_Filters")) {
                add_filter(MM_Filters::$PASSWORD_STRENGTH_VALIDATOR, array('MM_UserHooks', "passwordStrengthValidator"), 10, 1);
            }

            add_filter('submenu_file', 'MM_AuthUtils::highlight_account_menu_item');
            add_filter('mm_stripe_payment_method_title', 'MM_StripeConnect::payment_method_title');
            add_filter('site_status_tests', 'MM_StripeConnect::add_site_health_test');

            //configure local billing queue
            add_filter('resque_jobs_config', array('MM_LocalBillingEngine', 'configureResque'));
        }


        public function addActions()
        {
            if (function_exists("add_action")) {
                add_filter("plugin_action_links", array($this, "updateVersion"), 5, 3);
                add_filter('pre_set_site_transient_update_plugins', array($this, "preSetSiteTransient"));

                $user_hooks = new MM_UserHooks();
                add_action('wp_login_failed', array($user_hooks, 'loginFailed'), 9999);
                add_action('authenticate', array($user_hooks, 'authenticateLogin'));

                //session functions
                add_action('init', function () {
                    if (MM_Session::sessionExists()) {
                        MM_Session::sessionStart();
                    }
                });

                add_action('shutdown', array('MM_Session', 'sessionWrite'));
                add_action('wp_logout', function () {
                    MM_Session::sessionReap(MM_Session::getSessionId());
                });

                add_action('init', array($user_hooks, "doAutoLogin"));
                add_action('init', array($user_hooks, "doAutoLogout"));
                add_action('init', array($user_hooks, "removeWPAutoPOnCorePages"));
                add_action("init", array($user_hooks, "setupDefinitions"));
                add_action('init', "MM_Utils::loadMembermouseTextDomain", 0);
                add_action("template_redirect", array($user_hooks, "checkCorePageTypeInput"));
                add_action("wp_footer", array($user_hooks, "pageBasedActions"));
                add_action("delete_user", array($user_hooks, "deleteUser"));

                add_action("wp_head", array($this, "loadPreviewBar"));
                add_action("admin_bar_menu", array($this, "customizeAdminBar"));

                add_action("wp_enqueue_scripts", array($this, 'loadResources'));

                if (MM_Utils::inAdmin()) {
                    add_action("admin_enqueue_scripts", array($this, 'loadResources'));
                    add_action('admin_init', array($user_hooks, "checkEmployeeAccess"));
                    add_action('admin_menu', array($this, 'buildAdminMenu'));
                    add_action("in_admin_header", array($this, "enqueueAdminNotices"));
                    add_action("admin_notices", array($this, "activationFailed"));
                    add_action("admin_init", array($this, 'configurePostMeta'));
                }
                MM_AdminNotice::setup();
                $smartTagLibrary = new MM_SmartTagLibraryView();
                add_action("admin_footer", array($smartTagLibrary, "addDialogContainers"));
                add_action('media_buttons_context', array($smartTagLibrary, "customMediaButtons"));


                $paymentUtils = new MM_PaymentUtilsView();
                add_action("admin_footer", array($paymentUtils, "addDialogContainers"));
                add_action("wp_footer", array($paymentUtils, "addDialogContainers"));


                $post_hook = new MM_PostHooks();
                add_action("trashed_post", array($post_hook, "trashPostHandler"));
                add_action("deleted_post", array($post_hook, "deletePostHandler"));
                add_action('manage_posts_custom_column', array($post_hook, 'postCustomColumns'), 5, 2);
                add_action('manage_pages_custom_column', array($post_hook, 'postCustomColumns'), 5, 2);
                add_action('restrict_manage_posts', array($post_hook, 'editPostsFilter'));
                add_action('publish_page', array($post_hook, 'publishPageHandler'), 10, 2);

                /// saveCorePages
                $corePageEngine = new MM_CorePageEngine();
                add_action('save_post', array($corePageEngine, 'saveCorePages'), 10, 2);

                if (MM_Utils::inAdmin()) {
                    add_action('update_option_permalink_structure', array("MM_CorePageEngine", "createCorePageCache"));
                }

                $protectedContent = new MM_ProtectedContentEngine();
                add_action('save_post', array($protectedContent, 'saveSmartContent'), 10, 2);

                add_action('wp_ajax_module-handle', array($this, 'handleAjaxCallback'));
                add_action('wp_ajax_nopriv_module-handle', array($this, 'handleAjaxCallback'));
                add_action('wp_ajax_member-types', array($this, 'handleAjaxCallback'));

                // load MemberMouse widgets
                add_action('widgets_init', function () {
                    return register_widget("MM_SmartWidget");
                });


                add_action('widgets_init', function () {
                    return register_widget("MM_DripContentWidget");
                });

                // add event listeners

                $pne = new MM_PushNotificationEngine();
                add_action(MM_Event::$MEMBER_ADD, array($pne, 'memberAdded'), 100, 2);
                add_action(MM_Event::$MEMBER_STATUS_CHANGE, array($pne, 'memberStatusChanged'), 100, 2);
                add_action(MM_Event::$MEMBER_MEMBERSHIP_CHANGE, array($pne, 'memberMembershipChanged'), 100, 2);
                add_action(MM_Event::$MEMBER_ACCOUNT_UPDATE, array($pne, 'memberAccountUpdated'), 100, 2);
                add_action(MM_Event::$MEMBER_DELETE, array($pne, 'memberDeleted'), 100, 2);
                add_action(MM_Event::$BUNDLE_ADD, array($pne, 'bundleAdded'), 100, 2);
                add_action(MM_Event::$BUNDLE_STATUS_CHANGE, array($pne, 'bundleStatusChanged'), 100, 2);
                add_action(MM_Event::$PAYMENT_RECEIVED, array($pne, 'paymentReceived'), 100, 2);
                add_action(MM_Event::$PAYMENT_REBILL, array($pne, 'rebillPaymentReceived'), 100, 2);
                add_action(MM_Event::$PAYMENT_REBILL_DECLINED, array($pne, 'rebillPaymentDeclined'), 100, 2);
                add_action(MM_Event::$REFUND_ISSUED, array($pne, 'refundIssued'), 100, 2);
                add_action(MM_Event::$BILLING_SUBSCRIPTION_UPDATED, array($pne, 'billingSubscriptionUpdated'), 100, 2);
                add_action(MM_Event::$COMMISSION_INITIAL, array($pne, 'initialCommission'), 100, 2);
                add_action(MM_Event::$COMMISSION_REBILL, array($pne, 'rebillCommission'), 100, 2);
                add_action(MM_Event::$CANCEL_COMMISSION, array($pne, 'cancelCommission'), 100, 2);
                add_action(MM_Event::$PRODUCT_PURCHASE, array($pne, 'productPurchase'), 100, 2);
                add_action(MM_Event::$BILLING_SUBSCRIPTION_REBILL_DATE_CHANGED, array($pne, 'billingSubscriptionRebillDateChanged'), 100, 2);
                add_action(MM_Event::$BILLING_SUBSCRIPTION_CANCELED, array($pne, 'billingSubscriptionCanceled'), 100, 2);
                add_action(MM_Event::$AFFILIATE_INFO_CHANGED, array($pne, 'affiliateInfoChanged'), 100, 2);
                add_action(MM_Event::$MEMBER_MERGED, array($pne, 'memberMerged'), 100, 2);

                //set up event hooks needed for reminders
                MM_RemindersEngine::setup();

                //add listeners to update the estimated next rebill date for subscriptions
                add_action(MM_Event::$PURCHASE_COMPLETE, array("MM_OrderItem", 'updateEstimatedRebillDate'));
                add_action(MM_Event::$PAYMENT_REBILL, array("MM_OrderItem", 'updateEstimatedRebillDate'));

                //add a filter which creates custom wp cron intervals
                add_filter('cron_schedules', array('MM_CronEngine', 'addCustomIntervalRecurrenceOptions'));
                MM_CronEngine::setup();

                MM_AffiliateController::setup();

                MM_EmailServiceProviderController::setup();

                $mdv = new MM_MemberDetailsView();
                add_action(MM_Event::$PAYMENT_REBILL, array($mdv, 'handleRebillPaymentReceivedEvent'), 10, 2);
                add_action(MM_Event::$PAYMENT_REBILL_DECLINED, array($mdv, 'handleRebillPaymentDeclinedEvent'), 10, 2);

                //add payment service hooks
                add_action('init', array('MM_PaymentService', "performInitActions"), 9);


                //// capture any session changes to the override key 
                add_action('wp_loaded', array('MM_PaymentServiceFactory', "syncPaymentServiceKeyOveride"));

                //add extension hooks
                add_action("init", array('MM_Extension', "performInitActions"), 9);

                register_activation_hook(__FILE__, array($this, 'install'));
                register_deactivation_hook(__FILE__, array($this, 'onDeactivate'));


                add_action('wp_footer', function () {
                    if (MM_Session::sessionExists() && (MM_Session::$MM_SESSION_DELAYED_CREATE === true)) {
                        echo MM_Session::generateDelayedCreateJavascript();
                    }
                });

                $file   = basename(__FILE__);
                $folder = basename(dirname(__FILE__));
                $hook = "in_plugin_update_message-{$folder}/{$file}";
                add_action($hook, 'MemberMouse::majorVersionUpdateCheck', 10, 2);


                add_action('admin_init', 'MM_AuthenticatorService::clear_connection_data');
                add_action('init', 'MM_GrowthTools::init');
                add_action('init', 'MM_GroundLevelControl::init', 5);
                add_action('init', 'MM_AuthenticatorService::process_connect');
                add_action('init', 'MM_AuthenticatorService::process_disconnect');
                add_action('init', 'MM_PaymentNotifiers::parse_standalone_request');

                add_action('wp_ajax_mm_stripe_connect_update_creds', 'MM_StripeConnect::process_update_creds');
                add_action('wp_ajax_mm_stripe_connect_refresh', 'MM_StripeConnect::process_refresh_tokens');
                add_action('wp_ajax_mm_stripe_connect_disconnect', 'MM_StripeConnect::process_disconnect');
                add_action('wp_ajax_mm_stripe_connect_save_settings', 'MM_StripeConnect::stripe_connect_save_settings');
            }
        }

        public static function majorVersionUpdateCheck($plugin_data, $r)
        {
            if ($plugin_data["new_version"] ?? false) {
                if (version_compare(self::getPluginVersion(), "3.0.0", "<") && version_compare($plugin_data["new_version"], "3.0.0", ">=")) {
                    $moreInfoUrl = "https://membermouse.com/docs/important-changes-for-membermouse-3-0-read-this-first/";
            ?>
                    <br>
                    <br>
                    <strong>Version 3.0 and above include significant changes affecting many different aspects of the plugin, many of which cannot be rolled back after this upgrade. <br>
                        We strongly advise you to do <a href="<?php echo $moreInfoUrl; ?>" target="_blank">a full site backup (including the database)</a> prior to upgrading and ensure that your initial update is conducted in a staging environment. <br>
                        <a href="<?php echo $moreInfoUrl; ?>" target="_blank">Learn more</a>
                    </strong>
                <?php
                }
            }
        }

        public function filterTranslation($translatedText, $text, $domain)
        {
            return $translatedText;
        }


        public function addMMFooter($content)
        {
            echo "<div style='width: 100%; padding-top: 10px; text-align:right; height: 50px; font-size: 12px;'>" .
                "<a href=\"http://www.membermouse.com?ref=" . urlencode(get_option("siteurl")) . "&src=badge\">" .
                "Membership Software</a> Provided by MemberMouse &nbsp;&nbsp;</div>";
        }

        public function filterContent($content)
        {
            global $wp_query;

            if (!is_feed() && !is_search() && !is_archive()) {
                return $content;
            }

            $protectedContent = new MM_ProtectedContentEngine();
            $postId = $wp_query->query_vars["page_id"];

            if ($protectedContent->protectContent($postId)) {
                $wpPost = get_post($postId);
                setup_postdata($wpPost);

                if (
                    $wpPost && ($wpPost->post_status == "publish" || $wpPost->post_status == "inherit")
                    && ($wpPost->post_type == "post" || $wpPost->post_type == "page" || MM_Utils::isCustomPostType($wpPost->post_type))
                ) {
                    $excerptPosition = strpos($wpPost->post_content, "<!--more");
                    if (($excerptPosition !== false) && ($protectedContent->hasMoreTag($wpPost->post_content))) {
                        return substr($wpPost->post_content, 0, $excerptPosition) . " <a href=\"" . get_permalink($postId) . "\">Read More</a>";
                    }
                    return $content;
                } else {
                    return $content;
                }
            }

            $post = get_post($postId);
            setup_postdata($post);
            $excerptPosition = strpos($post->post_content, "<!--more");
            if (($excerptPosition !== false) && ($protectedContent->hasMoreTag($post->post_content))) {
                return substr($post->post_content, 0, $excerptPosition) . " <a href=\"" . get_permalink($postId) . "\">Read More</a>";
            }

            return __("This content is for members only", "membermouse");
        }


        protected function removeFromUpdateList()
        {
            $current = get_site_transient('update_plugins');
            if (isset($current->response["membermouse/index.php"])) {
                unset($current->response["membermouse/index.php"]);
                set_site_transient('update_plugins', $current);
            }
        }


        protected function addToUpdateList($returnStructure = false)
        {
            $crntMajorVersion = self::getPluginVersion();
            $upgradeVersion = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_UPGRADE_NOTICE);
            if (!empty($upgradeVersion) && $crntMajorVersion != $upgradeVersion) {
                if ($upgradeVersion !== false) {
                    $current = get_site_transient('update_plugins');
                    if (!is_object($current)) {
                        return false;
                    }
                    if (!isset($current->response)) {
                        $current->response = ["membermouse/index.php" => new stdClass()];
                    }
                    if (!isset($current->response["membermouse/index.php"])) {
                        $current->response["membermouse/index.php"] = new stdClass();
                    }
                    $current->response["membermouse/index.php"]->slug = "membermouse";
                    $current->response["membermouse/index.php"]->package = MM_PRETTY_CENTRAL_SERVER_URL . "/major-versions/" . $upgradeVersion . ".zip";
                    $current->response["membermouse/index.php"]->new_version = $upgradeVersion;
                    if ($returnStructure) {
                        return $current->response["membermouse/index.php"];
                    } else {
                        set_site_transient('update_plugins', $current);
                        return true;
                    }
                }
            }
            return false;
        }


        public function updateVersion($a, $b, $c)
        {
            if ((float)phpversion() < 7.0) {
                $this->removeFromUpdateList();
                return $a;
            }

            if ($b == plugin_basename(__FILE__)) {
                $this->addToUpdateList();
            }
            return $a;
        }


        public function checkVersion()
        {
            if ((float)phpversion() >= 7.0) {
                $this->addToUpdateList();
            } else {
                $this->removeFromUpdateList();
            }
        }


        public function getUpgradeIfAvail()
        {
            $mm = $this->addToUpdateList(true);
            if (is_object($mm)) {
                return $mm;
            }

            return false;
        }


        public function preSetSiteTransient($transient)
        {
            $current = $this->getUpgradeIfAvail();
            if (($current !== false) && is_object($current)) {
                $transient->response['membermouse/index.php'] = $current;
            } else {
                $crntMajorVersion = self::getPluginVersion();
                $item = (object) array(
                    'id'            => 'membermouse/index.php',
                    'slug'          => 'membermouse',
                    'plugin'        => 'membermouse/index.php',
                    'new_version'   => $crntMajorVersion,
                    'url'           => '',
                    'package'       => '',
                    'icons'         => array(),
                    'banners'       => array(),
                    'banners_rtl'   => array(),
                    'tested'        => '',
                    'requires_php'  => '',
                    'compatibility' => new stdClass(),
                );
                // Adding the "mock" item to the `no_update` property is required
                // for the enable/disable auto-updates links to correctly appear in UI.
                $transient->no_update['membermouse/index.php'] = $item;
            }

            return $transient;
        }


        public function loadPreviewBar()
        {
            if ((MM_Employee::isEmployee() == true || current_user_can('manage_options')) && !is_admin()) {
                MM_PreviewView::show();
            }
        }


        public function customizeAdminBar()
        {
            if (MM_Employee::isEmployee()) {
                global $wp_admin_bar;

                $wp_admin_bar->add_menu(array(
                    'id'    => 'mm-menu',
                    'title' => '<img src="' . MM_Utils::getImageUrl('mm-logo-svg') . '" style="width:22px; margin-bottom:2px; vertical-align:middle;" />',
                    'href'  => MM_ModuleUtils::getUrl(MM_MODULE_DASHBOARD),
                    'meta'  => array('title' => __('MemberMouse', 'membermouse')),
                ));

                $wp_admin_bar->add_menu(array(
                    "id" => "mm-manage-members",
                    "title" => __("Manage Members", "membermouse"),
                    "href" => MM_ModuleUtils::getUrl(MM_MODULE_MANAGE_MEMBERS),
                    "parent" => "mm-menu"
                ));

                $wp_admin_bar->add_menu(array(
                    "id" => "mm-browse-transactions",
                    "title" => __("Browse Transactions", "membermouse"),
                    "href" => MM_ModuleUtils::getUrl(MM_MODULE_MANAGE_TRANSACTIONS),
                    "parent" => "mm-menu"
                ));

                $wp_admin_bar->add_menu(array(
                    "id" => "mm-upcoming-payments",
                    "title" => __("Upcoming Payments", "membermouse"),
                    "href" => MM_ModuleUtils::getUrl(MM_MODULE_UPCOMING_PAYMENTS),
                    "parent" => "mm-menu"
                ));

                $wp_admin_bar->add_menu(array(
                    "id" => "mm-reporting-suite",
                    "title" => __("Reporting Suite", "membermouse"),
                    "href" => MM_ModuleUtils::getUrl(MM_MODULE_REPORTING),
                    "parent" => "mm-menu"
                ));

                $wp_admin_bar->add_menu(array(
                    "id" => "mm-product-settings",
                    "title" => __("Product Settings", "membermouse"),
                    "href" => MM_ModuleUtils::getUrl(MM_MODULE_PRODUCT_SETTINGS),
                    "parent" => "mm-menu"
                ));
                $wp_admin_bar->add_menu(array(
                    "id" => "mm-support-center",
                    "title" => __("Support Center", "membermouse"),
                    "href" => "https://membermouse.com/docs",
                    "parent" => "mm-menu",
                    "meta" => array("target" => "blank")
                ));
            }

            if (!is_admin() && MM_Preview::previewEnabled() && MM_Employee::isEmployee()) {
                ?>
                <style>
                    #wpadminbar {
                        background: linear-gradient(to top, #373737 0px, #464646 0px) repeat scroll 0 0 #464646;
                        background-image: -webkit-linear-gradient(bottom, #373737 0, #464646 1px);
                        border-bottom: 1px #555555 solid;
                    }

                    #wpadminbar .ab-top-secondary {
                        background: linear-gradient(to top, #373737 0px, #464646 0px) repeat scroll 0 0 #464646;
                        background-image: -webkit-linear-gradient(bottom, #373737 0, #464646 1px);
                        border-bottom: 1px #555555 solid;
                    }

                    body {
                        margin-bottom: 34px;
                    }
                </style>
                <script type='text/javascript'>
                    jQuery(document).ready(function() {
                        mmPreviewJs.hideNonMemberItems();
                    });
                </script>
<?php
            }
        }


        public function loadResources()
        {
            global $wp_scripts;

            $inMembermouseAdmin = MM_Utils::inMembermouseAdmin();

            if (is_admin() && !$inMembermouseAdmin) {
                $menuIconCSS = "#adminmenu #toplevel_page_mmdashboard .wp-menu-image img {padding: 9px 0 0 0px; width: 22px;}";
                wp_add_inline_style('admin-menu', $menuIconCSS);
            }

            $customCssFiles = array();

            if (!is_admin() || $inMembermouseAdmin) {
                $customCssFiles["main"] = 'resources/css/common/mm-main.css';
                $customCssFiles["buttons"] = 'resources/css/common/mm-buttons.css';
            }

            $module = MM_ModuleUtils::getModule();
            $submodule = MM_ModuleUtils::getSubModule();

            if ($module == MM_MODULE_EXTENSIONS && $submodule === MM_Extension::$MM_COURSES_TOKEN) {
                $customCssFiles["courses"] = 'resources/css/admin/mm-courses.css';
                wp_enqueue_style('wp-color-picker');
                wp_enqueue_media();
            }

            $extensionLandingSubmodules = [
                MM_Extension::$PDF_RECEIPTS_TOKEN,
            ];

            if ($module == MM_MODULE_EXTENSIONS && in_array($submodule, $extensionLandingSubmodules)) {
                $customCssFiles["ext_landing"] = 'resources/css/admin/mm-extensions_landing.css';
            }

            if ($module == MM_MODULE_REPORTING) {
                $customCssFiles["reporting"] = 'resources/css/admin/reporting/mm-reporting.css';
            }

            if ($inMembermouseAdmin) {
                $customCssFiles["admin"] = 'resources/css/common/mm-admin.css';
            }

            $useJQueryUI = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_USE_JQUERY_UI);

            if ((!is_admin() && $useJQueryUI == "1") || $inMembermouseAdmin) {
                if (function_exists("wp_scripts")) {
                    wp_scripts();
                }

                $jquery_ui_version = "1.11.2"; //fallback version
                if ($wp_scripts instanceof WP_Scripts) {
                    //if available, get registered script object for jquery-ui, use that to load the appropriate theme version from CDN to match the js version
                    $jqueryUICore = $wp_scripts->query('jquery-ui-core');
                    $jquery_ui_version = isset($jqueryUICore->ver) ? $jqueryUICore->ver : $jquery_ui_version;
                }
                wp_enqueue_style("membermouse-jquery-css", "//ajax.googleapis.com/ajax/libs/jqueryui/{$jquery_ui_version}/themes/" . MM_JQUERY_UI_THEME . "/jquery-ui.css", array(), $jquery_ui_version);
            }
            $version = self::getPluginVersion();

            foreach ($customCssFiles as $cssId => $cssFile) {
                wp_enqueue_style("membermouse-" . $cssId, plugins_url($cssFile, __FILE__), array(), $version);
            }

            $subfolder = ($module == MM_MODULE_REPORTING) ? "reporting/" : "";
            if (file_exists(MM_PLUGIN_ABSPATH . "/resources/css/admin/{$subfolder}mm-" . $module . ".css")) {
                wp_enqueue_style('membermouse-' . $module, plugins_url("resources/css/admin/{$subfolder}mm-" . $module . '.css', __FILE__), array());
            }

            if (!is_admin() || $inMembermouseAdmin) {
                wp_enqueue_style('membermouse-font-awesome', '//maxcdn.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css', array());
            }

            $this->loadJavascript($module);
        }


        public function loadJavascript($module = "")
        {
            $inMembermouseAdmin = MM_Utils::inMembermouseAdmin();
            $submodule = MM_ModuleUtils::getSubModule();

            $url = MM_OptionUtils::getOption("siteurl");
            $adminUrl = admin_url();

            if (isset($_SERVER["HTTP_HOST"])) {
                $thisUrl = $_SERVER["HTTP_HOST"] . $_SERVER["REQUEST_URI"];
                if (strpos($thisUrl, "http") === false) {
                    $thisUrl = (isset($_SERVER["HTTPS"]) && $_SERVER["HTTPS"] == "on") ? "https://{$thisUrl}" : "http://{$thisUrl}";
                }
            }

            if ((isset($_SERVER["HTTPS"]) && ($_SERVER["HTTPS"] == "on")) || (strpos($thisUrl, "https") !== false)) {
                $url = str_replace("http:", "https:", $url);
                $adminUrl = str_replace("http:", "https:", $adminUrl);
            }

            //first include global script
            $version = self::getPluginVersion(); //use plugin major version to control caching
            wp_enqueue_script("membermouse-global", plugins_url("/resources/js/global.js", __FILE__), array('jquery'), $version);
            $javascriptData = array(
                "jsIsAdmin" => $inMembermouseAdmin,
                "adminUrl" => $adminUrl,
                "globalurl" => MM_PLUGIN_URL,
                "ajaxurl" => admin_url('admin-ajax.php'),
                "checkoutProcessingPaidMessage" => MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_CHECKOUT_PAID_MESSAGE),
                "checkoutProcessingFreeMessage" => MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_CHECKOUT_FREE_MESSAGE),
                "checkoutProcessingMessageCSS" => MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_CHECKOUT_MESSAGE_CSS)
            );
            if ($module == MM_MODULE_REPORTING) {
                $javascriptData += array("reportingNoDataMessage" => __("No data was returned", "membermouse"));
            }

            $javascriptData["currencyInfo"] = MM_CurrencyUtil::getActiveCurrencyMetadata();

            wp_localize_script("membermouse-global", "MemberMouseGlobal", $javascriptData);

            $jsFiles = array();
            $userJSDir = "/resources/js/user/";
            $adminJSDir = "/resources/js/admin/";
            $commonJSDir = "/resources/js/common/";

            if ((isLocalInstall() && file_exists(MM_PLUGIN_ABSPATH . "{$commonJSDir}class.js")) || !file_exists(MM_PLUGIN_ABSPATH . "{$commonJSDir}mm-common-core.js")) {
                $commonJsFiles = array(
                    'class.js' => array(),
                    'mm-cache.js' => array(),
                    'mm-main.js' => array('wp-i18n'),
                    'class.ajax.js' => array('class.js'),
                    'mm-dialog.js' => array('class.js'),
                    'mm-core.js' => array('class.js', 'wp-i18n'),
                    'mm-iframe.js' => array('class.js'),
                    'class.form.js' => array('class.js'),
                    'mm-smarttag_library.js' => array('mm-core.js', 'mm-dialog.js', 'wp-i18n'),
                    'mm-payment_utils.js' => array('mm-core.js', 'wp-i18n'),
                );
            } else {
                $commonJsFiles = array('mm-common-core.js' => array('wp-i18n'));
            }

            $useJQueryUI = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_USE_JQUERY_UI);

            $wordPressjQueryUiHandles = array();

            if ($useJQueryUI == "1" || $inMembermouseAdmin) {
                $wordPressjQueryUiHandles[] = 'jquery-ui-accordion';
                $wordPressjQueryUiHandles[] = 'jquery-ui-button';
                $wordPressjQueryUiHandles[] = 'jquery-ui-datepicker';
                $wordPressjQueryUiHandles[] = 'jquery-ui-dialog';
                $wordPressjQueryUiHandles[] = 'jquery-ui-draggable';
                $wordPressjQueryUiHandles[] = 'jquery-ui-droppable';
                $wordPressjQueryUiHandles[] = 'jquery-ui-mouse';
                $wordPressjQueryUiHandles[] = 'jquery-ui-position';
                $wordPressjQueryUiHandles[] = 'jquery-ui-progressbar';
                $wordPressjQueryUiHandles[] = 'jquery-ui-resizable';
                $wordPressjQueryUiHandles[] = 'jquery-ui-selectable';
                $wordPressjQueryUiHandles[] = 'jquery-ui-sortable';
                $wordPressjQueryUiHandles[] = 'jquery-ui-widget';
            }

            if ($module == MM_MODULE_MEMBER_DETAILS_GENERAL) {
                $wordPressjQueryUiHandles[] = 'jquery-ui-autocomplete';
            }

            if ($inMembermouseAdmin) {
                wp_enqueue_script(
                    "membermouse-blockUI",
                    plugins_url(MM_PLUGIN_NAME . "/resources/js/common/jquery.blockUI.js"),
                    array("jquery"),
                    MemberMouse::getPluginVersion(),
                    true
                );
            }

            if ($module == MM_MODULE_EXTENSIONS && $submodule === MM_Extension::$MM_COURSES_TOKEN) {
                $wordPressjQueryUiHandles[] = 'wp-color-picker';
            }

            foreach ($commonJsFiles as $file => $dependencies) {
                if (file_exists(MM_PLUGIN_ABSPATH . "{$commonJSDir}{$file}")) {
                    $jsFiles[plugins_url("{$commonJSDir}{$file}", __FILE__)] = $dependencies;
                }
            }

            if (!is_admin()) {
                $userFiles = array(
                    'mm-preview.js',
                );

                foreach ($userFiles as $file) {
                    if (file_exists(MM_PLUGIN_ABSPATH . "{$userJSDir}{$file}")) {
                        $jsFiles[plugins_url("{$userJSDir}{$file}", __FILE__)] = array();
                    }
                }
            } else {
                $jsFiles[plugins_url("{$adminJSDir}mm-corepages.js", __FILE__)] = array("wp-i18n");
                $jsFiles[plugins_url("{$adminJSDir}mm-accessrights.js", __FILE__)] = array("wp-i18n");

                if (!empty($module)) {
                    $enhancedDatagridCandidates = [MM_MODULE_PRODUCTS, MM_MODULE_BUNDLES, MM_MODULE_COUPONS, MM_MODULE_MEMBERSHIP_LEVELS, MM_MODULE_MANAGE_TRANSACTIONS, MM_MODULE_BROWSE_MEMBERS, MM_MODULE_UPCOMING_PAYMENTS, MM_MODULE_JOBS_QUEUE_LOG, MM_MODULE_REMINDERS];
                    $ehdWithExporterCandidates = [MM_MODULE_MANAGE_TRANSACTIONS, MM_MODULE_BROWSE_MEMBERS];
                    $useEnhancedDatagrid = false;

                    // load JavaScript classes dynamically based on the module
                    $moduleJSDir = $adminJSDir;
                    $jsFileName = $module;
                    if ($module == MM_MODULE_REPORTING) {
                        $jsFiles[plugins_url("{$adminJSDir}mm-reportjsbase.js", __FILE__)] = array("wp-i18n");
                        $moduleJSDir .= "/reports/";
                        $jsFileName = MM_ModuleUtils::getPage();
                    }
                    /**
                     * <link href="https://cdn.jsdelivr.net/npm/select2@4.1.0-rc.0/dist/css/select2.min.css" rel="stylesheet" />
                     * <script src="https://cdn.jsdelivr.net/npm/select2@4.1.0-rc.0/dist/js/select2.min.js"></script>
                     */
                    if (in_array($module, $enhancedDatagridCandidates)) {
                        $useEnhancedDatagrid = true;
                        $ehdViewDependencies = (file_exists(MM_PLUGIN_ABSPATH . "{$commonJSDir}mm-core.js")) ? array("mm-core.js", "wp-i18n") : array("mm-common-core.js", "wp-i18n");
                        $jsFiles[plugins_url("{$adminJSDir}mm-enhanceddatagridview.js", __FILE__)] = $ehdViewDependencies;
                    }

                    if (in_array($module, $ehdWithExporterCandidates)) {
                        $jsFiles[plugins_url("{$adminJSDir}mm-ehdwithexporterview.js", __FILE__)] = array("mm-enhanceddatagridview.js");
                    }

                    if (file_exists(MM_PLUGIN_ABSPATH . "{$moduleJSDir}mm-{$jsFileName}.js")) {
                        $scriptDependencies = $useEnhancedDatagrid ? array('mm-enhanceddatagridview.js', 'wp-i18n') : array('wp-i18n');
                        $jsFiles[plugins_url("{$moduleJSDir}mm-{$jsFileName}.js", __FILE__)] = $scriptDependencies;
                    }
                }
            }

            foreach ($jsFiles as $file_to_include => $dependencies) {
                $scriptname = basename($file_to_include);
                $allDependencies = is_array($dependencies) ? $dependencies : array($dependencies);
                $allDependencies = array_merge(array('membermouse-global'), $allDependencies);
                wp_enqueue_script($scriptname, $file_to_include, $allDependencies, $version);

                if (in_array("wp-i18n", $allDependencies)) {
                    MM_Utils::setScriptTranslations($scriptname);
                }
            }

            foreach ($wordPressjQueryUiHandles as $handle) {
                wp_enqueue_script($handle);
            }
        }

        public function onDeactivate()
        {
            MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "plugin deactivated from index::onDeactivate()");
            MM_MemberMouseService::deactivatePlugin();

            // clean up MM crons
            MM_CronEngine::cleanup();
        }

        public function activationFailed()
        {
            if (isset($_GET[MM_Session::$PARAM_COMMAND_DEACTIVATE]) && isset($_GET[MM_Session::$PARAM_MESSAGE_KEY])) {
                echo "<div class='updated'>";
                echo "<p>" . urldecode($_GET[MM_Session::$PARAM_MESSAGE_KEY]) . "</p>";
                echo "</div>";
                MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "plugin deactivated from index::activationFailed()");
                @deactivate_plugins(MM_PLUGIN_ABSPATH . "/index.php", false);
            }
        }

        public function enqueueAdminNotices()
        {
            $this->checkVersion();
            // Check if database needs migration
            $dbNeedsUpgrade = get_option("mm-option-use-legacy-db", false);
            if ($dbNeedsUpgrade) {
                $title = sprintf(
                    '<i class="fa fa-exclamation-triangle"></i> <strong>%s</strong>',
                    __("Database Upgrade Required", "membermouse"),
                );
                $message = sprintf(
                    '<div style="margin-left:20px;"><p><i class="fa fa-caret-right"></i> %1$s. <a href="%2$s">%3$s</a></p></div>',
                    __("You have upgraded from a previous version of MemberMouse, and your database tables need to be upgraded", "membermouse"),
                    MM_ModuleUtils::getUrl(MM_MODULE_GENERAL_SETTINGS, MM_MODULE_MIGRATE_DB),
                    __("Click here to do this now", "membermouse"),
                );
                MM_AdminNotice::add("db-upgrade-required", $title, $message, MM_AdminNotice::TYPE_ERROR);
            }

            // Check if payment service is installed
            // Check for dependencies
            $algoMissing = false;
            if (!function_exists("hash_hmac")) {
                $algoMissing = true;
                $translatableMsg = __("hash_hmac PHP function and sha256 algorithm are not available on your server. Contact your hosting provide to address this", "membermouse");
            } else {
                $algorithms = hash_algos();
                if (is_array($algorithms) && !in_array("sha256", $algorithms)) {
                    $algoMissing = true;
                    $translatableMsg = __("The sha256 algorithm is not available on your server. Contact your hosting provide to address this", "membermouse");
                }
            }

            if ($algoMissing && ($translatableMsg ?? false)) {
                $activePaymentService = MM_PaymentServiceFactory::getOnsitePaymentService();
                if ($activePaymentService instanceof MM_ScheduledPaymentService) {
                    $errorTitle = sprintf(
                        '<i class="fa fa-exclamation-triangle"></i> <strong>%s</strong>',
                        __("MemberMouse Missing Components", "membermouse"),
                    );
                    $errorMsg = sprintf(
                        '<div style="margin-left:20px;"><p><i class="fa fa-caret-right"></i> %1$s <a href="%2$s">%3$s</a></p></div>',
                        $translatableMsg,
                        'https://membermouse.com/docs/warning-sha-256-and-hash_hmac-are-not-available-on-your-server',
                        __('Learn More', 'membermouse'),
                    );

                    MM_AdminNotice::add("missing-algos", $errorTitle, $errorMsg, MM_AdminNotice::TYPE_ERROR);
                }
            }

            // Check if MemberMouse is running unlicensed
            $license = MM_License::getCurrentLicense();
            if ($license->isUnlicensed() || ($license->getMemberId() === -1)) {
                $title = sprintf(
                    '<i class="fa fa-file-text"></i> <strong>%s</strong>',
                    __('This Version of MemberMouse is Unlicensed', 'membermouse'),
                );
                $messageText = sprintf(
                    __('To activate MemberMouse, please %1$spurchase a license here%2$s.', 'membermouse'),
                    '<a href="' . MM_MemberMouseService::getUpgradeUrl() . '" target="_blank">',
                    '</a>'
                );
                $message = '<div style="margin-left:20px;"><p><i class="fa fa-caret-right"></i>' . $messageText . '</p></div>';
                MM_AdminNotice::add("unlicensed", $title, $message, MM_AdminNotice::TYPE_ERROR, ["behavior" => false]);
            }

            // Check if billing is paused
            if (!MM_LocalBillingEngine::isBillingEnabled()) {
                $content = MM_LocalBillingEngine::getBillingPausedMessage();
                MM_AdminNotice::add("billing-paused", $content["title"], $content["message"], MM_AdminNotice::TYPE_WARNING);
            }

            // Check to see if this is a beta version
            if (MM_IS_BETA == true) {
                $mmVersion = self::getPluginVersion();
                $betaText = __("This is a Beta version of MemberMouse.", "membermouse");
                $notesText = __("Release Notes", "membermouse");
                $title = "MemberMouse {$mmVersion}";
                $message = "<div style='margin-left:20px;'><p><i class=\"fa fa-caret-right\"></i> {$betaText} ";
                $message .= "<a href='http://membermouse.com/beta-release-notes.php?version={$mmVersion}' target='_blank'>Beta {$mmVersion} {$notesText}</a>";
                $message .= "</p></div>";
                MM_AdminNotice::add("beta-version", $title, $message, MM_AdminNotice::TYPE_INFO);
            }

            // Check to see if there's a new version of MM available
            $crntMajorVersion = self::getPluginVersion();
            $upgradeVersion = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_UPGRADE_NOTICE);

            if (!empty($upgradeVersion)) {
                if ((version_compare($upgradeVersion, $crntMajorVersion, ">"))
                    || (MM_IS_BETA == true && version_compare($upgradeVersion, $crntMajorVersion, "=="))
                ) {
                    // verify that customer is on PHP 7.0+
                    if ((float)phpversion() < 7.0) {
                        $messageText = sprintf(
                            __('%1$sWarning%2$s: A new version of MemberMouse is available. In order to upgrade to the latest version of MemberMouse you will need to upgrade to PHP 7.0 or higher.', 'membermouse'),
                            '<strong>',
                            '</strong>',
                        );
                        $message = "<div><p><i class=\"fa fa-exclamation-triangle\"></i> {$messageText}</p></div>";
                        MM_AdminNotice::add("upgrade-notice", false, $message, MM_AdminNotice::TYPE_WARNING, ["behavior" => false]);
                    } else if ((strpos($_SERVER["PHP_SELF"], "plugins.php") === false) && !(isset($_GET["action"]) && ($_GET["action"] == "upgrade-plugin"))) {
                        $message = sprintf(
                            __('%1$sMemberMouse %2$s is available! %3$sPlease update now.%4$s', 'membermouse'),
                            '<p><a href="https://membermouse.com/docs/membermouse-versions" target="_blank">',
                            $upgradeVersion . '</a>',
                            '<a href="plugins.php?plugin_update=membermouse&version=' . $upgradeVersion . '">',
                            '</a></p>',
                        );
                        MM_AdminNotice::add("upgrade-notice", false, $message, MM_AdminNotice::TYPE_INFO, ["behavior" => false]);
                    }
                }
            }

            // Check to see if in Safe Mode
            $safeMode = MM_SafeMode::getMode();

            if ($safeMode == MM_SafeMode::$MODE_ENABLED) {
                $message = sprintf(
                    '<p><i class="fa fa-life-saver"></i> <strong>%1$s</strong> <a href="%2$s">%3$s</a></p>',
                    __("MemberMouse Safe Mode is Enabled", "membermouse"),
                    MM_ModuleUtils::getUrl(MM_MODULE_GENERAL_SETTINGS, MM_MODULE_SAFE_MODE),
                    __("Safe Mode Settings", "membermouse"),
                );
                MM_AdminNotice::add("safe-mode-enabled", false, $message, MM_AdminNotice::TYPE_WARNING, ["behavior" => false]);
            }


            // Check to see if payment subsystem is in test mode or if test data is being used.
            $testPaymentSrvcEnabled = MM_TestPaymentService::isSiteUsingTestService();
            $testDataEnabled = (MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_USE_CHECKOUT_FORM_TEST_DATA) == "1") ? true : false;

            if ($testPaymentSrvcEnabled || $testDataEnabled) {
                $titleText = __("MemberMouse Testing Enabled", "membermouse");
                $title = "<i class=\"fa fa-flask\"></i> <strong>{$titleText}</strong>";
                $message = "<div style='margin-left:20px;'><p>";

                if ($testPaymentSrvcEnabled) {
                    $paymentSettingsUrl = MM_ModuleUtils::getUrl(MM_MODULE_PAYMENT_SETTINGS, MM_MODULE_PAYMENT_METHODS);
                    $errorText = __("Test Payment Service is enabled. All charges will be processed by the test payment service. ", "membermouse");
                    $linkText = __("Payment Method Settings", "membermouse");
                    $message .= "<i class=\"fa fa-caret-right\"></i> {$errorText}";
                    $message .= "<a href='{$paymentSettingsUrl}'>{$linkText}</a><br/>";
                }

                if ($testDataEnabled) {
                    $testDataSettingsUrl = MM_ModuleUtils::getUrl(MM_MODULE_PAYMENT_SETTINGS, MM_MODULE_TEST_DATA);
                    $errorText = __("Test Data is enabled. All checkout forms will be prepopulated with test data.", "membermouse");
                    $linkText = __("Test Data Settings", "membermouse");
                    $message .= "<i class=\"fa fa-caret-right\"></i> {$errorText} <a href='{$testDataSettingsUrl}'>{$linkText}</a><br/>";
                }

                $message .= "</p></div>";
                MM_AdminNotice::add("testing-enabled", $title, $message, MM_AdminNotice::TYPE_WARNING);
            }

            // Check PHP version
            if ((float) phpversion() < 7.4) {
                $titleText = __("Obsolete PHP Version", "membermouse");
                $title = "<i class=\"fa fa-exclamation-triangle\"></i> <strong>{$titleText}</strong>";

                $message = "<div><p>";
                $message .= sprintf(
                    __('Your webserver is running PHP %1$s, which is an obsolete version of PHP. Future versions of MemberMouse require a minimum PHP version of 7.4. View this article to %2$slearn more%3$s.', 'membermouse'),
                    phpversion(),
                    '<a href="https://membermouse.com/docs/end-of-support-for-php-versions-older-than-7-4" target="_blank">',
                    '</a>',
                );
                $message .= "</p></div>";
                MM_AdminNotice::add("obsolete-php", $title, $message, MM_AdminNotice::TYPE_WARNING);
            }

            // Check Stripe Connect status
            MM_StripeConnect::upgrade_notice();
            MM_AuthenticatorService::mm_disconnect_notice();

            // Check for plugins with known compatibility issues
            MM_Utils::getPluginWarnings();
        }

        public function configurePostMeta()
        {
            if (is_admin()) {
                $protectedContentView = new MM_ProtectedContentView();
                add_meta_box('membermouse_post_access', __('MemberMouse Options', 'membermouse'), array($protectedContentView, 'postPublishingBox'), 'post', 'side', 'high');
                add_meta_box('membermouse_post_access', __('MemberMouse Options', 'membermouse'), array($protectedContentView, 'postPublishingBox'), 'page', 'side', 'high');

                // add meta box to custom post types
                $args = array(
                    'public'   => true,
                    '_builtin' => false
                );

                $post_types = get_post_types($args, 'names', 'and');

                foreach ($post_types as $post_type) {
                    add_meta_box('membermouse_post_access', __('MemberMouse Options', 'membermouse'), array($protectedContentView, 'postPublishingBox'), $post_type, 'side', 'high');
                }

                $commonJSDir = "/resources/js/common/";
                $dependency  = ((isLocalInstall() && file_exists(MM_PLUGIN_ABSPATH . "{$commonJSDir}mm-smarttag_library.js")) || !file_exists(MM_PLUGIN_ABSPATH . "{$commonJSDir}mm-common-core.js")) ? array('mm-smarttag_library.js') : array('mm-common-core.js');
                $dependency[] = "wp-i18n";
                $version = self::getPluginVersion(); //use plugin major version to control caching
                wp_enqueue_script('membermouse-postmeta', plugins_url('resources/js/admin/mm-accessrights.js', __FILE__), $dependency, $version);
                wp_enqueue_script('membermouse-corepages', plugins_url('resources/js/admin/mm-corepages.js', __FILE__), $dependency, $version);
                MM_Utils::setScriptTranslations("membermouse-postmeta");
                MM_Utils::setScriptTranslations("membermouse-corepages");
            }
        }

        public function install()
        {
            if (!$this->installerRan) {
                $installer = new MM_Install();
                $installer->activate();
                $this->installerRan = true;
            }
        }

        public function handleAjaxCallback()
        {
            $response = array();
            $data = stripslashes_deep($_REQUEST);

            if (isset($data["module"])) {
                if (isset($data["type"]) && $data["type"] == "displayonly") {
                    $obj = new $data["module"]();
                    $ret = $obj->callMethod($data);

                    if ($ret instanceof MM_Response) {
                        echo $ret->message;
                    } else {
                        echo $ret;
                    }

                    exit();
                } else if (class_exists($data["module"]) && is_subclass_of($data["module"], "MM_View")) {
                    $obj = new $data["module"]();
                    $response = $obj->callMethod($data);
                }
            }

            if (function_exists("wp_json_encode")) {
                echo wp_json_encode($response);
            } else {
                echo json_encode($response);
            }
            exit();
        }

        public function buildAdminMenu()
        {
            if (!isset($_GET[MM_Session::$PARAM_COMMAND_DEACTIVATE])) {
                global $current_user;

                $crntModule = MM_ModuleUtils::getPage();

                if ($crntModule == MM_MODULE_GENERAL_SETTINGS) {
                    add_thickbox();
                }

                if (class_exists("MM_Employee")) {
                    $employee = MM_Employee::findByUserId($current_user->ID);
                    $employee->buildMenu();
                }
            }
        }


        public static function allowJavascriptProtocol($protocols)
        {
            $protocols[] = "javascript";
            return $protocols;
        }


        public static function getPluginVersion()
        {
            return self::$pluginVersion;
        }
    }
    //Allow javascript pseudo-protocol
    add_filter('kses_allowed_protocols', 'MemberMouse::allowJavascriptProtocol');
    $mmplugin = new MemberMouse();
}
