<?php

/**
 * 
 * MemberMouse(TM) (http://www.membermouse.com)
 * (c) MemberMouse, LLC. All rights reserved.
 * 
 * MM_Install handles all install/update related tasks. The master schema is located in data/install_sql.php. 
 * Key things to remember when working with the master schema: 
 * 1. The wordpress dbdelta function is used to keep the schema in sync across versions. 
 *    Formatting is very important, please reference the initial schema when adding new tables
 * 2. PRIMARY KEYs need to be declared on a separate line (vs inline with the column definition. Two spaces need to
 *    follow the words PRIMARY KEY and the definition of that key, otherwise the dbdelta will fail
 * 3. FOREIGN KEYs are not currently supported and can potentially cause dbdeltas to fail.
 * 4. Each column or key definition needs to be on its own line. Lines are delimited by \n (unix-style newline)
 * 5. dbdelta() currently doesn't properly deal with backticks surrounding the column name in KEY definitions
 * 6. dbdelta() will add columns to existing tables, but will not drop removed columns. Those operations will need to be done manually 
 * 
 * With a properly formatted schema, the database structure will be synced to the schema on both install and update, eliminating the need for
 * version-specific update code and different code for installs vs updates. Default data with specified keys (either primary keys or unique column values) 
 * is inserted using INSERT IGNORE sql statements. This handles both insert and update scenarios because duplicate key errors are ignored if the data already exists.
 * 
 */
require_once(ABSPATH . 'wp-admin/includes/upgrade.php');

class MM_Install
{

    public function activate()
    {
        ob_start();
        $error = "";

        // reset minor version if this is a major version upgrade
        $crntMajorVersion = MemberMouse::getPluginVersion();
        $lastMajorVersion = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_MAJOR_VERSION);
        if (!empty($lastMajorVersion) && (version_compare($lastMajorVersion, $crntMajorVersion) < 0)) {
            MM_OptionUtils::setOption(MM_OptionUtils::$OPTION_KEY_MINOR_VERSION, MM_MemberMouseService::$DEFAULT_MINOR_VERSION);
        }

        // If the license url has changed, pause billing at installation when Site URL appears to be a staging or development url
        if (!MM_MemberMouseService::licenseUrlMatchesSiteUrl()) {
            $this->disableBillingForStagingUrls();
        }

        MM_MemberMouseService::authorize();

        if ($this->alterMMTables()) {
            if ($this->insertMMDefaultData()) {
                //if possible, set path to wp-load.php for use in standalone scripts
                $wplFile = ABSPATH . "/wp-load.php";
                $bbFile = dirname(__FILE__) . "/../bootstrap/barebones.php";
                if (file_exists($wplFile) && is_writable($bbFile)) {
                    $contents = file_get_contents($bbFile);
                    $updatedPath = "/* BEGIN WP-LOAD PATH */\n" .
                        "\$wpload_path = '{$wplFile}';\n" .
                        "/* END WP-LOAD PATH */";
                    $contents = preg_replace("/\/\* BEGIN WP-LOAD PATH \*\/\s.*\s\/\* END WP-LOAD PATH \*\//", $updatedPath, $contents);
                    file_put_contents($bbFile, $contents);
                }

                // set new major version
                $crntMajorVersion = MemberMouse::getPluginVersion();
                MM_OptionUtils::setOption(MM_OptionUtils::$OPTION_KEY_MAJOR_VERSION, $crntMajorVersion);

                //update version history
                $versionDescription = $crntMajorVersion;
                $minorVersion = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_MINOR_VERSION);

                if ($minorVersion !== false && intval($minorVersion) > 0) {
                    $versionDescription .= "-" . $minorVersion;
                } else {
                    $versionDescription .= "-" . MM_MemberMouseService::$DEFAULT_MINOR_VERSION;
                }

                $versionRelease = MM_VersionRelease::findByVersion($versionDescription);
                $versionRelease->setVersion($versionDescription);
                $versionRelease->commitData();

                MM_MemberMouseService::activatePlugin();

                // clear major version notice
                $crntMajorVersion = MemberMouse::getPluginVersion();
                $upgradeVersion = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_UPGRADE_NOTICE);
                if (!empty($upgradeVersion) && (version_compare($crntMajorVersion, $upgradeVersion, ">="))) {
                    MM_OptionUtils::setOption(MM_OptionUtils::$OPTION_KEY_UPGRADE_NOTICE, "");
                }

                ob_end_clean();
                return true;
            } else {
                $error = "The MemberMouse installer could not install default data into the database. Verify that the permissions of the database user defined in wp-config.php has all privileges.";
            }
        } else {
            $error = "The MemberMouse installer was unable to alter MemberMouse tables. Verify that the permissions of the database user defined in wp-config.php has all privileges.";
        }

        ob_end_clean();

        // an error occurred so deactivate the plugin
        $this->showError($error);
        exit;
    }


    private function showError($error)
    {
        echo $error;
        @deactivate_plugins(MM_PLUGIN_ABSPATH . "/index.php", false);
    }


    private function alterMMTables()
    {
        global $wpdb;

        $filename = "install_sql";
        require_once(MM_PLUGIN_ABSPATH . "/data/{$filename}.php");

        if (isset($sql) && count($sql) > 0) {
            dbdelta($sql);
        }

        // Since DBDelta doesn't handle dropping column, here's where we can do it manually & safely. We first 
        // check to see if the column even exists. This could happen if a client has deactivated the 
        // plugin for some reason and then decides to reactivate.

        $droppedColumns = array(); //currently empty

        foreach ($droppedColumns as $table => $columns) {
            foreach ($columns as $column) {
                if ($this->hasColumn($table, $column)) {
                    $wpdb->query("ALTER TABLE {$table} DROP COLUMN {$column}");
                }
            }
        }
        return true;
    }


    private function hasColumn($table, $column)
    {
        global $wpdb;

        $sql = $wpdb->prepare("SELECT COUNT(*) FROM information_schema.columns WHERE table_name = %s AND column_name = %s", $table, $column);
        $count = $wpdb->get_var($sql);
        return ($count > 0);
    }

    private function hasRecords($table, $column, $where)
    {
        global $wpdb;

        $sql = $wpdb->prepare("SELECT COUNT(*) FROM {$table} WHERE {$column} = %s", $where);
        $count = $wpdb->get_var($sql);
        return ($count > 0);
    }


    private function insertMMDefaultData()
    {
        global $wpdb, $current_user;

        // insert default option values
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ACCT_SECURITY_ENABLED, MM_OptionUtils::$DEFAULT_ACCT_SECURITY_ENABLED);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ACCT_SECURITY_MAX_IPS, MM_OptionUtils::$DEFAULT_ACCT_SECURITY_MAX_IPS);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ON_LOGIN_USE_WP_FRONTPAGE, MM_OptionUtils::$DEFAULT_ON_LOGIN_USE_WP_FRONTPAGE);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_AFTER_LOGIN_USE_WP_FRONTPAGE, MM_OptionUtils::$DEFAULT_AFTER_LOGIN_USE_WP_FRONTPAGE);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_SHOW_LOGIN_LOGOUT_LINK, MM_OptionUtils::$DEFAULT_SHOW_LOGIN_LOGOUT_LINK);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_HIDE_PROTECTED_MENU_ITEMS, MM_OptionUtils::$DEFAULT_HIDE_PROTECTED_MENU_ITEMS);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_MM_LOGIN_PAGE, MM_OptionUtils::$DEFAULT_USE_MM_LOGIN_PAGE);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_MM_RESET_PASSWORD_PAGE, MM_OptionUtils::$DEFAULT_USE_MM_RESET_PASSWORD_PAGE);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_DFLT_COUNTRY_SELECTION, MM_OptionUtils::$DEFAULT_COUNTRY_ISO);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_FORGOT_PASSWORD_SUBJECT, MM_OptionUtils::$DEFAULT_FORGOT_PASSWORD_SUBJECT);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_FORGOT_PASSWORD_BODY, MM_OptionUtils::$DEFAULT_FORGOT_PASSWORD_BODY);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_AFFILIATE, MM_OptionUtils::$DEFAULT_AFFILIATE);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_SUB_AFFILIATE, MM_OptionUtils::$DEFAULT_SUB_AFFILIATE);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_AFFILIATE_LIFESPAN, MM_OptionUtils::$DEFAULT_AFFILIATE_LIFESPAN);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_LOGIN_TOKEN_LIFESPAN, MM_OptionUtils::$DEFAULT_LOGIN_TOKEN_LIFESPAN);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_CHECKOUT_FORM_TEST_DATA, MM_OptionUtils::$DEFAULT_USE_CHECKOUT_FORM_TEST_DATA);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_PURCHASE_CONFIRMATION_DIALOG_WIDTH, MM_OptionUtils::$DEFAULT_PURCHASE_CONFIRMATION_DIALOG_WIDTH);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_PURCHASE_CONFIRMATION_DIALOG_HEIGHT, MM_OptionUtils::$DEFAULT_PURCHASE_CONFIRMATION_DIALOG_HEIGHT);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_DEFAULT_CHECKOUT_ITEM_TYPE, MM_OptionUtils::$DEFAULT_CHECKOUT_ITEM_TYPE);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_DEFAULT_CHECKOUT_ITEM_ID, MM_OptionUtils::$DEFAULT_CHECKOUT_ITEM_ID);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_SHOW_PREVIEW_BAR, MM_OptionUtils::$DEFAULT_SHOW_PREVIEW_BAR);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_HIDE_ADMIN_BAR, MM_OptionUtils::$DEFAULT_HIDE_ADMIN_BAR);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ENABLE_WP_AUTOP, MM_OptionUtils::$DEFAULT_ENABLE_WP_AUTOP);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ENABLE_USERNAME_CHANGE, MM_OptionUtils::$DEFAULT_ENABLE_USERNAME_CHANGE);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ALLOW_LOGGED_OUT_PURCHASES, MM_OptionUtils::$DEFAULT_ALLOW_LOGGED_OUT_PURCHASES);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ALLOW_DUPLICATE_SUBSCRIPTIONS, MM_OptionUtils::$DEFAULT_ALLOW_DUPLICATE_SUBSCRIPTIONS);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_CHECKOUT_PAID_MESSAGE, MM_OptionUtils::$DEFAULT_CHECKOUT_PAID_MESSAGE);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_CHECKOUT_FREE_MESSAGE, MM_OptionUtils::$DEFAULT_CHECKOUT_FREE_MESSAGE);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_CHECKOUT_MESSAGE_CSS, MM_OptionUtils::$DEFAULT_CHECKOUT_MESSAGE_CSS);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_CURRENCY, MM_OptionUtils::$DEFAULT_CURRENCY);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_JQUERY_UI, MM_OptionUtils::$DEFAULT_USE_JQUERY_UI);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ALLOW_OVERDUE_ACCESS, MM_OptionUtils::$DEFAULT_ALLOW_OVERDUE_ACCESS);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_CAPTCHA_ENABLED, MM_OptionUtils::$DEFAULT_CAPTCHA_ENABLED);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_MM_CSS_RESET_PASSWORD, MM_OptionUtils::$DEFAULT_USE_MM_CSS);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_DIAGNOSTIC_MODE, MM_DiagnosticLog::$MODE_OFF);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_CURRENCY_FORMAT_POSTFIX_ISO, 'false');
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_AFFILIATE_ALIAS, "");
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_SUB_AFFILIATE_ALIAS, "");
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_AFFILIATE_LIFESPAN, "");

        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_SITE_IN_TEST_MODE, "0"); //0 for false, 1 for true
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_CAPTCHA_KEY, "");
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_CAPTCHA_PRIVATE_KEY, "");

        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_FORGET_MEMBER_EMAIL_ADDRESS, "1");
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_FORGET_MEMBER_ADDRESS, "1");
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_FORGET_MEMBER_ADDRESS_COUNTRY, "1");
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_FORGET_MEMBER_ORDER_ADDRESS, "1");
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_FORGET_MEMBER_ORDER_COUNTRY, "1");
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_FORGET_MEMBER_ACTIVITY_LOG, "1");
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_FORGET_MEMBER_CUSTOM_FIELDS, "1");

        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_LIMELIGHT_USE_REBILL_PRODUCT, MM_OptionUtils::$DEFAULT_LIMELIGHT_USE_REBILL_PRODUCT);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_STICKYIO_USE_REBILL_PRODUCT, MM_OptionUtils::$DEFAULT_STICKYIO_USE_REBILL_PRODUCT);
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_LAST_BILLING_SWEEP, 'false');
        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_SCHEDULED_BILLING_ENABLED, MM_OptionUtils::$DEFAULT_SCHEDULED_BILLING_ENABLED);

        // options that vary based on if this is an upgrade versus a new install
        $lastMajorVersion = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_MAJOR_VERSION);

        if (empty($lastMajorVersion)) {
            // set options for new installs
            // for 2.0.7 release
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_DISABLE_EXPLICIT_LINKS, MM_OptionUtils::$DEFAULT_DISABLE_EXPLICIT_LINKS);
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_PURCHASE_LINK_STYLE, MM_OptionUtils::$DEFAULT_PURCHASE_LINK_STYLE);

            // for 2.0.8 release
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_SMARTTAG_VERSION, MM_OptionUtils::$DEFAULT_SMARTTAG_VERSION);

            // for 2.0.9 release
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ACTIVITY_LOG_CLEANUP_ENABLED, MM_OptionUtils::$DEFAULT_ACTIVITY_LOG_CLEANUP_ENABLED);
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ACTIVITY_LOG_CLEANUP_INTERVAL, MM_OptionUtils::$DEFAULT_ACTIVITY_LOG_CLEANUP_INTERVAL);

            // for 2.1.2 release
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_MM_CSS_CHECKOUT, MM_OptionUtils::$DEFAULT_USE_MM_CSS);
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_MM_CSS_MY_ACCOUNT, MM_OptionUtils::$DEFAULT_USE_MM_CSS);
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_MM_CSS_LOGIN, MM_OptionUtils::$DEFAULT_USE_MM_CSS);
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_MM_CSS_FORGOT_PASSWORD, MM_OptionUtils::$DEFAULT_USE_MM_CSS);
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ENABLE_MEMBERSHIP_PRORATION, MM_OptionUtils::$DEFAULT_ENABLE_MEMBERSHIP_PRORATION);

            // for 2.2.5 release
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_DRIP_CONTENT_TIME_SETTING, MM_OptionUtils::$DEFAULT_DRIP_CONTENT_TIME_SETTING);

            // for 2.4.0 release
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_CAPTCHA_VERSION, MM_OptionUtils::$DEFAULT_CAPTCHA_VERSION);
        } else {
            // set options for existing installs
            // for 2.0.7 release
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_DISABLE_EXPLICIT_LINKS, "0");
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_PURCHASE_LINK_STYLE, MM_LINK_STYLE_EXPLICIT);

            // for 2.0.8 release
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_SMARTTAG_VERSION, "2.0");

            // for 2.0.9 release
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ACTIVITY_LOG_CLEANUP_ENABLED, "0");
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ACTIVITY_LOG_CLEANUP_INTERVAL, "");

            // for 2.1.2 release
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_MM_CSS_CHECKOUT, "0");
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_MM_CSS_MY_ACCOUNT, "0");
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_MM_CSS_LOGIN, "0");
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_USE_MM_CSS_FORGOT_PASSWORD, "0");
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_ENABLE_MEMBERSHIP_PRORATION, "0");

            // for 2.2.5 release
            if (class_exists('MM_ProtectedContentEngine')) {
                MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_DRIP_CONTENT_TIME_SETTING, MM_ProtectedContentEngine::$TIME_SETTING_SERVER);
            } else {
                MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_DRIP_CONTENT_TIME_SETTING, MM_OptionUtils::$DEFAULT_DRIP_CONTENT_TIME_SETTING);
            }

            // for 2.4.0 release
            MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_CAPTCHA_VERSION, "2");


            // when upgrading from releases < 2.2.5
            if (version_compare($lastMajorVersion, '2.2.4', '<=')) {
                $wpdb->query("ALTER TABLE " . MM_TABLE_TRANSACTION_LOG . " MODIFY COLUMN order_id BIGINT(20) UNSIGNED NOT NULL");
                $wpdb->query("ALTER TABLE " . MM_TABLE_TRANSACTION_LOG . " MODIFY COLUMN order_item_id BIGINT(20) UNSIGNED");
            }
        }

        MM_Role::addRoles();

        // create default employee account if one doesn't exist, using the current user
        $install_sql = "select * from " . MM_TABLE_EMPLOYEE_ACCOUNTS . " where is_default='1' LIMIT 1;";
        $row = $wpdb->get_row($install_sql);
        if ($row == null) {
            //the current user isn't an employee
            $install_sql = "insert into " . MM_TABLE_EMPLOYEE_ACCOUNTS . " set display_name='%s', email='%s', is_default='%d', role_id='%s', user_id='%d'";
            $displayName = (!empty($current_user->display_name)) ? $current_user->display_name : "Support";
            $wpdb->query($wpdb->prepare($install_sql, $displayName, $current_user->user_email, 1, MM_Role::$ROLE_ADMINISTRATOR, $current_user->ID));
            $emailId = $wpdb->insert_id;
        } else {
            $emailId = $row->id;
        }

        // Install default overdue payment notification emails. Uses $emailId, set above.
        $crntValue = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_OVERDUE_PAYMENT_NOTIFICATION_INSTALLED);

        if (class_exists('MM_Action') && ($crntValue === false || $crntValue === "")) {
            $overdueActionData = [
                "emailToId"     => MM_Action::$CURRENT_MEMBER_PLACEHOLDER,
                "emailFromId"   => $emailId,
                "emailCC"       => "",
                "emailSubject"  => MM_OptionUtils::$DEFAULT_OVERDUE_PAYMENT_SUBJECT,
                "emailBody"     => MM_OptionUtils::$DEFAULT_OVERDUE_PAYMENT_BODY,
            ];
            $overdueStatusData = ["status_id" => MM_Status::$OVERDUE];

            $overdueMembershipEmail = $this->installPushNotification(
                true,
                MM_Event::$MEMBER_STATUS_CHANGE,
                MM_Action::$MM_ACTION_SEND_EMAIL,
                $overdueActionData,
                $overdueStatusData,
            );

            $overdueBundleEmail = $this->installPushNotification(
                true,
                MM_Event::$BUNDLE_STATUS_CHANGE,
                MM_Action::$MM_ACTION_SEND_EMAIL,
                $overdueActionData,
                $overdueStatusData,
            );

            if (MM_Response::isSuccess($overdueMembershipEmail) || MM_Response::isSuccess($overdueBundleEmail)) {
                MM_OptionUtils::setOption(MM_OptionUtils::$OPTION_KEY_OVERDUE_PAYMENT_NOTIFICATION_INSTALLED, "1");
            }
        }

        // Create default membership level. Uses $emailId, set above.
        if (!$this->hasRecords(MM_TABLE_MEMBERSHIP_LEVELS, 'is_default', '1')) {
            $install_sql = "INSERT INTO " . MM_TABLE_MEMBERSHIP_LEVELS . " SET " .
                "	name = 'Free Membership'," .
                "	is_free='1'," .
                "	is_default='1'," .
                "	status='1'," .
                "	description='Default Free Membership'," .
                "	email_subject='%s'," .
                "	email_body='%s'," .
                "	email_from_id='%d'" .
                "";

            $wpdb->query($wpdb->prepare($install_sql, MM_MembershipLevel::$DFLT_EMAIL_SUBJECT, MM_MembershipLevel::$DFLT_EMAIL_BODY, $emailId));
        }

        // create default commission profile
        if (!$this->hasRecords(MM_TABLE_COMMISSION_PROFILES, 'is_default', '1')) {
            $install_sql = "INSERT INTO " . MM_TABLE_COMMISSION_PROFILES . " SET " .
                "	name = 'Standard Commission Profile'," .
                "	is_default='1'," .
                "	description='This is the default commission profile'," .
                "	initial_commission_enabled='1'," .
                "	rebill_commissions_enabled='0'," .
                "	rebill_commission_type='default'," .
                "	rebill_commission_value='0'," .
                "	do_limit_rebill_commissions='0'," .
                "	rebill_commission_limit='0', " .
                "	do_reverse_commissions='1'" .
                "";

            $wpdb->query($install_sql);
        }

        //add smart tags
        $this->addSmartTags();

        $sql = "select count(*) as total from " . MM_TABLE_API_KEYS;
        $row = $wpdb->get_row($sql);

        if ($row->total <= 0) {
            $sql = "insert into " . MM_TABLE_API_KEYS . " set " .
                "name='Default Access', " .
                "api_key='" . MM_Utils::createRandomString(10) . "',	" .
                "api_secret='" . MM_Utils::createRandomString(10) . "', " .
                "status='1'";

            $wpdb->query($sql);
        }

        $sql = array();
        require_once(MM_PLUGIN_ABSPATH . "/data/countries.sql.php");
        $numCountries = count($sql);
        $countryCheckSql = "select count(*) as total from " . MM_TABLE_COUNTRIES;
        $row = $wpdb->get_row($countryCheckSql);
        if ($row->total < $numCountries) {
            foreach ($sql as $query) {
                if ($wpdb->query($query) === false) {
                    return false;
                }
            }
        }
        unset($sql);

        // set default country selections to have all countries selected
        $sql = "SELECT iso, printable_name from " . MM_TABLE_COUNTRIES . " order by printable_name asc";
        $results = $wpdb->get_results($sql, OBJECT_K);
        $defaultSelections = array();

        foreach ($results as $iso => $row) {
            $defaultSelections[$iso] = $iso;
        }

        MM_OptionUtils::setDefaultValue(MM_OptionUtils::$OPTION_KEY_COUNTRY_SELECTIONS, $defaultSelections);

        $sql = array();
        require_once(MM_PLUGIN_ABSPATH . "/data/country_subdivisions.sql.php");
        $numCountrySubdivisions = count($sql);
        $countrySubdivisionCheckSql = "select count(*) as total from " . MM_TABLE_COUNTRY_SUBDIVISIONS;
        $row = $wpdb->get_row($countrySubdivisionCheckSql);
        if ($row->total < $numCountrySubdivisions) {
            foreach ($sql as $query) {
                if ($wpdb->query($query) === false) {
                    return false;
                }
            }
        }
        unset($sql);

        $result = $this->setupCorePages();

        if (!$result) {
            return false;
        }

        $sql = array();

        //email service providers
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_EMAIL_SERVICE_PROVIDERS . " SET provider_name='None', provider_token='default', active='1';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_EMAIL_SERVICE_PROVIDERS . " SET provider_name='ActiveCampaign', provider_token='activecampaign', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_EMAIL_SERVICE_PROVIDERS . " SET provider_name='MailChimp', provider_token='mailchimp', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_EMAIL_SERVICE_PROVIDERS . " SET provider_name='iContact', provider_token='icontact', api_key='" . MM_IContactEmailServiceProvider::$API_KEY . "'";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_EMAIL_SERVICE_PROVIDERS . " SET provider_name='AWeber', provider_token='aweber', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_EMAIL_SERVICE_PROVIDERS . " SET provider_name='GetResponse', provider_token='getresponse', active='0';";

        //affiliate providers
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_AFFILIATE_PROVIDERS . " SET provider_name='None', provider_token='default', active='1';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_AFFILIATE_PROVIDERS . " SET provider_name='iDevAffiliate', provider_token='idevaffiliate';";

        //payment services
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='PAYPAL', name='PayPal', settings='', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='AUTHORIZENET', name='Authorize.net', settings='', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='AUTHORIZENETCIM', name='Authorize.net CIM', settings='', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='BRAINTREE', name='Braintree', settings='', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='CHARGIFY', name='Chargify', settings='', active='0';";

        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='STICKYIO', name='Sticky.io', settings='', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='LIMELIGHT', name='Lime Light (Legacy)', settings='', active='0';";

        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='STRIPE', name='Stripe', settings='', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='TWOCHECKOUT', name='2Checkout', settings='', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='CLICKBANK', name='ClickBank', settings='', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='LITLE', name='Litle', settings='', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='COINBASE', name='Coinbase (Wallet Required)', settings='', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='COINBASEMINIMAL', name='Coinbase (Wallet Ignored)', settings='', active='0';";
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_PAYMENT_SERVICES . " SET token='TEST', name='Test Payment Service', settings='', active='0';";

        //shipping methods
        $sql[] = "INSERT IGNORE INTO " . MM_TABLE_SHIPPING_METHODS . " SET token='FLATRATE', name='Flat Rate', settings='', active='1';"; //active by default

        foreach ($sql as $query) {
            $wpdb->query($query);
        }
        unset($sql);


        //apply any updates that may have been made to the schemas of the active payment services
        $activePaymentServices = MM_PaymentServiceFactory::getAvailablePaymentServices();
        foreach ($activePaymentServices as $ps) {
            $ps->install();
        }

        // Need to search and replace any remnants of [MM_Member_Data name='password'] in 
        // Forgot Password email with the new Reset Password Core Page Link Smart Tag
        $pattern               = "\[mm_member_data name=(['|\"]{1})password(['|\"]{1})\]";
        $forgotPasswordBody = MM_OptionUtils::getOption(MM_OptionUtils::$OPTION_KEY_FORGOT_PASSWORD_BODY);
        $forgotPasswordBody = preg_replace("/{$pattern}/i", "<a href=\"[MM_CorePage_Link type='resetpassword']\">click here to reset your password</a>", $forgotPasswordBody);
        MM_OptionUtils::setOption(MM_OptionUtils::$OPTION_KEY_FORGOT_PASSWORD_BODY, $forgotPasswordBody);

        return true;
    }


    /**
     * Install a push notification
     * @param  bool   $status Whether the push notification is enabled
     * @param  string $eventType  Event Type, @see MM_Event class definition
     * @param  string $actionType Action Type, @see MM_Action class definiton
     * @param  array  $actionData Data for the action:
     *					For $MM_ACTION_SEND_EMAIL - Array keys emailToId, emailFromId, emailCC, emailSubject, emailBody
     *					For $MM_ACTION_NOTIFY_ZAPIER - Array keys zapierMailbox, emailFromId, zapierAdditionalInfo
     *					For $MM_ACTION_CALL_SCRIPT - Array with one key, scriptUrl
     * @param  array  $attributes Attributes for the action, @see MM_PushNotificationEngine::checkEventAttributes()
     * @return object MM_Response indicating whether the action was successfully committed
     */
    public function installPushNotification($status, $eventType, $actionType, $actionData, $attributes = null): object
    {
        $action = new MM_Action();
        $action->setEventType($eventType);
        $action->setActionType($actionType);

        switch ($actionType) {
            case MM_Action::$MM_ACTION_SEND_EMAIL:
                $actionValue = MM_Action::prepareSendEmailValue(
                    $actionData["emailToId"],
                    $actionData["emailFromId"],
                    $actionData["emailCC"],
                    $actionData["emailSubject"],
                    $actionData["emailBody"],
                );
                break;
            case MM_Action::$MM_ACTION_NOTIFY_ZAPIER:
                $actionValue = MM_Action::prepareNotifyZapierValue(
                    $actionData["zapierMailbox"],
                    $actionData["emailFromId"],
                    $actionData["zapierAdditionalInfo"],
                );
                break;
            case MM_Action::$MM_ACTION_CALL_SCRIPT:
                $actionValue = MM_Action::prepareCallScriptValue($actionData["scriptUrl"]);
                break;
            default:
                return new MM_Response("Invalid action type passed to installPushNotification()", MM_Response::$ERROR);
        }

        $action->setStatus($status);
        $action->setActionValue($actionValue);

        if (isset($attributes)) {
            $action->setEventAttributes($attributes);
        }

        return $action->commitData();
    }


    public function setupCorePages()
    {
        global $wpdb, $current_user;

        // Core Page Types
        $core_page_types = [
            ["id" => MM_CorePageType::$MEMBER_HOME_PAGE, "name" => 'Member Home', "visible" => '1'],
            ["id" => MM_CorePageType::$SAVETHESALE, "name" => 'Save the Sale', "visible" => '1'],
            ["id" => MM_CorePageType::$ERROR, "name" => 'Error', "visible" => '1'],
            ["id" => MM_CorePageType::$LOGIN_PAGE, "name" => 'Login', "visible" => '1'],
            ["id" => MM_CorePageType::$FORGOT_PASSWORD, "name" => 'Forgot Password', "visible" => '1'],
            ["id" => MM_CorePageType::$RESET_PASSWORD, "name" => 'Reset Password', "visible" => '1'],
            ["id" => MM_CorePageType::$CHECKOUT, "name" => 'Checkout', "visible" => '1'],
            ["id" => MM_CorePageType::$REDEEM_GIFT, "name" => 'Redeem Gift', "visible" => '1'],
            ["id" => MM_CorePageType::$PAID_CONFIRMATION, "name" => 'Confirmation', "visible" => '1'],
            ["id" => MM_CorePageType::$FREE_CONFIRMATION, "name" => 'Free Confirmation', "visible" => '0'],
            ["id" => MM_CorePageType::$MY_ACCOUNT, "name" => 'My Account', "visible" => '1'],
            ["id" => MM_CorePageType::$LOGOUT_PAGE, "name" => 'Logout', "visible" => '1'],
        ];

        foreach ($core_page_types as $type) {
            $wpdb->query($wpdb->prepare(
                "INSERT IGNORE INTO " . MM_TABLE_CORE_PAGE_TYPES . " (id, name, visible) VALUES (%s, %s, %s)",
                $type['id'],
                $type['name'],
                $type['visible']
            ));
        }

        // Default Core Pages
        $core_pages = [
            ['1', MM_CorePageType::$MEMBER_HOME_PAGE],
            ['2', MM_CorePageType::$SAVETHESALE],
            ['3', MM_CorePageType::$ERROR],
            ['4', MM_CorePageType::$LOGIN_PAGE],
            ['5', MM_CorePageType::$FORGOT_PASSWORD],
            ['6', MM_CorePageType::$CHECKOUT],
            ['7', MM_CorePageType::$PAID_CONFIRMATION],
            ['9', MM_CorePageType::$FREE_CONFIRMATION],
            ['10', MM_CorePageType::$MY_ACCOUNT],
            ['11', MM_CorePageType::$LOGOUT_PAGE],
            ['12', MM_CorePageType::$REDEEM_GIFT],
        ];

        foreach ($core_pages as $page) {
            $wpdb->query($wpdb->prepare(
                "INSERT IGNORE INTO " . MM_TABLE_CORE_PAGES . " (id, core_page_type_id) VALUES (%s, %s)",
                $page[0],
                $page[1]
            ));
        }

        // Install WordPress pages for each CorePage type
        $templateParams = new stdClass();
        $templateParams->resourceDirectory = MM_RESOURCES_URL;
        $mm_template_base = MM_PLUGIN_ABSPATH . "/templates";

        $default_pages = [
            [MM_CorePageType::$MEMBER_HOME_PAGE, 'homepage.html.php', '[MM_Member_Data name="membershipName"] Home Page', 'home'],
            [MM_CorePageType::$SAVETHESALE, 'cancel.html.php', 'Cancel Membership', 'cancel'],
            [MM_CorePageType::$ERROR, 'error.html.php', 'Error', 'mm-error'],
            [MM_CorePageType::$LOGIN_PAGE, 'login.html.php', 'Login', 'login'],
            [MM_CorePageType::$LOGOUT_PAGE, 'logout.html.php', 'Logout', 'logout'],
            [MM_CorePageType::$FORGOT_PASSWORD, 'forgotpassword.html.php', 'Forgot Password', 'forgot-password'],
            [MM_CorePageType::$CHECKOUT, 'checkout.html.php', 'Checkout', 'checkout'],
            [MM_CorePageType::$REDEEM_GIFT, 'redeemgift.html.php', 'Redeem Gift', 'redeem-gift'],
            [MM_CorePageType::$MY_ACCOUNT, 'myaccount.html.php', 'My Account', 'myaccount'],
            [MM_CorePageType::$PAID_CONFIRMATION, 'confirmation.html.php', 'Thank You', 'confirmation'],
        ];

        foreach ($default_pages as $page) {
            $corePageSql = $wpdb->prepare(
                "SELECT COUNT(*) FROM " . MM_TABLE_CORE_PAGES . " WHERE core_page_type_id = %s AND ref_type IS NULL AND page_id IS NULL",
                $page[0]
            );
            $count = $wpdb->get_var($corePageSql);

            if ($count > 0) {
                $content = MM_TEMPLATE::generate($mm_template_base . "/default_templates/" . $page[1], $templateParams);
                $post_name = $this->getPostName($page[3]);

                $wpdb->insert(
                    $wpdb->posts,
                    array(
                        'post_type' => 'page',
                        'post_author' => $current_user->ID,
                        'post_date' => current_time('mysql'),
                        'post_status' => 'publish',
                        'post_content' => $content,
                        'post_title' => $page[2],
                        'post_name' => $post_name
                    ),
                    array('%s', '%d', '%s', '%s', '%s', '%s', '%s')
                );

                $this->linkCorePage($wpdb->insert_id, $page[0]);
            }
        }

        // Handle Reset Password Page separately due to its unique requirements
        $this->setupResetPasswordPage($mm_template_base, $templateParams, $current_user);

        return true;
    }


    private function setupResetPasswordPage($mm_template_base, $templateParams, $current_user)
    {
        global $wpdb;

        // Remove corrupt reset password core pages
        $wpdb->query($wpdb->prepare(
            "DELETE FROM " . MM_TABLE_CORE_PAGES . " WHERE core_page_type_id = %s AND (page_id IS NULL OR page_id = '')",
            MM_CorePageType::$RESET_PASSWORD
        ));

        $corePageSql = $wpdb->prepare(
            "SELECT id, page_id, COUNT(*) as total FROM " . MM_TABLE_CORE_PAGES . " WHERE core_page_type_id = %s",
            MM_CorePageType::$RESET_PASSWORD
        );
        $row = $wpdb->get_row($corePageSql);

        if ($row->total == 0 || $row->page_id == NULL) {
            // Create row in mm_core_pages table
            $wpdb->insert(
                MM_TABLE_CORE_PAGES,
                array('core_page_type_id' => MM_CorePageType::$RESET_PASSWORD),
                array('%s')
            );
            $resetCorePageId = $wpdb->insert_id;

            // Create reset password page
            $content = MM_TEMPLATE::generate($mm_template_base . "/default_templates/resetpassword.html.php", $templateParams);
            $post_name = $this->getPostName('reset-password');

            $wpdb->insert(
                $wpdb->posts,
                array(
                    'post_type' => 'page',
                    'post_author' => $current_user->ID,
                    'post_date' => current_time('mysql'),
                    'post_status' => 'publish',
                    'post_content' => $content,
                    'post_title' => 'Reset Password',
                    'post_name' => $post_name
                ),
                array('%s', '%d', '%s', '%s', '%s', '%s', '%s')
            );

            // Link row in mm_core_pages table to reset password page
            $this->linkCorePage($wpdb->insert_id, $resetCorePageId);
        }
    }


    private function addSmartTags()
    {
        global $wpdb;

        $wpdb->query("DELETE FROM " . MM_TABLE_SMARTTAGS);
        require_once(MM_PLUGIN_ABSPATH . "/data/smarttags.sql.php");
        if (isset($sql) && is_array($sql)) {
            foreach ($sql as $query) {
                if ($wpdb->query($query) === false) {
                    return false;
                }
            }
        }

        //clear global SmartTag cache
        MM_OptionUtils::setOption(MM_OptionUtils::$OPTION_KEY_SMARTTAGS);

        return true;
    }

    private function getPostName($suggestedName = "")
    {
        global $wpdb;
        $sql = "select count(*) as total from {$wpdb->posts} where post_name='" . $suggestedName . "'";
        $row = $wpdb->get_row($sql);
        if ($row->total > 0) {
            $sql = "select post_name from {$wpdb->posts} where post_name like '" . $suggestedName . "-%'";
            $rows = $wpdb->get_results($sql);

            $index = 1;
            while (true) {
                $newName = $suggestedName . "-" . $index;
                $hasName = false;
                foreach ($rows as $row) {
                    if ($row->post_name == $newName) {
                        $hasName = true;
                    }
                }
                if (!$hasName) {
                    return $newName;
                }
                $index++;
            }
        }
        return $suggestedName;
    }


    private function linkCorePage($page_id, $id)
    {
        global $wpdb;
        $sql = "update " . MM_TABLE_CORE_PAGES . " set page_id='{$page_id}' where id='{$id}'";

        if (!$wpdb->query($sql)) {
            echo $sql;
            return false;
        }
    }

    private function field_exists($column, $table)
    {
        global $wpdb;
        $sql = "describe {$table}";
        $rows = $wpdb->get_results($sql);

        for ($i = 0; $i < count($rows); $i++) {
            if ($column == $rows[$i]->Field) {
                return true;
            }
        }

        return false;
    }

    private function getFieldType($column, $table)
    {
        global $wpdb;
        $sql = "describe {$table}";
        $rows = $wpdb->get_results($sql);

        $numRows = count($rows);
        for ($i = 0; $i < $numRows; $i++) {
            if ($column == $rows[$i]->Field) {
                return $rows[$i]->Type;
            }
        }

        return false;
    }

    private function disableBillingForStagingUrls()
    {
        $siteUrl = strtolower(site_url());
        $parsedUrl = wp_parse_url($siteUrl);
        $shouldDisableBilling = false;

        // Check hostname
        $hostname = $parsedUrl['host'] ?? '';
        if (in_array($hostname, ['localhost', '127.0.0.1'])) {
            $shouldDisableBilling = true;
        }

        // Check if the subdomain is 'staging' (if there is a subdomain)
        $subdomains = explode('.', $hostname);
        if ((count($subdomains) > 2) && ($subdomains[0] === 'staging')) {
            $shouldDisableBilling = true;; // Any one condition is true
        }

        // Check if the TLD is '.dev' or '.local'
        $tld = end($subdomains);
        if (in_array($tld, ['dev', 'local'])) {
            $shouldDisableBilling = true;  // Any one condition is true
        }

        if ($shouldDisableBilling) {
            MM_LocalBillingEngine::disableBilling();
            MM_DiagnosticLog::log(MM_DiagnosticLog::$MM_ERROR, "Potential development/staging URL detected during installation, billing paused");
        }
    }


    public static function getRequired2xTableNames()
    {
        $requiredTables = [
            'mm_actions',
            'mm_affiliate_partner_payouts',
            'mm_affiliate_provider_mappings',
            'mm_affiliate_providers',
            'mm_affiliate_rebill_commissions',
            'mm_api_keys',
            'mm_applied_bundles',
            'mm_bundle_categories',
            'mm_bundle_products',
            'mm_bundles',
            'mm_card_on_file',
            'mm_commission_profiles',
            'mm_core_page_types',
            'mm_core_pages',
            'mm_countries',
            'mm_country_subdivisions',
            'mm_coupon_restrictions',
            'mm_coupon_usage',
            'mm_coupons',
            'mm_custom_field_data',
            'mm_custom_field_options',
            'mm_custom_fields',
            'mm_diagnostic_log',
            'mm_email_provider_bundle_mappings',
            'mm_email_provider_mappings',
            'mm_email_service_providers',
            'mm_employee_accounts',
            'mm_flatrate_shipping_options',
            'mm_log_captcha',
            'mm_log_events',
            'mm_login_token',
            'mm_membership_level_categories',
            'mm_membership_level_products',
            'mm_membership_levels',
            'mm_order_item_access',
            'mm_order_items',
            'mm_orders',
            'mm_payment_services',
            'mm_posts_access',
            'mm_products',
            'mm_queued_scheduled_events',
            'mm_report_data_cache',
            'mm_scheduled_events',
            'mm_scheduled_payments',
            'mm_sessions',
            'mm_shipping_methods',
            'mm_smarttag_groups',
            'mm_smarttags',
            'mm_social_login_linked_profiles',
            'mm_social_login_providers',
            'mm_top_level_domains',
            'mm_transaction_key',
            'mm_transaction_log',
            'mm_user_data',
            'mm_user_defined_pages',
            'mm_version_releases'
        ];
        return $requiredTables;
    }


    public static function getConditional2xTableNames()
    {
        $conditionalTables = [
            'mm_authorizenet_aim_transactions',
            'mm_authorizenet_arb_subscriptions',
            'mm_authorizenet_arb_subscription_history',
            'mm_authorizenet_pending_overdue_subscriptions',
            'mm_paypal_ipn_log',
            'mm_paypal_subscr_links',
            'mm_coinbase_ipn_log',
            'mm_coinbase_subscr_links',
            'mm_coinbase_button_links',
            'mm_coinbaseminimal_ipn_log',
            'mm_coinbaseminimal_button_links',
            'mm_coinbaseminimal_scheduled_events',
            'mm_coinbaseminimal_transactions',
            'mm_stripe_customer_links',
            'mm_stripe_charges',
            'mm_stripe_memberships',
            'mm_stripe_subscription_payments',
            'mm_stripe_webhooks',
            'mm_stripe_products',
            'mm_stripe_coupons',
            'mm_clickbank_product_links',
            'mm_clickbank_ipn_log',
            'mm_clickbank_order_item_links',
            'mm_braintree_customer_links',
            'mm_braintree_charges',
            'mm_authorizenet_cim_customer_links',
            'mm_authorizenet_cim_charges',
            'mm_limelight_products',
            'mm_limelight_shipping_methods',
            'mm_limelight_subscription_links',
            'mm_limelight_charges',
            'mm_limelight_ipn_log',
            'mm_stickyio_products',
            'mm_stickyio_shipping_methods',
            'mm_stickyio_subscription_links',
            'mm_stickyio_charges',
            'mm_stickyio_ipn_log',
            'mm_test_cardonfile',
            'mm_scheduled_payment_job_links',
            'mm_reminders',
            'mm_reminders_products',
            'mm_scheduled_reminders',
            'mm_resque_job_links',
        ];
        return $conditionalTables;
    }


    public static function tableExists($tableName)
    {
        global $wpdb;

        $query = $wpdb->prepare("SHOW TABLES LIKE %s", $wpdb->esc_like($tableName));
        return ($wpdb->get_var($query) === $tableName);
    }
}
