<?php

namespace membermouse\resque;

use \DateTime;
use \Exception;

if (!defined('ABSPATH')) {
    die('You are not allowed to call this page directly.');
}

class Resque
{
    public $config;

    public function __construct()
    {
        // Setup job configuration
        $this->config = apply_filters('resque_jobs_config', (object)array(
            'status'  => (object)array(
                'pending'  => 'pending',
                'complete' => 'complete',
                'failed'   => 'failed',
                'working'  => 'working'
            ),
            'worker' => (object)array(
                'interval' => self::minutes(1),
                'retry_after' => self::minutes(30) // Standard retries after a failure
            ),
            'cleanup' => (object)array(
                'num_retries'             => 5, // "num_retries" before transactions fail
                'interval'                => self::hours(1),
                'retry_after'             => self::hours(1), // Purely for zombie jobs left in a bad state
                'delete_completed_after'  => self::days(2),
                'delete_failed_after'     => self::days(30)
            )
        ));

        // Set a wp-cron
        add_filter('cron_schedules', array($this, 'intervals'));
        add_action(SUB_PREFIX . 'resque_jobs_worker', array($this, 'worker'));
        add_action(SUB_PREFIX . 'resque_jobs_cleanup', array($this, 'cleanup'));

        if (!wp_next_scheduled(SUB_PREFIX . 'resque_jobs_worker')) {
            wp_schedule_event(time() + $this->config->worker->interval, 'resque_jobs_interval', SUB_PREFIX . 'resque_jobs_worker');
        }

        if (!wp_next_scheduled(SUB_PREFIX . 'resque_jobs_cleanup')) {
            wp_schedule_event(time() + $this->config->cleanup->interval, 'resque_jobs_cleanup_interval', SUB_PREFIX . 'resque_jobs_cleanup');
        }
    }

    public function intervals($schedules)
    {
        $schedules['resque_jobs_interval'] = array(
            'interval' => $this->config->worker->interval,
            'display' => __('WP Resque Jobs Worker', 'membermouse')
        );

        $schedules['resque_jobs_cleanup_interval'] = array(
            'interval' => $this->config->cleanup->interval,
            'display' => __('WP Resque Jobs Cleanup', 'membermouse')
        );

        return $schedules;
    }

    public function worker()
    {
        $max_run_time = 45;
        $start_time = time();

        // We want to allow for at least 15 seconds of buffer
        while (((time() - $start_time) <= $max_run_time) and
            ($job = $this->next_job())
        ) {
            try {
                $this->work($job);
                if (isset($job->class)) {
                    $obj = JobFactory::fetch($job->class, $job);
                    //resqueUtils::debug_log(sprintf(__('Starting Job - %1$s (%2$s): %3$s', 'membermouse'), $job->id, $job->class, resqueUtils::object_to_string($obj)));
                    $obj->perform(); // Run the job's perform method
                    //resqueUtils::debug_log(sprintf(__('Job Completed - %1$s (%2$s)', 'membermouse'), $job->id, $job->class));
                    $this->complete($job); // When we're successful we complete the job
                } else {
                    $this->fail($job, __('No class was specified in the job config', 'membermouse'));
                    //resqueUtils::debug_log(__('Job Failed: No class', 'membermouse'));
                }
            } catch (Exception $e) {
                $this->fail($job, $e->getMessage());
                //resqueUtils::debug_log(sprintf(__('Job Failed: %s', 'membermouse'), $e->getMessage()));
            }
        }
    }

    public function cleanup()
    {
        global $wpdb;

        // Retry lingering jobs
        $query = "UPDATE " . self::jobs_table_name() . " 
                 SET status = %s
               WHERE status IN (%s,%s)
                 AND tries <= %d
                 AND TIMESTAMPDIFF(SECOND,lastrun,%s) >= %d";
        $query = $wpdb->prepare(
            $query,
            $this->config->status->pending, // Set status to pending
            $this->config->status->working, // if status = working or
            $this->config->status->failed, // status = failed and
            $this->config->cleanup->num_retries, // number of tries <= num_retries
            self::db_now(),
            $this->config->cleanup->retry_after // and the correct number of seconds since lastrun has elapsed
        );
        $wpdb->query($query);

        // Delete completed jobs that have been in the system for over a day?
        $query = "DELETE FROM " . self::jobs_table_name() . " 
               WHERE status = %s
                 AND TIMESTAMPDIFF(SECOND,lastrun,%s) >= %d";
        $query = $wpdb->prepare(
            $query, // Delete jobs
            $this->config->status->complete, // which have a status = complete
            self::db_now(),
            $this->config->cleanup->delete_completed_after // and the correct number of seconds since lastrun has elapsed
        );
        $wpdb->query($query);

        // Delete jobs that have been retried and are still in a working state
        $query = "DELETE FROM " . self::jobs_table_name() . " 
               WHERE tries > %d
                 AND TIMESTAMPDIFF(SECOND,lastrun,%s) >= %d";
        $query = $wpdb->prepare(
            $query, // Delete jobs
            $this->config->cleanup->num_retries, // which have only been 'n' retries
            self::db_now(),
            $this->config->cleanup->delete_failed_after // and the correct number of seconds since lastrun has elapsed
        );
        $wpdb->query($query);
    }

    /** Returns a full list of all the pending jobs in the queue */
    public function queue()
    {
        global $wpdb;

        $query = "
      SELECT * FROM " . self::jobs_table_name() . " 
       WHERE status = %s
         AND runtime <= %s
       ORDER BY priority ASC, runtime ASC
    ";
        $query = $wpdb->prepare($query, $this->config->status->pending, self::db_now());

        return $wpdb->get_results($query, OBJECT);
    }

    public function next_job()
    {
        global $wpdb;

        $query = "SELECT * FROM " . self::jobs_table_name() . " 
               WHERE status = %s
                 AND runtime <= %s
               ORDER BY priority ASC, runtime ASC
               LIMIT 1";
        $query = $wpdb->prepare($query, $this->config->status->pending, self::db_now());

        return $wpdb->get_row($query, OBJECT);
    }

    public function enqueue_in($in, $classname, $args = array(), $priority = 10)
    {
        $when = time() + $this->interval2seconds($in);
        return $this->enqueue($classname, $args, $when, $priority);
    }

    public function enqueue_at($at, $classname, $args = array(), $priority = 10)
    {
        $when = $at;
        return $this->enqueue($classname, $args, $when, $priority);
    }

    public function enqueue($classname, $args = array(), $when = 'now', $priority = 10)
    {
        global $wpdb;

        if ($when === 'now') {
            $when = time();
        }

        $config = array(
            'runtime' => gmdate('c', $when),
            'firstrun' => gmdate('c', $when),
            'priority' => $priority,
            'tries' => 0,
            'class' => $classname,
            'args' => json_encode($args),
            'reason' => '',
            'status' => $this->config->status->pending,
            'lastrun' => gmdate('c')
        );

        // returns the job id to dequeue later if necessary
        return $this->create_record(self::jobs_table_name(), $config, true);
    }

    public function dequeue($job_id)
    {
        if ($job_id == 0) {
            return;
        }

        global $wpdb;
        return $this->delete_records(self::jobs_table_name(), array('id' => $job_id));
    }

    public function work($job)
    {
        global $wpdb;

        $args = array(
            'status'  => $this->config->status->working,
            'tries'   => $job->tries + 1,
            'lastrun' => gmdate('c')
        );

        $this->update_record(self::jobs_table_name(), $job->id, $args);
    }

    public function retry($job, $reason = '')
    {
        global $wpdb;

        $when = time() + $this->config->worker->retry_after;

        $args = array(
            'status'  => $this->config->status->pending,
            'runtime' => gmdate('c', $when),
            'reason'  => $reason
        );

        $this->update_record(self::jobs_table_name(), $job->id, $args);
    }

    public function complete($job)
    {
        global $wpdb;

        $args = array(
            'status' => $this->config->status->complete,
            'reason' => ''
        );

        $this->update_record(self::jobs_table_name(), $job->id, $args);

        //create a permanent entry in the completed jobs table
        $completed_job = $args + get_object_vars($job);
        $this->create_record(self::completed_jobs_table_name(), $completed_job, false);

        //remove job from queue table after it is archived
        $this->dequeue($job->id);

        do_action(SUB_PREFIX . 'resque_job_complete', (object)$completed_job);
    }

    public function fail($job, $reason = '')
    {
        global $wpdb;

        // We fail and then re-enqueue for an hour later 5 times before giving up
        if ($job->tries >= $this->config->cleanup->num_retries) {
            $args = array(
                'status' => $this->config->status->failed,
                'reason' => $reason
            );

            $this->update_record(self::jobs_table_name(), $job->id, $args);

            //create a permanent entry in the failed jobs table
            $failed_job = $args + get_object_vars($job);
            $this->create_record(self::failed_jobs_table_name(), $failed_job, false);

            //remove job from queue table after it is archived
            $this->dequeue($job->id);

            do_action(SUB_PREFIX . 'resque_job_failed', (object)$failed_job);
        } else {
            $this->retry($job, $reason);
        }
    }

    private function interval2seconds($interval)
    {
        $units = array('m', 'h', 'd', 'w', 'M', 'y');
        $seconds = 0;

        foreach ($units as $u) {
            preg_match_all("/(\d+){$u}/", $interval, $matches);
            if (isset($matches[1])) {
                foreach ($matches[1] as $m) {
                    if ($u == 'm') {
                        $seconds += self::minutes($m);
                    } elseif ($u == 'h') {
                        $seconds += self::hours($m);
                    } elseif ($u == 'd') {
                        $seconds += self::days($m);
                    } elseif ($u == 'w') {
                        $seconds += self::weeks($m);
                    } elseif ($u == 'M') {
                        $seconds += self::months($m);
                    } elseif ($u == 'y') {
                        $seconds += self::years($m);
                    }
                }
            }
        }

        return $seconds;
    }

    public function unschedule_events()
    {
        $timestamp = wp_next_scheduled(SUB_PREFIX . 'resque_jobs_worker');
        wp_unschedule_event($timestamp, SUB_PREFIX . 'resque_jobs_worker');

        $timestamp = wp_next_scheduled(SUB_PREFIX . 'resque_jobs_cleanup');
        wp_unschedule_event($timestamp, SUB_PREFIX . 'resque_jobs_cleanup');
    }

    // Utils
    private static function ts_to_mysql_date($ts, $format = 'Y-m-d H:i:s')
    {
        return gmdate($format, $ts);
    }

    private static function db_now($format = 'Y-m-d H:i:s')
    {
        return self::ts_to_mysql_date(time(), $format);
    }

    public static function minutes($n = 1)
    {
        return $n * 60;
    }

    public static function hours($n = 1)
    {
        return $n * self::minutes(60);
    }

    public static function days($n = 1)
    {
        return $n * self::hours(24);
    }

    public static function weeks($n = 1)
    {
        return $n * self::days(7);
    }

    public static function months($n, $base_ts = false, $backwards = false, $day_num = false)
    {
        $base_ts = empty($base_ts) ? time() : $base_ts;

        $month_num  = gmdate('n', $base_ts);
        $day_num    = ((int) $day_num < 1 || (int) $day_num > 31) ? gmdate('j', $base_ts) : $day_num;
        $year_num   = gmdate('Y', $base_ts);
        $hour_num   = gmdate('H', $base_ts);
        $minute_num = gmdate('i', $base_ts);
        $second_num = gmdate('s', $base_ts);

        // We're going to use the FIRST DAY of month for our calc date, then adjust the day of month when we're done
        // This allows us to get the correct target month first, then set the right day of month afterwards
        try {
            $calc_date = new DateTime("{$year_num}-{$month_num}-1 {$hour_num}:{$minute_num}:{$second_num}", new DateTimeZone("UTC"));
        } catch (Exception $e) {
            return 0;
        }

        if ($backwards) {
            $calc_date->modify("-{$n} month");
        } else {
            $calc_date->modify("+{$n} month");
        }

        $days_in_new_month = $calc_date->format('t');

        // Now that we have the right month, let's get the right day of month
        if ($days_in_new_month < $day_num) {
            $calc_date->modify("last day of this month");
        } elseif ($day_num > 1) {
            $add_days = ($day_num - 1); // $calc_date is already at the first day of the month, so we'll minus one day here
            $calc_date->modify("+{$add_days} day");
        }

        // If $backwards is true, this will most likely be a negative number so we'll use abs()
        return abs($calc_date->getTimestamp() - $base_ts);
    }

    public static function years($n, $base_ts = false, $backwards = false, $day_num = false, $month_num = false)
    {
        $base_ts = empty($base_ts) ? time() : $base_ts;

        $day_num    = ((int) $day_num < 1 || (int) $day_num > 31) ? gmdate('j', $base_ts) : $day_num;
        $month_num  = ((int) $month_num < 1 || (int) $month_num > 12) ? gmdate('n', $base_ts) : $month_num;
        $year_num   = gmdate('Y', $base_ts);
        $hour_num   = gmdate('H', $base_ts);
        $minute_num = gmdate('i', $base_ts);
        $second_num = gmdate('s', $base_ts);

        try {
            $calc_date = new DateTime("{$year_num}-{$month_num}-{$day_num} {$hour_num}:{$minute_num}:{$second_num}", new DateTimeZone("UTC"));
        } catch (Exception $e) {
            return 0;
        }

        if ($backwards) {
            $calc_date->modify("-{$n} year");
        } else {
            $calc_date->modify("+{$n} year");
        }

        // If we're counting from Feb 29th on a Leap Year to a non-leap year we need to minus 1 day
        // or we'll end up with a March 1st date
        if ($day_num == 29 && $month_num == 2 && $calc_date->format('L') == 0) {
            $calc_date->modify('-1 day');
        }

        // If $backwards is true, this will most likely be a negative number so we'll use abs()
        return abs($calc_date->getTimestamp() - $base_ts);
    }

    public static function jobs_table_name()
    {
        global $wpdb;
        return apply_filters(SUB_PREFIX . 'jobs_table_name', $wpdb->prefix . SUB_PREFIX . 'jobs');
    }

    public static function completed_jobs_table_name()
    {
        global $wpdb;
        return apply_filters(SUB_PREFIX . 'completed_jobs_table_name', $wpdb->prefix . SUB_PREFIX . 'completed_jobs');
    }

    public static function failed_jobs_table_name()
    {
        global $wpdb;
        return apply_filters(SUB_PREFIX . 'failed_jobs_table_name', $wpdb->prefix . SUB_PREFIX . 'failed_jobs');
    }

    public function create_record($table, $args, $record_created_at = true)
    {
        global $wpdb;

        $cols = array();
        $vars = array();
        $values = array();

        $i = 0;

        foreach ($args as $key => $value) {
            if ($key == 'id' && empty($value)) {
                // To prevent issue with SQL MODE NO_AUTO_VALUE_ON_ZERO
                continue;
            }

            $cols[$i] = "`$key`";
            if (is_numeric($value) and preg_match('!\.!', $value)) {
                $vars[$i] = '%f';
            } else if (is_int($value) or is_bool($value)) {
                $vars[$i] = '%d';
            } else {
                $vars[$i] = '%s';
            }

            if (is_bool($value)) {
                $values[$i] = $value ? 1 : 0;
            } else if (is_array($value) || is_object($value)) {
                $values[$i] = serialize($value);
            } else {
                $values[$i] = $value;
            }

            $i++;
        }

        if ($record_created_at) {
            $cols[$i] = 'created_at';
            $vars[$i] = "'" . date('c') . "'";
        }

        if (empty($cols))
            return false;

        $cols_str = implode(',', $cols);
        $vars_str = implode(',', $vars);

        $query = "INSERT INTO {$table} ({$cols_str}) VALUES ({$vars_str})";
        $query = $wpdb->prepare($query, $values);

        // This will fail to insert the row if NO_AUTO_VALUE_ON_ZERO is set in the database SQL_MODE
        $query_results = $wpdb->query($query);

        if ($query_results) {
            return $wpdb->insert_id;
        } else {
            return false;
        }
    }

    public function update_record($table, $id, $args)
    {
        global $wpdb;

        if (empty($args) or empty($id))
            return false;

        $set = '';
        $values = array();
        foreach ($args as $key => $value) {
            if (empty($set)) {
                $set .= ' SET';
            } else {
                $set .= ',';
            }

            $set .= " `{$key}`=";

            if (is_numeric($value) and preg_match('!\.!', $value)) {
                $set .= "%f";
            } else if (is_int($value) or is_bool($value)) {
                $set .= "%d";
            } else {
                $set .= "%s";
            }

            if (is_bool($value)) {
                $values[] = $value ? 1 : 0;
            } else if (is_array($value) || is_object($value)) {
                $values[] = serialize($value);
            } else {
                $values[] = $value;
            }
        }

        $values[] = $id;
        $query = "UPDATE {$table}{$set} WHERE id=%d";
        $query = $wpdb->prepare($query, $values);
        $wpdb->query($query);

        return $id;
    }

    public function delete_records($table, $args)
    {
        global $wpdb;
        extract($this->get_where_clause_and_values($table, $args));

        $query = "DELETE FROM {$table}{$where}";
        if (!empty($values)) {
            $query = $wpdb->prepare($query, $values);
        }

        return $wpdb->query($query);
    }

    public function get_where_clause_and_values($table, $args)
    {
        $where = '';
        $values = array();
        foreach ($args as $key => $value) {
            $where .= (empty($where) ? ' WHERE' : ' AND');
            $where .= " `{$key}`=";
            $where .= $this->get_column_prepare_type($table, $key, $value);

            if (is_bool($value)) {
                $values[] = $value ? 1 : 0;
            } else {
                $values[] = $value;
            }
        }

        return compact('where', 'values');
    }

    public function get_column_prepare_type($table, $column, $value = null)
    {
        $info = $this->get_column_info($table);

        $patterns = (object)array(
            'integer' => '/(bit|binary|bool|boolean|tinyint|smallint|mediumint|integer|int|bigint)/',
            'float' => '/(real|double|float|dec|decimal|numeric)/',
            'string' => '/(datetime|date|timestamp|time|year|tinyblob|blob|mediumblob|longblob|varchar|char|varbinary|tinytext|text|mediumtext|longtext|enum|set)/'
        );

        $prepare_type = '%s';
        if (isset($info[$column])) {
            if (preg_match($patterns->integer, strtolower($info[$column]['DATA_TYPE']))) {
                $prepare_type = '%d';
            } else if (preg_match($patterns->float, strtolower($info[$column]['DATA_TYPE']))) {
                $prepare_type = '%f';
            }
        } else {
            if (is_numeric($value) and preg_match('!\.!', $value)) {
                $prepare_type = '%f';
            } else if (is_int($value) or is_bool($value)) {
                $prepare_type = '%d';
            }
        }

        return $prepare_type;
    }

    public function get_column_info($table)
    {
        global $wpdb;

        if (false === ($table_columns_info = get_transient('_' . SUB_PREFIX . 'resque_table_columns_info'))) {
            $table_columns_info = array();
        }

        if (!isset($table_columns_info[$table])) {
            $q = $wpdb->prepare(
                "
          SELECT *
            FROM INFORMATION_SCHEMA.COLUMNS
           WHERE table_name = %s
        ",
                $table
            );

            $columns_info = $wpdb->get_results($q, ARRAY_A);

            $table_columns_info[$table] = array();
            foreach ($columns_info as $c) {
                $table_columns_info[$table][$c['COLUMN_NAME']] = $c;
            }

            set_transient('_' . SUB_PREFIX . 'resque_table_columns_info', $table_columns_info);
        }

        return $table_columns_info[$table];
    }
}
