<?php

/** This file will load the appropriate wp-resque database and files you'll
 * need to use the job queue in your application
 *
 * We'll just include this file somewhere in the boot up sequence in
 * MemberMouse, like so:
 *
 * require_once('lib/resque/autoloader.php');
 *
 * Once this is loaded we should be able to do the following
 * wherever we want in the code:
 *
 * $job = new \membermouse\resque\EmailJob();
 * $job->values  = $values;
 * $job->subject = $subject;
 * $job->body    = $body;
 * $job->class   = get_class($this);
 * $job->to      = $this->to;
 * $job->headers = $this->headers;
 * $job->use_template = $use_template;
 * $job->content_type = $content_type;
 * $job->enqueue();
 *
 */

namespace membermouse\resque;

const SUB_PREFIX = 'mm_';
define("JOBS_PATH", dirname(__FILE__) . '/jobs');

class Init
{
    // When changes are made to the jobs table, increment schema here
    const DB_VERSION = 1;

    public function __construct()
    {
        $this->load_db();
        $this->load_classes();

        new Resque();
    }

    public function jobs_table_name()
    {
        global $wpdb;
        return apply_filters(SUB_PREFIX . 'jobs_table_name', $wpdb->prefix . SUB_PREFIX . 'jobs');
    }

    public function completed_jobs_table_name()
    {
        global $wpdb;
        return apply_filters(SUB_PREFIX . 'completed_jobs_table_name', $wpdb->prefix . SUB_PREFIX . 'completed_jobs');
    }

    public function failed_jobs_table_name()
    {
        global $wpdb;
        return apply_filters(SUB_PREFIX . 'failed_jobs_table_name', $wpdb->prefix . SUB_PREFIX . 'failed_jobs');
    }

    private function load_db()
    {
        global $wpdb;

        //make sure dbDelta function is accessible
        require_once(ABSPATH . 'wp-admin/includes/upgrade.php');

        $old_db_version = get_option(SUB_PREFIX . 'resque_db_version', 0);
        if ($old_db_version < self::DB_VERSION) {
            $char_col = "";
            $collation = $wpdb->get_row("SHOW FULL COLUMNS FROM {$wpdb->posts} WHERE field = 'post_content'");

            if (isset($collation->Collation)) {
                $charset = explode('_', $collation->Collation);

                if (is_array($charset) && count($charset) > 1) {
                    $charset = $charset[0]; //Get the charset from the collation
                    $char_col = "DEFAULT CHARACTER SET {$charset} COLLATE {$collation->Collation}";
                }
            }

            if (empty($char_col)) {
                $char_col = $wpdb->get_charset_collate();
            }

            $jobs = [];
            $jobs[] =
                "CREATE TABLE {$this->jobs_table_name()} (
          id bigint(20) NOT NULL auto_increment,
          runtime datetime NOT NULL,
          firstrun datetime NOT NULL,
          lastrun datetime DEFAULT NULL,
          priority bigint(20) DEFAULT 10,
          tries bigint(20) DEFAULT 0,
          class varchar(255) NOT NULL,
          batch varchar(255) DEFAULT NULL,
          args text,
          reason text,
          status varchar(255) DEFAULT 'pending',
          created_at datetime NOT NULL,
          PRIMARY KEY  (id),
          KEY job_runtime (runtime),
          KEY job_firstrun (firstrun),
          KEY job_lastrun (lastrun),
          KEY job_status (status(191)),
          KEY job_priority (priority),
          KEY job_tries (tries),
          KEY job_class (class(191)),
          KEY job_batch (batch(191)),
          KEY job_created_at (created_at)
        ) {$char_col};";

            $jobs[] =
                "CREATE TABLE {$this->completed_jobs_table_name()} (
          id bigint(20) NOT NULL,
          runtime datetime NOT NULL,
          firstrun datetime NOT NULL,
          lastrun datetime DEFAULT NULL,
          priority bigint(20) DEFAULT 10,
          tries bigint(20) DEFAULT 0,
          class varchar(255) NOT NULL,
          batch varchar(255) DEFAULT NULL,
          args text,
          reason text,
          status varchar(255) DEFAULT 'pending',
          created_at datetime NOT NULL,
          PRIMARY KEY  (id),
          KEY job_runtime (runtime),
          KEY job_firstrun (firstrun),
          KEY job_lastrun (lastrun),
          KEY job_priority (priority),
          KEY job_tries (tries),
          KEY job_created_at (created_at)
        ) {$char_col};";

            $jobs[] =
                "CREATE TABLE {$this->failed_jobs_table_name()} (
          id bigint(20) NOT NULL,
          runtime datetime NOT NULL,
          firstrun datetime NOT NULL,
          lastrun datetime DEFAULT NULL,
          priority bigint(20) DEFAULT 10,
          tries bigint(20) DEFAULT 0,
          class varchar(255) NOT NULL,
          batch varchar(255) DEFAULT NULL,
          args text,
          reason text,
          status varchar(255) DEFAULT 'pending',
          created_at datetime NOT NULL,
          PRIMARY KEY  (id),
          KEY job_runtime (runtime),
          KEY job_firstrun (firstrun),
          KEY job_lastrun (lastrun),
          KEY job_priority (priority),
          KEY job_tries (tries),
          KEY job_created_at (created_at)
        ) {$char_col};";

            foreach ($jobs as $job) {
                dbDelta($job);
            }
            update_option(SUB_PREFIX . 'resque_db_version', self::DB_VERSION);
        }
    }

    private function load_classes()
    {
        $path = dirname(__FILE__);
        require_once("{$path}/Resque.php");
        require_once("{$path}/Job.php");
        require_once("{$path}/JobFactory.php");

        foreach (JobFactory::paths() as $path) {
            foreach (@glob("{$path}/*Job.php", GLOB_NOSORT) as $job_file) {
                require_once($job_file);
            }
        }
    }
}

// Start Job Processor / Scheduler
new Init();
